/*
  Copyright 2009,2010,2014
  John Culp, All Rights Reserved
*/

#include <stdlib.h>
#include <string.h>

#include <iostream>
#include <ios>
#include <iomanip>

#include "AssemFile.hh"
#include "AssemError.hh"

using namespace std;

AssemFile::AssemFile(void)
{}

AssemFile::~AssemFile(void)
{}

bool AssemFile::LoadFile(istream &rfin)
{
  //load the file
  char lineBuffer[512];
  unsigned cLine  = 1;
  while(rfin)
    {
      memset(lineBuffer, '\0', 512);
      rfin.getline(lineBuffer, 511);
      if(!rfin) break;
      lineBuffer[511] = '\0';

      Line line;
      line.lineBuffer = lineBuffer;
      line.num = cLine;
      if( !line.lineBuffer.empty() )
	lineList.push_back(line);

      cLine++;
    };

  cout << "done loading file" << endl;
  return true;
}

void AssemFile::OutputList(std::ostream &rOut)
{
  unsigned long cLine = 0;
  list<Line>::iterator cPlace = lineList.begin();
  while( cPlace != lineList.end() )
    {
      Line &rLine = *cPlace;
      rOut << cLine << ": " << rLine.address << "    " << rLine.lineBuffer << '\n';
      cLine++;
      cPlace++;
    };
}

bool AssemFile::ParseFile(void)
{
  list<Line>::iterator cPlace = lineList.begin();
  while( cPlace != lineList.end() )
    {
      Line &rLine = *cPlace;

      try{
	ParseFileLine(rLine);
      }catch(AssemError asError)
	{
	  cerr << "caught assem error: " << asError.error << " on line: " << rLine.num << endl;
	  return false;
	};

      cPlace++;
    };

  cout << "done parsing file" << endl;

  return true;
}

void AssemFile::ParseFileLine(Line &rLine)
{
  //truncate all lines begining with the first ';' so we can ignore comments
  for(unsigned ii= 0; ii < rLine.lineBuffer.size(); ii++)
    {
      if( rLine.lineBuffer[ii] == ';' )
	{
	  rLine.lineBuffer = rLine.lineBuffer.substr(0, ii);
	}
    };

  //if the line is completly blank (or has only whitespace characters), skip it
  if( rLine.Blank() ) return;

  //convert all characters to capital
  //rLine.ToCaps(rLine.lineBuffer);

  if( rLine.lineBuffer[0] == '\t' ) { // a (probable) instruction
    unsigned startPlace = 1;
    unsigned endPlace = 0;
    if( !ParseValue(rLine.lineBuffer, startPlace, endPlace) )
      {
	throw( AssemError("failed to find insn") );
      }else
      {
	rLine.insn = rLine.lineBuffer.substr(1, endPlace-1);
	rLine.ToCaps(rLine.insn);
	//cout << "set instruction to: "  << rLine.insn  << " " << endPlace << endl;
	if( rLine.insn == "CALL" || rLine.insn == "RETURN" || rLine.insn == "FNSETUP" ||
	    rLine.insn == "PUSH" || rLine.insn == "PULL" ||
	    rLine.insn == "LI" )
	  rLine.lt = Line::LT_INSN_MACRO;
	else
	  rLine.lt = Line::LT_INSN;
	
	//parse as many args as exist (and we could possibly use)
	unsigned startPlace = endPlace;
	if( ParseValue(rLine.lineBuffer, startPlace, endPlace) )
	  {
	    rLine.argA = rLine.lineBuffer.substr(startPlace, endPlace - startPlace);
	    unsigned startPlace = endPlace;
	    if( ParseValue(rLine.lineBuffer, startPlace, endPlace) )
	      {
		rLine.argB = rLine.lineBuffer.substr(startPlace, endPlace - startPlace);
		unsigned startPlace = endPlace;
		if( ParseValue(rLine.lineBuffer, startPlace, endPlace) )
		  {
		    rLine.argC = rLine.lineBuffer.substr(startPlace, endPlace - startPlace);
		  };
	      };
	  }
      }
  } else { // either an empty line or a label
    unsigned startPlace, endPlace;
    
    if( rLine.lineBuffer.substr(0,12) == ".SEGMENT_END")
      {
	rLine.lt = Line::LT_SEGMENT_END;
      }else
      if( rLine.lineBuffer.substr(0,8) == ".SEGMENT" )
	{
	  startPlace = 8;
	  if( !ParseValue( rLine.lineBuffer, startPlace, endPlace) ) throw( AssemError("unable to find segment name") );
	  rLine.lt = Line::LT_SEGMENT;
	rLine.seg_name = rLine.lineBuffer.substr(startPlace, endPlace-startPlace);

	startPlace = endPlace;
	bool res = ParseValue( rLine.lineBuffer, startPlace, endPlace);
	if(res)
	  {
	    char *pTail;
	    const char *pStart = rLine.lineBuffer.c_str()+startPlace;
	    long increment = strtol(pStart, &pTail, 10);
	    if( pTail == pStart) throw( AssemError("unable to parse segment increment") );

	    rLine.seg_increment = increment;
	  };
      }else
	if( rLine.lineBuffer.substr(0, 4) == ".ORG" )
	{
	  startPlace = 4;
	  if( !ParseValue( rLine.lineBuffer, startPlace, endPlace) ) throw( AssemError("unable to find org value") );
	  string num_string = rLine.lineBuffer.substr(startPlace, endPlace-startPlace);
	  const char *pStart = num_string.c_str();
	  const char *pEnd = 0;
	  
	  long value = strtol(pStart, const_cast<char**>(&pEnd), 10);
	  if(pEnd == pStart) throw( AssemError("unable to parse org value") );
	  
	  rLine.lt = Line::LT_ORG;
	  rLine.org_address = value;
	}else
	if( !isspace(rLine.lineBuffer[0]) ) //assume a label
	  { 
	    rLine.lt = Line::LT_LABEL;
	    
	    unsigned startPlace = 0;
	    unsigned endPlace = 0;
	    if( ParseValue(rLine.lineBuffer, startPlace, endPlace) )
	      {
		rLine.label = rLine.lineBuffer.substr(0, endPlace);
		cout << "found label: " << rLine.label << endl;
	      }else
	      {
		throw( AssemError("failed to find label") );
	      };
	  };   
  };
}

bool AssemFile::ParseValue(std::string &rStr, unsigned &rStart, unsigned &rEnd)
{
  bool startFound = false;
  unsigned start = rStart;

  for(unsigned ii = start; ii < rStr.size(); ii++)
    {
      if( !isspace( rStr[ii] ) )
	{
	  start = ii;
	  startFound = true;
	  break;
	};
    };

  if(!startFound)
    {
      rEnd = start;
      return false;
    };

  for(unsigned ii = start; ii < rStr.size(); ii++)
    {
      if( isspace(rStr[ii]) )
	{
	  rStart = start;
	  rEnd = ii;
	  return true;
	}
    };

  //ran into end  of string
  rStart = start;
  rEnd = rStr.size();
  return true;
};

