/*
  Copyright 2009,2014
  John Culp, All Rights Reserved
*/

#include <stdlib.h>

#include <iostream>
#include <iomanip>
#include <sstream>

#include "Assembler.hh"

using namespace std;;

Assembler::Assembler(void) :
  AssemFile()
{}

Assembler::~Assembler(void)
{}

bool Assembler::DoAssem(std::ostream &rfout)
{
  long lastSegmentAddress = 0;
  long lastSegmentIncrement = 1;

  long cAddress = 0;
  long c_increment = 1;

  list<Line>::iterator cPlace = lineList.begin();
  while( cPlace != lineList.end() )
    {
      Line &rLine = *cPlace;
      rLine.address = cAddress; //by default address is set here
      //cout << rLine.num << " set address: " << rLine.address << " " << cAddress << endl;

      switch(rLine.lt)
	{
	case Line::LT_ORG:
	  cAddress = rLine.org_address;
	  rLine.address = cAddress; //reset this line's address
	  break;
	case Line::LT_INSN:
	  cAddress += c_increment;
	  break;
	case Line::LT_INSN_MACRO:
	  if( rLine.insn == "LI" ) {
	    cAddress += (2*c_increment);
	  } else {
	    if(rLine.insn == "CALL" ) {
	      cAddress += (3*c_increment);
	    } else {
	      if( rLine.insn == "RETURN" ) {
		cAddress += (9*c_increment);
	      } else {
		if( rLine.insn == "PUSH" ) {
		  cAddress += (2*c_increment);
		} else {
		  if(rLine.insn == "PULL") {
		    cAddress += (2*c_increment);
		  }else{
		    if( rLine.insn == "FNSETUP" ){
		      cAddress += 8*c_increment;
		    }else
		    throw( AssemError("unknown assembler macro") );
		  };
		};
	      };
	    }
	  }
	  break;
	case Line::LT_LABEL:
	  break; //nothing required except address.
	case Line::LT_SEGMENT:
	  lastSegmentAddress = cAddress;
	  lastSegmentIncrement = c_increment;
	  cAddress = 0;
	  c_increment = rLine.seg_increment;
	  break;
	case Line::LT_SEGMENT_END:
	  cAddress = lastSegmentAddress;
	  c_increment = lastSegmentIncrement;	  
	  break;
	default:
	  break; //do nothing
	};

      cPlace++;
    };  

  cout << "done generating addresses" << endl;

  cPlace = lineList.begin();
  try {
    while( cPlace != lineList.end() )
      {
	Line &rLine = *cPlace;
	debugLine = rLine;
	if( rLine.lt == Line::LT_INSN )
	  {
	    AssembleInsnLine(rLine, rfout);
	  }else
	  if( rLine.lt == Line::LT_INSN_MACRO )
	    AssembleInsnMacroLine(rLine, rfout);

	cPlace++;
      };
  }catch(AssemError error)
    {
      unsigned cLine = 0;
      string linec;
      if( cPlace != lineList.end() )
	{
	  cLine = (*cPlace).num;
	  linec = (*cPlace).lineBuffer;
	};
      cerr << "caught following error line: " << cLine << endl;
      cerr << "error: " << error.error << endl;
      cerr << "error: line contents: " << linec << endl;
      cerr << "bailing out now" << endl;
      return false;
    };

  return true;
}

long Assembler::LookupAddress(const std::string &rLabel) const
{
  list<Line>::const_iterator cPlace = lineList.begin();
  while( cPlace != lineList.end() )
    {
      const Line &rLine = *cPlace;

      if( rLine.lt == Line::LT_LABEL )
	if( rLine.label == rLabel )
	  {
	    //cerr << "got matching label!" << endl;
	    return rLine.address;
	  };
      
      cPlace++;
    };

  throw( AssemError("no matching label name found") );
  //return 0;
}

unsigned long Assembler::TxtToCmd(const std::string &rTxt) const
{
  if( rTxt == "NOP" ) return MN_NOP;

  if( rTxt == "LUI" ) return MN_LUI;

  if( rTxt == "LLI" ) return MN_LLI;

  if( rTxt == "ADD" ) return MN_ADD;

  if( rTxt == "SUB" ) return MN_SUB;

  if( rTxt == "ADDL" ) return MN_ADDL;

  if( rTxt == "AND" ) return MN_AND;

  if( rTxt == "OR" ) return MN_OR;

  if( rTxt == "XOR" ) return MN_XOR;

  if( rTxt == "NOT" ) return MN_NOT;

  if( rTxt == "BSET" ) return MN_BSET;

  if( rTxt == "BCLR" ) return MN_BCLR;

  if( rTxt == "RSL" ) return MN_RSL;

  if( rTxt == "RSR" ) return MN_RSR;

  if( rTxt == "MUL" ) return MN_MUL;

  if( rTxt == "CMP" ) return MN_CMP;

  if( rTxt == "SRI" ) return MN_SRI;

  if( rTxt == "SRR" ) return MN_SRR;

  if( rTxt == "LRI" ) return MN_LRI;

  if( rTxt == "LRR" ) return MN_LRR;

  if( rTxt == "SEQZ") return MN_SEQZ;

  if( rTxt == "SNEQZ") return MN_SNEQZ;

  if( rTxt == "SBSET") return MN_SBSET;

  if( rTxt == "SBCLR") return MN_SBCLR;

  if( rTxt == "JI") return MN_JI;

  if( rTxt == "JR") return MN_JR;

  if( rTxt == "HLT" ) return MN_HLT;

  throw( AssemError( string("unknown assembler insn: ") + rTxt ) );
  //return 0xFFFF;
}

unsigned long Assembler::TxtToReg(const std::string &rTxt) const
{
  if( rTxt == "%A" ) return 0;
  if( rTxt == "%B" ) return 1;
  if( rTxt == "%C" ) return 2;
  if( rTxt == "%D" ) return 3;
  if( rTxt == "%E" ) return 4;
  if( rTxt == "%F" ) return 5;
  if( rTxt == "%G" ) return 6;
  if( rTxt == "%H" ) return 7;

  throw( AssemError("unknown register specification") );
}

/*
unsigned long Assembler::TxtToULong(const std::string &rTxt) const
{
  const char *pStart = rTxt.c_str();
  char *pTail = 0;

  unsigned long value = strtoul(pStart, &pTail, 10);
  if(pTail == pStart) throw( AssemError( string("unable to parse value from string: ") + rTxt ) );

  if( (pTail - pStart) != rTxt.length() )
    throw( AssemEror("trailing trash after parsing ulong") );

  return value;
}

long Assembler::TxtToLong(const std::string &rTxt) const
{
  const char *pStart = rTxt.c_str();
  char *pTail = 0;

  long value = strtol(pStart, &pTail, 10);
  if(pTail == pStart) throw( AssemError( string("unable to parse value from string: ") + rTxt ) );

  if( (pTail - pStart) != rTxt.length() )
    throw( AssemEror("trailing trash after parsing ulong") );

  return value;
}
*/
long long Assembler::TxtToValueLL(const std::string &rTxt) const
{
  if( rTxt.length() == 0)
    throw( AssemError("attempt to parse value from zero length string") );

  try{
    long long dest = LookupAddress(rTxt);
    return dest;
  }catch(...){}; //try something else

  cout << "on line: " << debugLine.num << " failed to find address of: " << rTxt << " desperately trying something else..." << endl;
  cout << "line text is: " << debugLine.lineBuffer << endl;

  const char *pStart = rTxt.c_str();
  char *pTail = 0;

  if( rTxt[ rTxt.length() - 1 ] == 'f' )
    {
      //floating point constant
      float value = strtod(pStart, &pTail);
      if( pTail == pStart )
	throw( AssemError("failed to parse float") );

      if( (pTail - pStart) != rTxt.length()-1 )
	throw( AssemError("trailing trash after parsing float") );

      //? will negative fp numbers get sign extended through top 32 bits?
      return *((unsigned long*)(&value));
    };

  if( rTxt[ rTxt.length() - 1 ] == 'h' )
    {
      //hex constant
      unsigned long value = strtoul(pStart, &pTail, 16);
      if( pTail == pStart )
	throw( AssemError("failed to parse hex number") );

      if( (pTail - pStart) != rTxt.length()-1 )
	throw( AssemError("trailing trash after parsing hex number") );

      return value;
    };

  //parse a standard signed integer value, stored as signed complements
  long value = strtol(pStart, &pTail, 10);
  if( pTail == pStart )
    throw( AssemError("failed to parse long") );

  if( (pTail - pStart) != rTxt.length() )
    throw( AssemError("trailing trash after long") );

  return value;
}

unsigned long Assembler::TxtToImm(const std::string &rTxt) const
{
  long long value = TxtToValueLL(rTxt);

  if( value < 0)
    {
      if( -value > 32768 )
	throw( AssemError("value out of range for imm: < -32768") );

      long negValue = value;
      unsigned long full_value = *((unsigned long*)&negValue);
      return 0xFFFF&full_value;
    }else
    {
      if( value > 65535 )
	throw( AssemError("value out of range for imm: > 65535") );

      long posValue = value;
      return posValue;
    };
}

unsigned long Assembler::TxtToValue(const std::string &rTxt) const
{
  long long value = TxtToValueLL(rTxt);
  return value;

  if( value < 0)
    {
      if( -value > 0x8000000 )
	throw( AssemError("value out of range for negative 32bit value: < -2^32/2") );

      long negValue = value;
      return *((unsigned long*)&negValue);
    }else
    {
      if( value > 0xFFFFFFFF )
	throw( AssemError("value out of rage for positive 32 bit value: > 2^32") );

      unsigned long posValue = value;
      return posValue;
    };
}

/*
unsigned long Assembler::ParseTarget(const std::string &rArg) const
{
  unsigned long dest = 0;

  try{
    dest = TxtToULong(rArg);
  }catch(...)
    {
      try{
	dest = LookupAddress(rArg);
      }catch(...)
	{
	  throw( AssemError( string("unable to get number or valid label for target address from string: ") + rArg ) );
	}
    };

  return dest;
};
*/

void Assembler::AssembleInsnMacroLine(const Line &rLine, ostream &rOut) const
{
  Line fakeInsn;

  if(rLine.insn == "LI" )
    {
      unsigned long arg = TxtToValue( rLine.argB );
      unsigned long lword = arg&0xFFFF;
      unsigned long uword = (arg&0xFFFF0000)>>16;

      ostringstream lstr;
      ostringstream ustr;

      lstr << lword;
      ustr << uword;

      fakeInsn.insn = "LLI";
      fakeInsn.argA = rLine.argA;
      fakeInsn.argB = lstr.str();
      AssembleInsnLine(fakeInsn, rOut);

      fakeInsn.insn = "LUI"; //store the return address
      fakeInsn.argA = rLine.argA;
      fakeInsn.argB = ustr.str();
      AssembleInsnLine(fakeInsn, rOut);

      return;
    };

  if(rLine.insn == "CALL" ) {
    //take the address of this 'instruction' and add 7. this will be the return address
    unsigned long retAddr = rLine.address;
    retAddr += 3;
    ostringstream valStr;
    valStr << retAddr;

    fakeInsn.insn = "LI"; //note actually two instructions
    fakeInsn.argA = "%G";
    fakeInsn.argB = valStr.str();
    AssembleInsnMacroLine(fakeInsn, rOut);

    fakeInsn.insn = "JI";
    fakeInsn.argA = rLine.argA;
    fakeInsn.argB = "";
    fakeInsn.argC = "";
    AssembleInsnLine(fakeInsn, rOut);
    return;
  };

  if(rLine.insn == "FNSETUP" ) {
    //store the return address on the stack
    fakeInsn.insn = "SRR";
    fakeInsn.argA = "%G";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "0";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "SRR";
    fakeInsn.argA = "%A";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "-1";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "SRR";
    fakeInsn.argA = "%B";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "-2";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "SRR";
    fakeInsn.argA = "%C";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "-3";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "SRR";
    fakeInsn.argA = "%D";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "-4";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "SRR";
    fakeInsn.argA = "%E";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "-5";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "SRR";
    fakeInsn.argA = "%F";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "-6";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "ADDL";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "-7";
    AssembleInsnLine(fakeInsn, rOut);

    /* the stack pointer again points to an empty slot in memory */
    return;
  };

  if( rLine.insn == "RETURN" ) {
    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%F";
    fakeInsn.argC = "1";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%E";
    fakeInsn.argC = "2";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%D";
    fakeInsn.argC = "3";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%C";
    fakeInsn.argC = "4";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%B";
    fakeInsn.argC = "5";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%A";
    fakeInsn.argC = "6";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%G";
    fakeInsn.argC = "7";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "ADDL"; //modify the stack pointer
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "7";
    AssembleInsnLine(fakeInsn, rOut);

    //jump to the return address
    fakeInsn.insn = "JR";
    fakeInsn.argA = "%G";
    fakeInsn.argB = "";
    fakeInsn.argC = "";
    AssembleInsnLine(fakeInsn, rOut);

    return;
  };

  if( rLine.insn == "PUSH" ) {
    fakeInsn.insn = "SRR";
    fakeInsn.argA = rLine.argA;
    fakeInsn.argB = "%H";
    fakeInsn.argC = "0";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "ADDL";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "-1";
    AssembleInsnLine(fakeInsn,rOut);
    return;
  };

  if(rLine.insn == "PULL") {
    //the stack pointer normally points to a blank location. to get
    //to the top value on the stack, we need to decrement the pointer first
    fakeInsn.insn = "ADDL";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "1";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%H";
    fakeInsn.argB = rLine.argA;
    fakeInsn.argC = "0";
    AssembleInsnLine(fakeInsn, rOut);
    return;
  };

  throw( AssemError("unhandled macro instruction") );
}

void Assembler::AssembleInsnLine(const Line &rLine, ostream &rOut) const
{
  unsigned long cmd = TxtToCmd(rLine.insn);
  unsigned long insn;
  switch(cmd)
    {
    case MN_NOP:
      insn = 0;
      break;
    case MN_LUI:
      HandleLUI(rLine, insn);
      break;
    case MN_LLI:
      HandleLLI(rLine, insn);
      break;
    case MN_ADD:
      HandleADD(rLine, insn);
      break;
    case MN_SUB:
      HandleSUB(rLine, insn);
      break;
    case MN_ADDL:
      HandleADDL(rLine, insn);
      break;
    case MN_AND:
      HandleAND(rLine, insn);
      break;
    case MN_OR:
      HandleOR(rLine, insn);
      break;
    case MN_XOR:
      HandleXOR(rLine, insn);
      break;
    case MN_NOT:
      HandleNOT(rLine, insn);
      break;
    case MN_BSET:
      HandleBSET(rLine, insn);
      break;
    case MN_BCLR:
      HandleBCLR(rLine, insn);
      break;
    case MN_RSL:
      HandleRSL(rLine, insn);
      break;
    case MN_RSR:
      HandleRSR(rLine, insn);
      break;
    case MN_MUL:
      HandleMUL(rLine, insn);
      break;
    case MN_CMP:
      HandleCMP(rLine, insn);
      break;
    case MN_SRI:
      HandleSRI(rLine, insn);
      break;
    case MN_SRR:
      HandleSRR(rLine, insn);
      break;
    case MN_LRI:
      HandleLRI(rLine, insn);
      break;
    case MN_LRR:
      HandleLRR(rLine, insn);
      break;
    case MN_SEQZ:
      HandleSEQZ(rLine, insn);
      break;
    case MN_SNEQZ:
      HandleSNEQZ(rLine, insn);
      break;
    case MN_SBSET:
      HandleSBSET(rLine, insn);
      break;
    case MN_SBCLR:
      HandleSBCLR(rLine, insn);
      break;
    case MN_JI:
      HandleJI(rLine, insn);
      break;
    case MN_JR:
      HandleJR(rLine, insn);
      break;
    case MN_HLT:
      HandleHLT(rLine, insn);
      break;
    default:
      throw( AssemError("unhandled assembler mneumonic") );
    };

  rOut << hex << setw(8) << setfill('0') << insn << endl;
}

void Assembler::HandleLUI(const Line & rLine, unsigned long & rinsn) const
{
  //source and target register
  unsigned long reg = TxtToReg(rLine.argA);
  unsigned long imm = TxtToImm(rLine.argB);
  rinsn = (MN_LUI<<I_CMD_O) | (reg<<I_SRCA_O) | (reg<<I_DEST_O) | (imm&0xFFFF);
}

void Assembler::HandleLLI(const Line & rLine, unsigned long & rinsn) const
{
  //source and target register
  unsigned long reg = TxtToReg(rLine.argA);
  unsigned long imm = TxtToImm(rLine.argB);
  rinsn = (MN_LLI<<I_CMD_O) | (reg<<I_SRCA_O) | (reg<<I_DEST_O) | (imm&0xFFFF);
}

void Assembler::HandleADD(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long r1 = TxtToReg(rLine.argA);
  unsigned long r2 = TxtToReg(rLine.argB);
  unsigned long r3 = TxtToReg(rLine.argC);

  rinsn = (MN_ADD<<I_CMD_O) | (r1<<I_SRCA_O) | (r2<<I_SRCB_O) | (r3<<I_DEST_O);
}

void Assembler::HandleSUB(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long r1 = TxtToReg(rLine.argA);
  unsigned long r2 = TxtToReg(rLine.argB);
  unsigned long r3 = TxtToReg(rLine.argC);

  rinsn = (MN_SUB<<I_CMD_O) | (r1<<I_SRCA_O) | (r2<<I_SRCB_O) | (r3<<I_DEST_O);  
}

void Assembler::HandleADDL(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long r1 = TxtToReg(rLine.argA);
  unsigned long r2 = TxtToReg(rLine.argB);
  unsigned long imm = TxtToImm(rLine.argC);

  rinsn = (MN_ADDL<<I_CMD_O) | (r1<<I_SRCA_O) | (r2<<I_DEST_O) | (0xFFFF&imm);
}

void Assembler::HandleAND(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long srcA = TxtToReg(rLine.argA);
  unsigned long srcB = TxtToReg(rLine.argB);
  unsigned long dest = TxtToReg(rLine.argC);

  rinsn = (MN_AND<<I_CMD_O) | (srcA<<I_SRCA_O) | (srcB<<I_SRCB_O) | (dest<<I_DEST_O);
}

void Assembler::HandleOR(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long srcA = TxtToReg(rLine.argA);
  unsigned long srcB = TxtToReg(rLine.argB);
  unsigned long dest = TxtToReg(rLine.argC);

  rinsn = (MN_OR<<I_CMD_O) | (srcA<<I_SRCA_O) | (srcB<<I_SRCB_O) | (dest<<I_DEST_O);
}

void Assembler::HandleXOR(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long srcA = TxtToReg(rLine.argA);
  unsigned long srcB = TxtToReg(rLine.argB);
  unsigned long dest = TxtToReg(rLine.argC);

  rinsn = (MN_XOR<<I_CMD_O) | (srcA<<I_SRCA_O) | (srcB<<I_SRCB_O) | (dest<<I_DEST_O);
}

void Assembler::HandleNOT(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long srcA = TxtToReg(rLine.argA);
  unsigned long dest = TxtToReg(rLine.argB);

  rinsn = (MN_NOT<<I_CMD_O) | (srcA<<I_SRCA_O) | (dest<<I_DEST_O);
}

void Assembler::HandleBSET(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long srcA = TxtToReg(rLine.argA);
  unsigned long dest = TxtToReg(rLine.argB);
  unsigned long imm = TxtToImm(rLine.argC);
  if( imm > 31 ) throw( AssemError("shift value out of range for BSET") );
  imm &= 0x1F; //just to be safe

  rinsn = (MN_BSET<<I_CMD_O) | (srcA<<I_SRCA_O) | (dest<<I_DEST_O) | (imm&0xFFFF);
}

void Assembler::HandleBCLR(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long srcA = TxtToReg(rLine.argA);
  unsigned long dest = TxtToReg(rLine.argB);
  unsigned long imm = TxtToImm(rLine.argC);
  if( imm > 31 ) throw( AssemError("shift value out of range for BCLR") );
  imm &= 0x1F; //just to be safe

  rinsn = (MN_BCLR<<I_CMD_O) | (srcA<<I_SRCA_O) | (dest<<I_DEST_O) | (imm&0xFFFF);
}

void Assembler::HandleRSL(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long srcA = TxtToReg(rLine.argA);
  unsigned long dest = TxtToReg(rLine.argB);
  unsigned long imm = TxtToImm(rLine.argC);
  if( imm > 31 ) throw( AssemError("shift value out of range for RSL") );
  imm &= 0x1F; //just to be safe

  rinsn = (MN_RSL<<I_CMD_O) | (srcA<<I_SRCA_O) | (dest<<I_DEST_O) | (imm&0xFFFF);
}

void Assembler::HandleRSR(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long srcA = TxtToReg(rLine.argA);
  unsigned long dest = TxtToReg(rLine.argB);
  unsigned long imm = TxtToImm(rLine.argC);
  if( imm > 31 ) throw( AssemError("shift value out of range for RSR") );
  imm &= 0x1F;

  rinsn = (MN_RSR<<I_CMD_O) | (srcA<<I_SRCA_O) | (dest<<I_DEST_O) | (imm&0xFFFF);
}

void Assembler::HandleMUL(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long srcA = TxtToReg(rLine.argA);
  unsigned long srcB = TxtToReg(rLine.argB);
  unsigned long dest = TxtToReg(rLine.argC);

  rinsn = (MN_MUL<<I_CMD_O) | (srcA<<I_SRCA_O) | (srcB<<I_SRCB_O) | (dest<<I_DEST_O);
}

void Assembler::HandleCMP(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long srcA = TxtToReg(rLine.argA);
  unsigned long srcB = TxtToReg(rLine.argB);
  unsigned long dest = TxtToReg(rLine.argC);
  rinsn = (MN_CMP<<I_CMD_O) | (srcA<<I_SRCA_O) | (srcB<<I_SRCB_O) | (dest<<I_DEST_O);
}

void Assembler::HandleSRI(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long reg = TxtToReg(rLine.argA);
  unsigned long imm = TxtToImm(rLine.argB);

  rinsn = (MN_SRI<<I_CMD_O) | (reg<<I_SRCA_O) | (imm&0xFFFF);  
}

void Assembler::HandleSRR(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long reg = TxtToReg(rLine.argA);
  unsigned long r2 = TxtToReg(rLine.argB);
  unsigned long imm = TxtToImm(rLine.argC);

  cout << reg << " " << r2 << " " << imm << " " << rLine.lineBuffer << endl;

  rinsn = (MN_SRR<<I_CMD_O) | (reg<<I_SRCA_O) | (r2<<I_SRCB_O) | (imm&0xFFFF);
}

void Assembler::HandleLRI(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long reg = TxtToReg(rLine.argA);
  unsigned long imm = TxtToImm(rLine.argB);

  rinsn = (MN_LRI<<I_CMD_O) | (reg<<I_DEST_O) | (imm&0xFFFF);    
}

void Assembler::HandleLRR(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long reg = TxtToReg(rLine.argA);
  unsigned long r2 = TxtToReg(rLine.argB);
  unsigned long imm = TxtToImm(rLine.argC);

  //now register address is encoded as B instead of A to save
  //an address offset adder in the cpu.
  rinsn = (MN_LRR<<I_CMD_O) | (reg<<I_SRCB_O) | (r2<<I_DEST_O) | (imm&0xFFFF);  
}

void Assembler::HandleSEQZ(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long reg = TxtToReg(rLine.argA);
  rinsn = (MN_SEQZ<<I_CMD_O) | (reg<<I_SRCA_O);
}

void Assembler::HandleSNEQZ(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long reg = TxtToReg(rLine.argA);
  rinsn = (MN_SNEQZ<<I_CMD_O) | (reg<<I_SRCA_O);
}

void Assembler::HandleSBSET(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long reg = TxtToReg(rLine.argA);
  unsigned long imm = TxtToImm(rLine.argB);
  if( imm > 31 ) throw( AssemError("bit value out of range for SBSET") );
  imm &= 0x1F;

  rinsn = (MN_SBSET<<I_CMD_O) | (reg<<I_SRCA_O) | (imm&0xFFFF);
}

void Assembler::HandleSBCLR(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long reg = TxtToReg(rLine.argA);
  unsigned long imm = TxtToImm(rLine.argB);
  if( imm > 31 ) throw( AssemError("bit value out of range for SBCLR") );
  imm &= 0x1F;

  rinsn = (MN_SBCLR<<I_CMD_O) | (reg<<I_SRCA_O) | (imm&0xFFFF);
}

void Assembler::HandleJI(const Line & rLine, unsigned long & rinsn) const
{
  unsigned long dest = TxtToImm(rLine.argA);

  rinsn = (MN_JI<<I_CMD_O) | (dest&0xFFFF);
}

void Assembler::HandleJR(const Line &rLine, unsigned long & rinsn) const
{
  unsigned long reg = TxtToReg(rLine.argA);
  rinsn = (MN_JR<<I_CMD_O) | (reg<<I_SRCA_O);
}

void Assembler::HandleHLT(const Line & rLine, unsigned long & rinsn) const
{
  rinsn = MN_HLT<<I_CMD_O;
}

/*
 #if 0
  if( rLine.insn == "RETURN" ) {
    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%F";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "1";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%E";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "1";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%D";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "2";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%C";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "3";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%B";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "4";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%A";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "5";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = "%G";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "6";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "ADDL"; //modify the stack pointer
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "6";
    AssembleInsnLine(fakeInsn, rOut);

    //jump to the return address
    fakeInsn.insn = "JR";
    fakeInsn.argA = "%G";
    fakeInsn.argB = "";
    fakeInsn.argC = "";
    AssembleInsnLine(fakeInsn, rOut);

    return;
  };

  if( rLine.insn == "PUSH" ) {
    fakeInsn.insn = "SRR";
    fakeInsn.argA = rLine.argA;
    fakeInsn.argB = "%H";
    fakeInsn.argC = "0";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "ADDL";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "-1";
    AssembleInsnLine(fakeInsn,rOut);
    return;
  };

  if(rLine.insn == "PULL") {
    //the stack pointer normally points to a blank location. to get
    //to the top value on the stack, we need to decrement the pointer first
    fakeInsn.insn = "ADDL";
    fakeInsn.argA = "%H";
    fakeInsn.argB = "%H";
    fakeInsn.argC = "1";
    AssembleInsnLine(fakeInsn, rOut);

    fakeInsn.insn = "LRR";
    fakeInsn.argA = rLine.argA;
    fakeInsn.argB = "%H";
    fakeInsn.argC = "0";
    AssembleInsnLine(fakeInsn, rOut);
    return;
  };
#else
*/
