/*
  Copyright 2010, 2014
  John Culp, All Rights Reserved
*/

#ifndef OPPS_INCLUDED
#define OPPS_INCLUDED

#include <string>

#include "ExecState.hh"
#include "PT_State.hh"
#include "IExpr.hh"
#include "TVCC_Parser_Objs.hh"
#include "TVCC_Defs.hh"
#include "FullOppList.hh"
#include "IE_Call.hh"

//Every Opp owns the pointers to the objects within them

class Opp
{
public:
  Opp(void);
  virtual ~Opp(void);

  virtual Opp* Clone(void) const = 0;
  virtual void GetGName(std::string &rName) = 0;

  virtual void Exec(ExecState &rState);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);

private:
  const Opp& operator=(const Opp &rhs);
};

// declare a code named variable on the stack
class CodeVarDec : public Opp
{
public:
  CodeVarDec(P_TYPE t_var_type, const std::string &rName);
  CodeVarDec(P_TYPE t_var_type, const std::string &rName, const IE &rValue);
  CodeVarDec(const CodeVarDec &rhs);
  virtual ~CodeVarDec(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);

  void UpdateSrc(IE *pNewSrc);

  P_TYPE var_type;
  std::string name; //name from c code or auto generated name
  IE *pValueSrc; //this can be NULL if the variable is not initialized
  unsigned long address; //not assigned initially

private:
  const CodeVarDec& operator=(const CodeVarDec &rhs);
};

class CodeVarWrite : public Opp
{
public:
  CodeVarWrite(const std::string &rName, const IE &rValue);
  CodeVarWrite(const CodeVarWrite &rhs);
  virtual ~CodeVarWrite(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);

  void UpdateSrc(IE *pNewSrc);

  std::string name;
  bool isGlobal;
  IE *pValueSrc;

private:
  const CodeVarWrite& operator=(const CodeVarWrite &rhs);
};

class CodeVarWriteOffset : public Opp
{
public:
  CodeVarWriteOffset(const std::string &rName, const IE &rValue, const IE &rOffset);
  CodeVarWriteOffset(const CodeVarWriteOffset &rhs);
  virtual ~CodeVarWriteOffset(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);

  void UpdateValueSrc(IE *pNewSrc);
  void UpdateOffsetSrc(IE *pNewSrc);

  std::string name;
  bool isGlobal;
  IE *pValueSrc;
  IE *pOffsetSrc;

private:
  const CodeVarWriteOffset& operator=(const CodeVarWriteOffset &rhs);
};

class CompTempStore : public Opp
{
public:
  CompTempStore(unsigned t_ct_id);
  CompTempStore(unsigned t_ct_id, const IE &rValue); //used in loop constructors
  CompTempStore(const CompTempStore &rhs);
  virtual ~CompTempStore(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void Exec(ExecState &rState);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);

  unsigned ct_id;
  IE *pSrc;

private:
  const CompTempStore& operator=(const CompTempStore &rhs);
};


class OppCall : public Opp //function call
{
public:
  OppCall(const std::string &rFnName);
  OppCall(const std::string &rFnName, ExprList &rEL);
  OppCall(const OppCall &rhs);
  virtual ~OppCall(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void Exec(ExecState &rState);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);

  std::string functionName;
  ExprList exprList;
  //unsigned oppID;

private:
  const OppCall& operator=(const OppCall &rhs);
};

class OppReturn : public Opp
{
public:
  OppReturn(void);
  OppReturn(const IE &rValue);
  OppReturn(const OppReturn &rhs);
  virtual ~OppReturn(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);

  bool RetNull(void);

  unsigned ctID; //I need a temp var
  //will be empty if not returning anything, but can't use
  //that as indicator of ret null as we empty the list
  //when we convert the program to a straight line.
  FullOppList oppList;

protected:
  bool retNull;
private:
  const OppReturn& operator=(const OppReturn &rhs);
};

class OppRegLoadFromVar : public Opp
{
public:
  OppRegLoadFromVar(REG_N tReg, const std::string &rVarName);
  OppRegLoadFromVar(const OppRegLoadFromVar &rhs);
  virtual ~OppRegLoadFromVar(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void Exec(ExecState &rState);

  REG_N reg;
  std::string varName;

private:
  const OppRegLoadFromVar& operator=(const OppRegLoadFromVar &rhs);
};

class SkipIfZero : public Opp
{
public:
  SkipIfZero(REG_N tReg, bool invert_logic = false);
  virtual ~SkipIfZero(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void Exec(ExecState &rState);

  REG_N reg;
  bool inv_logic;
private:
  const OppRegLoadFromVar& operator=(const OppRegLoadFromVar &rhs);
};

class Jump : public Opp
{
public:
  Jump(const std::string &rTarget);
  virtual ~Jump(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void Exec(ExecState &rState);

  std::string target;

private:
  const Jump& operator=(const Jump &rhs);
};

//does not do anything, remains as a marker as Opps are inserted into instruction stream
class Target : public Opp
{
public:
  Target(const std::string &rName);
  virtual ~Target(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  //virtual void Exec(ExecState &rState);

  std::string name;

private:
  const Target& operator=(const Target &rhs);
};

class OppIf : public Opp
{
public:
  OppIf(const IE &rIE, const FullOppList &rFOL);
  OppIf(const IE &rIE, const FullOppList &rFOL, const FullOppList &rFOL_Else);
  OppIf(const OppIf &rhs);
  virtual ~OppIf(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void Exec(ExecState &rState);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);
  
  //  IE *pTest;
  FullOppList testList;
  FullOppList oppList;
  FullOppList elseList;

  unsigned oppID;

private:
  const OppIf& operator=(const OppIf &rhs);
};

class OppWhile : public Opp
{
public:
  OppWhile(const IE &rOpp, const FullOppList &rFOL);
  OppWhile(const OppWhile& rhs);
  virtual ~OppWhile(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void Exec(ExecState &rState);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);

  FullOppList testList;
  FullOppList oppList;

  unsigned oppID;

private:
  const OppWhile& operator=(const OppWhile &rhs);
};

class OppFor : public Opp
{
public:
  OppFor(const Opp &rOppL, const IE &rTest, const Opp &rOppR, const FullOppList &rFOL);
  OppFor(const OppFor& rhs);
  virtual ~OppFor(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void Exec(ExecState &rState);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);

  FullOppList initList;
  FullOppList testList;
  FullOppList incList;
 
  FullOppList oppList;

  unsigned oppID;

private:
  const OppFor& operator=(const OppFor &rhs);
};

/*
class OppStore : public Opp
{
public:
  OppStore(unsigned long tAddress, const IE &rOpp);
  OppStore(const OppStore &rhs);
  virtual ~OppStore(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void Exec(ExecState &rState);

  virtual void PrintTree(unsigned myID, PT_State &rPTS);

  unsigned long address;
  IE *pOpp;

private:
  const OppStore& operator=(const OppStore &rhs);
};

class OppRegStore : public Opp
{
public:
  OppRegStore(REG_N tReg, const IE &rOpp);
  OppRegStore(const OppRegStore &rhs);
  virtual ~OppRegStore(void);

  virtual Opp* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual void Exec(ExecState &rState);
  virtual void AssignUniqueIDs(ID_Gen &rGen);

  REG_N reg;
  IE *pOpp;

private:
  const OppRegStore& operator=(const OppRegStore &rhs);
};
*/


#endif //OPPS_INCLUDED
