/*
  Copyright 2010,2014
  John Culp, All Rights Reserved
*/

#include <iomanip>
#include <fstream>
#include <sstream>

#include <libGx/GxMainInterface.hh>
#include <libGx/GxFileSelector.hh>
#include <libGx/GxGeomControl.hh>

#include "SimGui.hh"

using namespace std;

void FormatValue(CPU::REG_TYPE value, DATA_FMT format, std::string &rStr)
{
  ostringstream out;
  long signed_value;
  switch(format)
    {
    case DATA_UINT:
      out << value;
      break;
    case DATA_SIGNED:
      signed_value = (CPU::SIGN_REG_TYPE)value;
      out << signed_value;
      break;
    case DATA_FP:
      out << *((float*)&value);
      break;
    default: //case DATA_HEX:
      out << hex << setfill('0') << setw(8) << value;
      break;
    };
  rStr = out.str();
}


RegRow::RegRow(GxRealOwner *pOwner, const char *pName) :
  GxRow(pOwner), regLabel(this, pName), valueWin(this), format(this),
  fmtHex("Hex"), fmtUInt("UInt"), fmtSigned("Signed"), fmtFp("FP")
{
  SetAdditionalGap(1);
  format.AddItem( &fmtHex );
  format.AddItem( &fmtUInt );
  format.AddItem( &fmtSigned );
  format.AddItem( &fmtFp );

  fmtHex.cb.Assign( CbVoidMemberObj<RegRow, DATA_FMT>(this, &RegRow::SetFormat, DATA_HEX) );
  fmtUInt.cb.Assign( CbVoidMemberObj<RegRow, DATA_FMT>(this, &RegRow::SetFormat, DATA_UINT) );
  fmtSigned.cb.Assign( CbVoidMemberObj<RegRow, DATA_FMT>(this, &RegRow::SetFormat, DATA_SIGNED) );
  fmtFp.cb.Assign( CbVoidMemberObj<RegRow, DATA_FMT>(this, &RegRow::SetFormat, DATA_FP) );

  format.SetCurrentItem(1);
  fmtHex.cb();
}

RegRow::~RegRow(void)
{}

void RegRow::SetValue(unsigned long newValue)
{
  value = newValue;
  Fill();
}

void RegRow::SetFormat(DATA_FMT newFmt)
{
  fmt = newFmt;
  Fill();
}

void RegRow::Fill(void)
{
  string out;
  FormatValue(value, fmt, out);
  valueWin.SetText( out.c_str() );
}

// *********************

WatchRow::WatchRow(GxRealOwner *pOwner) :
  GxRow(pOwner), address(this), name(this), valueWin(this), format(this),
  fmtHex("Hex"), fmtUInt("UInt"), fmtSigned("Signed"), fmtFp("FP")
{
  address.SetNumVisibleChars(10);
  valueWin.SetNumVisibleChars(15);

  format.AddItem( &fmtHex );
  format.AddItem( &fmtUInt );
  format.AddItem( &fmtSigned );
  format.AddItem( &fmtFp );

  fmtHex.cb.Assign( CbVoidMemberObj<WatchRow, DATA_FMT>(this, &WatchRow::SetFormat, DATA_HEX) );
  fmtUInt.cb.Assign( CbVoidMemberObj<WatchRow, DATA_FMT>(this, &WatchRow::SetFormat,DATA_UINT) );
  fmtSigned.cb.Assign( CbVoidMemberObj<WatchRow, DATA_FMT>(this, &WatchRow::SetFormat, DATA_SIGNED) );
  fmtFp.cb.Assign( CbVoidMemberObj<WatchRow, DATA_FMT>(this, &WatchRow::SetFormat, DATA_FP) );

  SetAdditionalGap(2);

  format.SetCurrentItem(1);
  fmtHex.cb();
}

WatchRow::~WatchRow(void){}

unsigned WatchRow::ParseAddr(void)
{
  const char *pStart = address.GetText();
  char *pTail = 0;
  return strtoul(pStart, &pTail, 10);
}

void WatchRow::SetValue(unsigned tValue)
{
  value = tValue;
  Fill();
}

void WatchRow::SetFormat(DATA_FMT newFmt)
{
  fmt = newFmt;
  Fill();
}

void WatchRow::Fill(void)
{
  string out;
  FormatValue(value, fmt, out);
  valueWin.SetText( out.c_str() );
}

// *********************

ExecUntilDialog::ExecUntilDialog(GxTopLevelWin *pTLWin) :
  GxPopupWin(pTLWin), mainColumn(this),
  addLabel(&mainColumn, "address"), address(&mainColumn),
  buttonRow(&mainColumn), ok(&buttonRow, "Ok"), cancel(&buttonRow, "Cancel")
{
  mainColumn.SetAdditionalGap(true);
  mainColumn.SetGeomControl( GxSIBasic(GX_WD_INT, GX_HT_INT, GX_FLOW_LEFT, GX_FLOW_UP, 1,1,1,1) );

  ok.cb.Assign( CbVoidMemberObj<ExecUntilDialog, GX_STATUS>(this, &ExecUntilDialog::DoExit, GX_OK) );
  cancel.cb.Assign( CbVoidMemberObj<ExecUntilDialog, GX_STATUS>(this, &ExecUntilDialog::DoExit, GX_CANCELED) );

  buttonRow.SetWidthIdentical(true);
  buttonRow.SetHeightIdentical(true);
  buttonRow.SetAdditionalGap(true);
			     
  Place();
  Create();
}

ExecUntilDialog::~ExecUntilDialog(void)
{}

GX_STATUS ExecUntilDialog::DoDialog(void)
{
  Display();
  EventLoop();
  XUnmapWindow(dInfo.display, xWin);
  return eStat;
}
 
unsigned ExecUntilDialog::GetAddress(void)
{
  const char *pStart = address.GetText();
  char *pTail = 0;
  unsigned res = strtoul(pStart, &pTail, 10);
  return res;
}

void ExecUntilDialog::DoExit(GX_STATUS rStat)
{
  eStat = rStat;
  processEvents = false;
}

// *********************
DispWin::DispWin(GxRealOwner *pOwner) :
  GxCoreWin(pOwner)
{
  width = 640;
  height = 480;

  img.Initialize(640, 480);
}

DispWin::~DispWin(void)
{}

void DispWin::Draw(void)
{
  for(unsigned ii = 0; ii < 480; ii++)
    for(unsigned jj = 0; jj < 640; jj++)
      {
	XSetForeground(dInfo.display, vData.borderGC, dwgData.GetXPixel( img.Pixel(jj,ii) ));
	XDrawPoint(dInfo.display, xWin, vData.borderGC, jj, ii);
      };
}

void DispWin::HandleEvent(const XEvent &rEvent)
{
  /* becase we have a backing pixmap, I don't have to redraw on expose */
  if(rEvent.type == Expose )
    {
      if( rEvent.xexpose.count != 0 ) return;

      //draw it.
      Draw();
    };
}

void DispWin::CreateXWindow(void)
{
  GxCoreWin::CreateXWindow();

  dwgData.Setup( dInfo.cVisualInfo );
}

void DispWin::GetWindowData(XSetWindowAttributes &winAttributes,
			    ULINT &valueMask)
{
  winAttributes.event_mask = ExposureMask;

  valueMask |= CWEventMask;
}

// *********************

SimGui::SimGui(GxDisplay *pOwner, Sim &rTSim) :
  GxMainWin(pOwner), rSim(rTSim), mainRow(this), bCol(&mainRow),
  loadFile(&bCol, "Load"), exec(&bCol, "Exec"), execUntil(&bCol, "Exec Until..."), stop(&bCol, "Stop"),
  singleStep(&bCol, "Step"), saveCfg(&bCol, "Save Cfg"),
  loadCfg(&bCol, "Load Cfg"), blankMem(&bCol, "Blank Mem"), loadMem(&bCol, "Load Mem"), blankFB(&bCol, "Blank FB"),
  updateGui(&bCol, "Update Gui"), quit(&bCol, "Quit"),
  regCol(&mainRow),
  insRow(&regCol), pcRow(&insRow, "PC"), disassembly(&insRow),
  regTable(&regCol),
  aRow(&regTable, " A"), bRow(&regTable, " B"), cRow(&regTable, " C"), dRow(&regTable, " D"),
  eRow(&regTable, " E"), fRow(&regTable, " F"), gRow(&regTable, " G"), hRow(&regTable, " H"),
  wrLabel(&regCol, "address/name/value/format"),
  wr0(&regCol), wr1(&regCol), wr2(&regCol), wr3(&regCol), wr4(&regCol), wr5(&regCol),
  wr6(&regCol), wr7(&regCol), wr8(&regCol), wr9(&regCol), 
  dwin(&mainRow)
{  
  mainRow.SetGeomControl( GxSIBasic(GX_WD_INT, GX_HT_INT, GX_FLOW_LEFT, GX_FLOW_UP, 1,1,1,1) );
  mainRow.SetAdditionalGap(2);

  bCol.SetWidthIdentical(true);
  bCol.SetHeightIdentical(true);
  bCol.SetAdditionalGap(1);

  loadFile.cb.Assign( CbVoidMember<SimGui>(this, &SimGui::DoLoad) );
  exec.cb.Assign( CbVoidMember<SimGui>(this, &SimGui::DoExec) );
  execUntil.cb.Assign( CbVoidMember<SimGui>(this, &SimGui::DoExecUntil) );
  stop.cb.Assign( CbVoidMember<SimGui>(this, &SimGui::DoStop) );
  singleStep.cb.Assign( CbVoidMember<SimGui>(this, &SimGui::DoStep) );
  saveCfg.cb.Assign( CbVoidMember<SimGui>(this, &SimGui::DoSaveCfg) );
  loadCfg.cb.Assign( CbVoidMember<SimGui>(this, &SimGui::DoLoadCfg) );
  blankMem.cb.Assign( CbVoidMember<SimGui>(this, &SimGui::DoBlankMem) );
  loadMem.cb.Assign( CbVoidMember<SimGui>(this, &SimGui::DoLoadMem) );
  blankFB.cb.Assign( CbVoidMember<SimGui>(this, &SimGui::DoBlankFB) );
  updateGui.cb.Assign(  CbVoidMember<SimGui>(this, &SimGui::DoUpdateGui) );
  quit.cb.Assign(  CbVoidMember<SimGui>(this, &SimGui::DoQuit) );

  regCol.SetAdditionalGap(1);

  insRow.SetAdditionalGap(1);

  regTable.SetNumObjectsPerRow(2);
  regTable.SetAdditionalHGap(1);
  regTable.SetAdditionalVGap(1);

  wrs[0] = &wr0;
  wrs[1] = &wr1;
  wrs[2] = &wr2;
  wrs[3] = &wr3;
  wrs[4] = &wr4;
  wrs[5] = &wr5;
  wrs[6] = &wr6;
  wrs[7] = &wr7;
  wrs[8] = &wr8;
  wrs[9] = &wr9;
}

SimGui::~SimGui(void)
{}

void SimGui::DoLoad(void)
{
  GxFileSelector fileSel(this);
  string fileName;

  if( fileSel.OpenFile(fileName, "*.txt") != GX_OK ) return;

  rSim.LoadProg(fileName);

  rSim.SetupForExec();
  UpdateGui();
}

void SimGui::DoExec(void)
{
  stopExec = false;
  ExecCore(false, 0);
}

void SimGui::DoExecUntil(void)
{
  stopExec = false;
  ExecUntilDialog execUntil(this);
  if( execUntil.DoDialog() == GX_OK )
    {
      unsigned stopAddress = execUntil.GetAddress();
      cout << "SimGui::DoExecUntil got address: " << stopAddress << endl;
      ExecCore(true, stopAddress);
    };
}

void SimGui::DoStop(void)
{
  stopExec = true;
}

void SimGui::DoStep(void)
{
  cout << "before: " << endl;
  rSim.PrintState();
  rSim.SingleStep();
  cout << "after: " << endl;
  rSim.PrintState();
  UpdateGui();
}

void SimGui::DoSaveCfg(void)
{
  GxFileSelector fileSel(this);
  string fileName;

  if( fileSel.SaveFile(fileName, "*.cfg") != GX_OK ) return;
  
  ofstream fout( fileName.c_str() );
  if( !fout ) return;

  for(unsigned ii = 0; ii < 10; ii++)
    {
      if( strlen( wrs[ii]->address.GetText() ) ) //there actually is an address value
	fout << ii << " " << (wrs[ii])->ParseAddr() << " " << (wrs[ii])->name.GetText() << endl;
    };

  fout.close();
}

void SimGui::DoLoadCfg(void)
{
  GxFileSelector fileSel(this);
  string fileName;

  if( fileSel.OpenFile(fileName, "*.cfg") != GX_OK ) return;

  ifstream fin( fileName.c_str() );
  if(!fin) return;

  unsigned lineNum = 1;
  while(fin)
    {
      char lineBuffer[512];
      fin.getline(lineBuffer, 511);
      lineBuffer[511] = '\0';

      if(!fin) return;

      char *pStart = lineBuffer;
      char *pTail = 0;

      unsigned num = strtoul(pStart, &pTail, 10);
      if( pStart == pTail )
	{
	  cerr << "error parsing num on line: " << lineNum << endl;
	  continue;
	};

      pStart = pTail;
      unsigned lineAddr = strtoul(pStart, &pTail, 10);
      if( pTail == pStart )
	{
	  cerr << "error parsing address on line: " << lineNum << endl;
	  continue;
	};

      pStart = pTail;
      while( isspace(*pStart) && *pStart != '\0' )
	pStart++;

      pTail = pStart;
      while( !isspace(*pTail) && *pTail != '\0' )
	pTail++;

      string name(pStart, pTail-pStart);

      if(num < 10)
	{
	  ostringstream out;
	  out << lineAddr;
	  (wrs[num])->address.SetText( out.str().c_str() );
	  (wrs[num])->name.SetText( name.c_str() );
	};

      lineNum++;
    };

}

void SimGui::DoBlankMem(void)
{
  rSim.BlankData();
  UpdateGui();
}

void SimGui::DoLoadMem(void)
{
  GxFileSelector fileSel(this);
  string fileName;

  if( fileSel.OpenFile(fileName, "*.txt") != GX_OK ) return;

  rSim.LoadData(fileName);
  UpdateGui();
}

void SimGui::DoBlankFB(void)
{
  rSim.BlankFB();
  //hackish.
  rSim.dmaToFBOccurred = true;
  UpdateGui();
}

void SimGui::DoUpdateGui(void)
{
  rSim.dmaToFBOccurred = true;
  UpdateGui();
}

void SimGui::DoQuit(void)
{
  dInfo.rMainInterface.EndEventLoop();
}

void SimGui::ExecCore(bool stop_at_addr, unsigned pc_address)
{
  while(true)
    {
      if(stop_at_addr && pc_address == rSim.GetPC() ) break;
      if( stopExec ) break;
      bool res = rSim.SingleStep();
      UpdateGui();
      dInfo.rMainInterface.EventStep(false);
      if( !res ) break;
    };
}

void SimGui::UpdateGui(void)
{
  unsigned pc, a, b, c, d, e, f, g, h;
  rSim.GetState(pc, a, b, c, d, e, f, g, h);

  pcRow.SetValue(pc);
  aRow.SetValue(a);
  bRow.SetValue(b);
  cRow.SetValue(c);
  dRow.SetValue(d);
  eRow.SetValue(e);
  fRow.SetValue(f);
  gRow.SetValue(g);
  hRow.SetValue(h);

  //unsigned long GetData(unsigned dword);
  //unsigned long GetProg(unsigned dword);
  string diss;
  rSim.Disassemble(diss);

  disassembly.SetText( diss.c_str() );

  for(unsigned ii = 0; ii < 10; ii++)
    {
      unsigned address = (wrs[ii])->ParseAddr();
      (wrs[ii])->SetValue( rSim.GetData(address) );
    };

  if( rSim.dmaToFBOccurred ) //too expensive to do blindly every Update
    {
      for(unsigned ii = 0; ii < 480; ii++)
	for(unsigned jj = 0; jj < 640; jj++)
	  dwin.img.Pixel(jj,ii) = U32RGBAPixel( rSim.fb[ii*640 + jj] );

      dwin.Draw();
      rSim.dmaToFBOccurred = false;
    };
}
