/*
  Copyright 2009,2014
  John Culp, All Rights Reserved
*/

#include <sstream>
#include <fstream>
#include <assert.h>
#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <stdexcept>
#include <string.h>

#include "cpu.hh"

using namespace std;

const unsigned RAM_SIZE = 1024;
const unsigned ROM_SIZE = 8192;

// class Blank
// {
// public:
//   void operator()(unsigned long rValue)
//   {
//     rValue = 0;
//   };
// };

void Blank(CPU::REG_TYPE &rValue)
{
  rValue = 0;
}

CPU::CPU(void) :
  regA(0), regB(0), regC(0), regD(0), regE(0), regF(0), regG(0), regH(0), PC(0)
{
  verbose = 0;

  assert(sizeof(REG_TYPE) == 4);
  assert(sizeof(SIGN_REG_TYPE) == 4);

  data_ram.resize(RAM_SIZE);
  program_rom.resize(ROM_SIZE);

  for_each(data_ram.begin(), data_ram.end(), Blank);
  for_each(program_rom.begin(), program_rom.end(), Blank);
}

CPU::~CPU(void)
{}

void CPU::LoadProg(const string &rFileName)
{
  ifstream fin( rFileName.c_str() );
  assert( fin );

  char lineBuffer[512];

  unsigned cAddr = 0;
  while(1)
    {
      fin.getline(lineBuffer, 511);
      if(!fin) break;
      lineBuffer[511] = '\0';

      char *pTail;
      unsigned long insn_raw = strtoul(lineBuffer, &pTail, 16);
      REG_TYPE insn = insn_raw;
      assert( insn == insn_raw);

      if( pTail == lineBuffer )
	{
	  cerr << "error parsing line: " << cAddr << " terminating read" << endl;
	  break;
	};
      
      program_rom[cAddr] = insn;
      cAddr++;
    };

  program_size = cAddr;

  if(verbose) //any verbosity level except off.
    cout << "loaded: " << dec << cAddr << " instructions" << endl;;
  fin.close();
}

void CPU::BlankData(void)
{
  for_each(data_ram.begin(), data_ram.end(), Blank);
}

void CPU::LoadData(const std::string &rFileName)
{
  ifstream fin( rFileName.c_str() );
  if(!fin) return;

  unsigned lineNum = 0;
  while(fin)
    {
      lineNum++;
      char lineBuffer[512];
      fin.getline(lineBuffer, 511);
      lineBuffer[511] = '\0';

      if(!fin) return;

      if(lineBuffer[0] == '#') continue;

      char *pStart = lineBuffer;

      //if the line is empty except white spaces, ignore it
      bool nwsFound = false;
      while(*pStart != '\0' )
	{
	  if( !isspace( *pStart ) )
	    {
	      nwsFound = true;
	      break;
	    }
	  pStart++;
	};
      if(!nwsFound) continue;

      pStart = lineBuffer;
      char *pTail = 0;

      unsigned lineAddr = strtoul(pStart, &pTail, 10);
      if( pStart == pTail )
	{
	  cerr << "error parsing address on line: " << lineNum << endl;
	  continue;
	};

      int startValOffset = pTail - pStart;
      while(isspace( lineBuffer[startValOffset]) )
	{
	  startValOffset++;
	  if(startValOffset >= 512) continue; //next line.
	};
      
      int endValOffset = startValOffset;
      while( !isspace( lineBuffer[endValOffset] ) && lineBuffer[endValOffset] != '\0')
	{
	  endValOffset++; //relying on termination.
	};
      
      unsigned value = 0;
      if(lineBuffer[endValOffset-1] == 'f')
	{
	  cout << "parsing memory value as float line: " << lineNum << endl;
	  pStart = pTail;
	  float tvalue = strtod(pStart, &pTail);
	  if( pTail == pStart )
	    {
	      cerr << "error parsing float value on line: " << lineNum << endl;
	      continue;
	    };
	  memcpy(&value, &tvalue, 4);
	}else{
	cout << "parsing memory value as uint line: " << lineNum << endl;
	pStart = pTail;
	value = strtoul(pStart, &pTail, 10);
	if( pTail == pStart )
	  {
	    cerr << "error parsing value on line: " << lineNum << endl;
	    continue;
	  };
      };

      DataWrite(lineAddr, value);
    };
}

void CPU::Reset(void)
{
    PC = 0;
}

bool CPU::SingleStep(void)
{      

  REG_TYPE inst = 0;
  try{
    inst = program_rom[PC];  //fetch
  }catch(std::out_of_range arg)
    {
      //terminate execution with error.
      cout << "invalid instruction address for fetch" << endl;
      return false;
    } 

  //execute
  REG_TYPE nextPC = PC+1;
  REG_TYPE cmd, srcA, srcB, dest, immediate;
  Dissassemble(inst, cmd, srcA, srcB, dest, immediate);

  static unsigned v_count = 0;
  v_count++;

  if(verbose)
    {
      if( !(v_count%verbose) )
	PrintState();
    };

  //cout << hex << cmd << srcA << " " << srcB << " " << dest << endl;
  /*
    case MN_INC:
      IncReg(dest);
      break;
    case MN_DEC:
      DecReg(dest);
      break;
  */

  switch(cmd)
    {
    case MN_NOP:
      break;
    case MN_LUI:
      I_LUI(srcA, dest, inst&0xFFFF);
      break;
    case MN_LLI:
      I_LLI(srcA, dest, inst&0xFFFF);
      break;
    case MN_ADD:
      I_ADD(srcA, srcB, dest);
      break;
    case MN_SUB:
      I_SUB(srcA, srcB, dest);
      break;
    case MN_ADDL:
      I_ADDL(srcA, dest, immediate);
      break;
    case MN_AND:
      I_AND(srcA, srcB, dest);
      break;
    case MN_OR:
      I_OR(srcA, srcB, dest);
      break;
    case MN_XOR:
      I_XOR(srcA, srcB, dest);
      break;
    case MN_NOT:
      I_NOT(srcA, dest);
      break;
    case MN_BSET:
      I_BSET(srcA, dest, immediate);
      break;
    case MN_BCLR:
      I_BCLR(srcA, dest, immediate);
      break;
    case MN_RSL:
      I_RSL(srcA, dest, immediate);
      break;
    case MN_RSR:
      I_RSR(srcA, dest, immediate);
      break;
    case MN_MUL:
      I_MUL(srcA, srcB, dest);
      break;
    case MN_CMP:
      I_CMP(srcA, srcB, dest);
      break;
    case MN_SRI:
      I_SRI(srcA, immediate);
      break;
    case MN_SRR:
      I_SRR(srcA, srcB, immediate);
      break;
    case MN_LRI:
      I_LRI(dest, immediate);
      break;
    case MN_LRR:
      I_LRR(srcB, dest, immediate);
      break;
    case MN_SEQZ:
      if( GetRegValue(srcA) == 0 ) nextPC++; //skip next instruction.
      break;
    case MN_SNEQZ:
      if( GetRegValue(srcA) != 0 ) nextPC++;
      break;
    case MN_SBSET:
      if( ( GetRegValue(srcA) & GetImmBitMask(immediate) ) != 0 ) nextPC++;
      break;
    case MN_SBCLR:
      if( ( GetRegValue(srcA) & GetImmBitMask(immediate) ) == 0 ) nextPC++;
      break;
    case MN_JI: //branch always
      nextPC = inst&0xFFFF;
      break;
    case MN_JR: //branch always
      nextPC = GetRegValue(srcA)&0xFFFF;
      break;
    case MN_HLT: //terminate
      nextPC = PC; //don't advance program counter
      return false;
    default:
      cerr << "invalid instruction. terminate" << endl;
      return false;
      //invalid instruction
    };

  //  PrintState();

  PC = nextPC;
  return true;
}

void CPU::Execute(void)
{
  execCount = 0;

  while( SingleStep() )
    execCount++;
}

unsigned long CPU::ExecCount(void)
{
  return execCount;
}

void CPU::PrintState(void)
{
  cout << "PC: " << hex << setw(8) << setfill('0') << PC
       << " A:"  << setw(8) << regA
       << " B:"  << setw(8) << regB
       << " C:"  << setw(8) << regC
       << " D:"  << setw(8) << regD
       << " E:"  << setw(8) << regE
       << " F:"  << setw(8) << regF
       << " G:"  << setw(8) << regG
       << " H:"  << setw(8) << regH << endl;
}

void CPU::Disassemble(std::string &rStr)
{
  ostringstream out;

  REG_TYPE inst = 0;
  try{
    inst = program_rom[PC];  //fetch
  }catch(std::out_of_range arg)
    {
      //terminate execution with error.
      cout << "ERROR: Invalid instruction address for Disassemble" << endl;
    }

  REG_TYPE cmd, srcA, srcB, dest, immediate;
  Dissassemble(inst, cmd, srcA, srcB, dest, immediate);  

  switch(cmd)
    {
    case MN_NOP:
      out << "NOP";
      break;
    case MN_LUI:
      out << "LUI";
      break;
    case MN_LLI:
      out << "LLI";
      break;
    case MN_ADD:
      out << "ADD";
      break;
    case MN_SUB:
      out << "SUB";
      break;
    case MN_ADDL:
      out << "ADDL";
      break;
    case MN_AND:
      out << "AND";
      break;
    case MN_OR:
      out << "OR";
      break;
    case MN_XOR:
      out << "XOR";
      break;
    case MN_NOT:
      out << "NOT";
      break;
    case MN_BSET:
      out << "BSET";
      break;
    case MN_BCLR:
      out << "BCLR";
      break;
    case MN_RSL:
      out << "RSL";
      break;
    case MN_RSR:
      out << "RSR";
      break;
    case MN_MUL:
      out << "MUL";
      break;
    case MN_CMP:
      out << "CMP";
      break;
    case MN_SRI:
      out << "SRI";
      break;
    case MN_SRR:
      out << "SRR";
      break;
    case MN_LRI:
      out << "LRI";
      break;
    case MN_LRR:
      out << "LRR";
      break;
    case MN_SEQZ:
      out << "SEQZ";
      break;
    case MN_SNEQZ:
      out << "SNEQZ";
      break;
    case MN_SBSET:
      out << "SBSET";
      break;
    case MN_SBCLR:
      out << "SBCLR";
      break;
    case MN_JI:
      out << "JI";
      break;
    case MN_JR:
      out << "JR";
      break;
    case MN_HLT:
      out << "HLT";
      break;
    default:
      out << "<ERROR>";
      break;
    };

  out << ' ' << RegToStr(srcA) << ' ' << RegToStr(srcB) << ' ' << RegToStr(dest) << ' ' << hex << setw(8) << setfill('0') << immediate;

  rStr = out.str();
}

void CPU::Dissassemble(REG_TYPE inst, REG_TYPE &cmd, REG_TYPE &srcA, REG_TYPE &srcB, REG_TYPE &dest,
		       REG_TYPE &imm) const
{
  
  cmd  = (inst & I_CMD_MASK)>>I_CMD_O;
  srcA = (inst & I_SRCA_MASK)>>I_SRCA_O;
  srcB = (inst & I_SRCB_MASK)>>I_SRCB_O;
  dest = (inst & I_DEST_MASK)>>I_DEST_O;
  imm = (inst&0xFFFF);
}

const char* CPU::RegToStr(unsigned reg)
{
  switch(reg)
    {
    case 0:
      return "A";
     case 1:
      return "B";
    case 2:
      return "C";
    case 3:
      return "D";
    case 4:
      return "E";
    case 5:
      return "F";
    case 6:
      return "G";
    case 7:
      return "H";
    default:
      return "ERROR";
    };
}

CPU::REG_TYPE CPU::DataRead(REG_TYPE dwordAddr)
{
  CPU::REG_TYPE value;

  if(dwordAddr >= data_ram.size() )
    value = InvalidDataRead(dwordAddr);
  else
    {
      assert( dwordAddr < data_ram.size() );
      value = data_ram[dwordAddr];
    };

  //cout << "RL " << dwordAddr << " " << value << endl;
  return value;
}

void CPU::DataWrite(CPU::REG_TYPE dwordAddr, REG_TYPE value)
{
  //cout << "RS " << dwordAddr << " " << value << endl;

  if(dwordAddr >= data_ram.size() )
    {
      InvalidDataWrite(dwordAddr, value);
      return;
    };

  assert( dwordAddr < data_ram.size() );
  data_ram[dwordAddr] = value;
}


CPU::REG_TYPE CPU::InvalidDataRead(REG_TYPE dwordAddr)
{
  cerr << "CPU::InvalidDataRead ERROR READING FROM INVALID dwordAddr: " << dwordAddr << endl;
  PrintState();
  return 0xFFFFFFFF;
}

void CPU::InvalidDataWrite(REG_TYPE dwordAddr, REG_TYPE value)
{
  cerr << "CPU::InvalidDataWrite ERROR IGNORED WRITING TO INVALID dwordAddr: " << dwordAddr << " of value: " << value << endl;
  PrintState();
}

CPU::REG_TYPE CPU::GetRegValue(unsigned reg) const
{
  switch(reg)
    {
    case 0:
      return regA;
    case 1:
      return regB;
    case 2:
      return regC;
    case 3:
      return regD;
    case 4:
      return regE;
    case 5:
      return regF;
    case 6:
      return regG;
    case 7:
      return regH;
    default:
      assert(false);
      return 0;
    }
}

void CPU::SetRegValue(unsigned reg, REG_TYPE value)
{
    switch(reg)
    {
    case 0:
      regA = value;
      break;
    case 1:
      regB = value;
      break;
    case 2:
      regC = value;
      break;
    case 3:
      regD = value;
      break;
    case 4:
      regE = value;
      break;
    case 5:
      regF = value;
      break;
    case 6:
      regG = value;
      break;
    case 7:
      regH = value;
      break;
    default:
      assert(false);
    }
}

long CPU::SignExtendImm(REG_TYPE imm) const
{
  long res = imm;
  if( imm&1<<15 ) //negative, so sign extend
    res = 0xFFFF0000 | (imm&0x0000FFFF);

  return res;
}

CPU::REG_TYPE CPU::GetImmBitMask(REG_TYPE imm) const
{
  imm &= 0x01F;
  REG_TYPE mask = 0x1<<imm;
  return mask;
}

void CPU::I_LUI(unsigned srcReg, unsigned destReg, REG_TYPE data)
{
  if(srcReg != destReg)
    cout << "CPU::I_LUI assembled code bug! srcReg != destReg" << endl;

  REG_TYPE value = GetRegValue(destReg);

  value &= 0x0000FFFF;
  value |= data<<16;

  SetRegValue(destReg, value);
}

void CPU::I_LLI(unsigned srcReg, unsigned destReg, REG_TYPE data)
{
  if(srcReg != destReg)
    cout << "CPU::I_LLI assembled code bug! srcReg != destReg" << endl;

  REG_TYPE value = GetRegValue(destReg);

  value &= 0xFFFF0000;
  value |= data&0x0000FFFF;

  SetRegValue(destReg, value);
}

void CPU::I_ADD(unsigned srcA, unsigned srcB, unsigned dest)
{
  REG_TYPE valA = GetRegValue(srcA);
  REG_TYPE valB = GetRegValue(srcB);

  REG_TYPE valDest = valA + valB;

  SetRegValue(dest, valDest);
}

void CPU::I_SUB(unsigned srcA, unsigned srcB, unsigned dest)
{
  REG_TYPE valA = GetRegValue(srcA);
  REG_TYPE valB = GetRegValue(srcB);

  REG_TYPE valDest = valA - valB; /* (REG_TYPE)( (SIGN_REG_TYPE)valA - (SIGN_REG_TYPE)valB ); */

  SetRegValue(dest, valDest);
}

void CPU::I_ADDL(unsigned srcA, unsigned dest, unsigned imm)
{
  REG_TYPE valA = GetRegValue(srcA);

  long val_dest = valA + SignExtendImm(imm);

  SetRegValue(dest, val_dest);
}

void CPU::I_AND(unsigned srcA, unsigned srcB, unsigned dest)
{
  REG_TYPE valA = GetRegValue(srcA);
  REG_TYPE valB = GetRegValue(srcB);

  REG_TYPE valDest = valA & valB;

  SetRegValue(dest, valDest);
}

void CPU::I_OR(unsigned srcA, unsigned srcB, unsigned dest)
{
  REG_TYPE valA = GetRegValue(srcA);
  REG_TYPE valB = GetRegValue(srcB);

  REG_TYPE valDest = valA | valB;

  SetRegValue(dest, valDest);
}

void CPU::I_XOR(unsigned srcA, unsigned srcB, unsigned dest)
{
  REG_TYPE valA = GetRegValue(srcA);
  REG_TYPE valB = GetRegValue(srcB);

  REG_TYPE valDest = valA ^ valB;

  SetRegValue(dest, valDest);
}

void CPU::I_NOT(unsigned srcA, unsigned dest)
{
  REG_TYPE valA = GetRegValue(srcA);

  REG_TYPE valDest = ~valA;

  SetRegValue(dest, valDest);
}

void CPU::I_BSET(unsigned srcA, unsigned dest, unsigned imm)
{
  REG_TYPE val = GetRegValue(srcA);

  val |= GetImmBitMask(imm);

  SetRegValue(dest, val);
}

void CPU::I_BCLR(unsigned srcA, unsigned dest, unsigned imm)
{
  REG_TYPE val = GetRegValue(srcA);

  REG_TYPE mask = GetImmBitMask(imm);
  val &= ~mask;

  SetRegValue(dest, val);
}

void CPU::I_RSL(unsigned src, unsigned dest, unsigned imm)
{
  REG_TYPE val = GetRegValue(src);

  REG_TYPE valDest = val<<(imm & 0x1F);

  SetRegValue(dest, valDest);
}

void CPU::I_RSR(unsigned src, unsigned dest, unsigned imm)
{
  REG_TYPE val = GetRegValue(src);

  REG_TYPE valDest = val>>(imm & 0x1F);

  SetRegValue(dest, valDest);
}

void CPU::I_MUL(unsigned srcA, unsigned srcB, unsigned dest)
{
  REG_TYPE valA = GetRegValue(srcA);
  //valA &= 0xFFFF; // can't do this; unsigned values y'all

  REG_TYPE valB = GetRegValue(srcB);
  // valB &= 0xFFFF; // can't do this; unsigned values y'all

  REG_TYPE valDest = valA*valB;
  
  SetRegValue(dest, valDest);
}

void CPU::I_CMP(unsigned srcA, unsigned srcB, unsigned dest)
{
  REG_TYPE valA = GetRegValue(srcA);
  REG_TYPE valB = GetRegValue(srcB);

  REG_TYPE valDest = 0;

  if( valA == valB ) valDest |= 0x1;

  if( valA > valB ) valDest |= 0x2;

  if( (SIGN_REG_TYPE)valA > (SIGN_REG_TYPE)valB ) valDest |= 4;

  if( valA < valB ) valDest |= 0x8;

  if( (SIGN_REG_TYPE)valA < (SIGN_REG_TYPE)valB ) valDest |= 16;

  SetRegValue(dest, valDest);
}

void CPU::I_SRI(unsigned srcA, REG_TYPE imm)
{
  REG_TYPE valA = GetRegValue(srcA);
  REG_TYPE address = imm&0x0000FFFF;

  DataWrite(address, valA);
}

void CPU::I_SRR(unsigned srcA, unsigned srcB, REG_TYPE imm)
{
  REG_TYPE valA = GetRegValue(srcA);
  REG_TYPE address = ( GetRegValue(srcB) + SignExtendImm(imm) ) & 0x0000FFFF;

  DataWrite(address, valA);
}

void CPU::I_LRI(unsigned dest, REG_TYPE imm)
{
  REG_TYPE address = imm&0x0000FFFF;

  REG_TYPE value = DataRead( address );

  SetRegValue(dest, value);
}

void CPU::I_LRR(unsigned srcB, unsigned dest, REG_TYPE imm)
{
  REG_TYPE address = ( GetRegValue(srcB) + SignExtendImm(imm) ) & 0x0000FFFF;

  REG_TYPE value = DataRead( address );

  //cout << "cpu::I_LRR: " << RegToStr(srcA) << " " << RegToStr(dest) << " address: " << address << " value: " << value << endl;

  SetRegValue(dest, value);
}
