/*
  file created 2014
  based on files from 2006,2009
  modified 2014
  released to the Public Domain by the author
*/

#include <math.h>

#include <algorithm>
#include <ext/functional>
#include <assert.h>
#include <iomanip>
#include <string.h>
#include <arpa/inet.h>
#include <unistd.h>

#include <iostream>
#include <fstream>

#include "Driver_HL.hh"

//the location of the depth buffer in the overall framebuffer
// ********** THESE ARE DWORD OFFSETS FOR THE PROCESSOR ************
//************ multiply x4 to get bytes *******************
const unsigned FB_OFFSET = 0;
const unsigned DB_OFFSET = 1000000;

const unsigned TEX_START = 200000;
const unsigned TEX_DIM   = 256;
const unsigned TEX_SIZE  = TEX_DIM*TEX_DIM;

using namespace std;

unsigned F2R(float in)
{
  unsigned res;
  memcpy(&res, &in, 4);
  return res;
}

// *****************

Driver_HL::Driver_HL(CB_Base &rTCB) :
  forSim(false), currentTextureID(0), rCB(rTCB)
{}

Driver_HL::~Driver_HL(void)
{}

void Driver_HL::PrintStatus(void)
{
  unsigned status = rCB.CB_Read(0xFF);
  cout << "status: " << status << endl;

  if( status & 0x08 )
     cout << "poly scanline idle" << endl;
  else
    cout << "poly scanline busy" << endl;

  if( status & 0x01 )
     cout << "block reader idle" << endl;
  else
    cout << "block reader busy" << endl;

  if( status & 0x02 )
     cout << "block writer idle" << endl;
  else
    cout << "block writer busy" << endl;

  if( status & 0x04 )
     cout << "block set idle" << endl;
  else
    cout << "block set busy" << endl;
}

/*
  this was a quick (very old) test that I though might be useful (with changes) to test the 
  framebuffer.  I just stuck the code here where I might find it later.
*/
void Driver_HL::TestRam(void)
{
  const unsigned write_count = 255;
  const unsigned read_count = 255;

  unsigned bad_count = 0;
  unsigned presumed_bad_writes = 0;

  for( unsigned ii = 0; ii < write_count; ii++)
    {
      unsigned dataValue = (unsigned)rint( (((double)random())/((double)RAND_MAX))* 0xFFFFFFFF);
      cout << "storing value: " << hex << dataValue << dec << endl;
      
      Write_FB(0x00, dataValue);

      unsigned bad_read_count = 0;
      for(unsigned jj = 0; jj < read_count; jj++)
	{
	  
	  unsigned result = 0;
	  Read_FB(0, result);
	  //driver.DumpRegisters();
	  if( result != dataValue )
	    {
	      cout << setfill('0') << dec << setw(5) << jj
		   << "\t error stored != read stored: " << setfill('0') << setw(8) << hex << dataValue
		   << " result: " << setfill('0') << setw(8) << hex << result << dec << endl;
	      bad_read_count++;
	    };
	};
      bad_count += bad_read_count;

      if(bad_read_count > read_count/2)
	presumed_bad_writes++;
    };

  cout << "bad count: " << bad_count << endl;
  cout << "error factor: " << setprecision(5) << (double)bad_count/(double)(write_count*read_count) << endl;
  cout << "factor that could be due to bad writes: " << setprecision(5)
       << (double)(presumed_bad_writes*write_count)/(double)(write_count*read_count) << endl;
}

void Driver_HL::ConfigForRTLSim(bool tForSim)
{
  forSim = tForSim;
}

void Driver_HL::LoadProgram(const std::string &rPrgm)
{
  ifstream fin( rPrgm.c_str() );
  assert(fin);

  unsigned cAddr = 0;
  while(true)
    {
      char lineBuffer[512];
      fin.getline(lineBuffer, 511);
      if(!fin) break;
      lineBuffer[511] = '\0';

      char *pTail;
      unsigned insn_raw = strtoul(lineBuffer, &pTail, 16);
      unsigned insn = insn_raw; //64 to 32 bit maybe.
      assert( insn == insn_raw);

      if( pTail == lineBuffer )
	{
	  cerr << "error parsing line: " << cAddr << " terminating read" << endl;
	  assert(false);
	  break;
	};
      
      //write the instruction to address E0
      rCB.CB_Write(0xE0, insn);
      //write the memory address to write to E1
      rCB.CB_Write(0xE1, cAddr);
      //write a junk value to E2 to cause the program memory write
      rCB.CB_Write(0xE2, 0xFFFFFFFF);

      cout << "wrote addr: " << cAddr << endl;
      cAddr++;
    };
  cout << "load program done" << endl;
}

void Driver_HL::LoadDataRam(unsigned addr, unsigned value)
{
  // write the memory value to address E0
  rCB.CB_Write(0xE0, value);
  // write the memory address to write to E1
  rCB.CB_Write(0xE1, addr);
  // write a junk value to E4 to cause the data memory write
  rCB.CB_Write(0xE4, 0xFFFFFFFF);
}

void Driver_HL::ClearHLTandReset(void)
{
  // write a junk value to E8 to clear any halt and reset the processor
  rCB.CB_Write(0xE8, 0xFFFFFFFF);
}

void Driver_HL::SimWaitForHalt(void)
{
  rCB.CB_Wait_Equ(0xE6, 0x1);
}

bool Driver_HL::Halted(bool printStatus)
{
#if 1
  if(forSim)
    {
      SimWaitForHalt();
      return true;
    };
#endif

  cout << "halted initial start read" << endl;
  unsigned value = rCB.CB_Read(0xE6);
  cout << "halted initial read done" << endl;

    do{
      if(printStatus)
	cout << "read status: " << setw(8) << setfill('0') << hex << value << dec << endl;

      cout << "halted start read" << endl;
      value = rCB.CB_Read(0xE6);
      cout << "halted read done" << endl;
      value &= 0xFF;
    }while( value != 0 && value != 1);

  return (value & 0x1);
}

void Driver_HL::Blank(unsigned char pixValue)
{
  unsigned busValue = HWPixelToBusWidth( pixValue );

  #if 0
  //ii is start in bytes, we clear 256 dwords at a time
 for(unsigned ii = 0; ii < 640*480; ii += 4)
   {
     Write_FB((4*FB_OFFSET)+ii, busValue, true);
   };
 #else
   const unsigned LEN = 64;
   //ii is start in bytes, we clear 256 dwords at a time
   for(unsigned ii = 0; ii < 640*480; ii += LEN*4)
     {
       //we must wait here because the usb is faster than the BS unit
       Block_Set((4*FB_OFFSET)+ii, LEN, busValue, true);
     };
#endif
}

void Driver_HL::DrawPixel(int xPos, int yPos, unsigned char pixValue)
{
  unsigned address = yPos*640 + xPos;

  unsigned fbData = 0;
  Read_FB(address, fbData);

  /*
  switch(address & 0x00000003)
    {
    case 0:
      fbData = ((unsigned)pixValue)<<24 | (fbData&0x00FFFFFF);
      break;
    case 1:
      fbData = ((unsigned)pixValue)<<16 | (fbData&0xFF00FFFF);
      break;
    case 2:
      fbData = ((unsigned)pixValue)<<8 | (fbData&0xFFFF00FF);
      break;
    default: //3
      fbData = ((unsigned)pixValue) | (fbData&0xFFFFFF00);
      break;
    };
  */

  switch(address & 0x00000003)
    {
    case 0:
      fbData = ((unsigned)pixValue)<<24 | (fbData&0x00FFFFFF);
      break;
    case 1:
      fbData = ((unsigned)pixValue)<<16 | (fbData&0xFF00FFFF);
      break;
    case 2:
      fbData = ((unsigned)pixValue)<<8 | (fbData&0xFFFF00FF);
      break;
    default: //3
      fbData = ((unsigned)pixValue) | (fbData&0xFFFFFF00);
      break;
    };


  Write_FB(address, fbData);
}

void Driver_HL::DrawLine(int xStart, int yStart, int xEnd, int yEnd,
			 unsigned char pixValue)
{
  int deltaX = xEnd - xStart;
  int deltaY = yEnd - yStart;
  bool xRev = false;
  bool yRev = false;

  if(deltaX < 0)
    {
      xRev = true;
      deltaX *= -1;
    };

  if(deltaY < 0)
    {
      yRev = true;
      deltaY *= -1;
    };

  int lXStart = xStart;
  int lXEnd = xEnd;
  int lYStart = yStart;
  int lYEnd = yEnd;

  if(deltaY > deltaX)
    {
      cout << "y dominant line" << endl;
      if(yRev)
	{
	  lXStart = xEnd;
	  lXEnd = xStart;
	  lYStart = yEnd;
	  lYEnd = yStart;
	  deltaX *= -1;
	};

      for(int cY = lYStart; cY <= lYEnd; cY++)
	{
	  int xVal = lXStart +
	    (int)rint( ((float)deltaX)*((float)(cY - lYStart))/((float)deltaY) );
	  DrawPixel(xVal, cY, pixValue);
	  cout << xVal << " " << cY << endl;
	};
    }else
      {
	cout << "x dominant line" << endl;
	if(xRev)
	  {
	    lXStart = xEnd;
	    lXEnd = xStart;
	    lYStart = yEnd;
	    lYEnd = yStart;
	    deltaY *= -1;
	  };

	for(int cX = lXStart; cX < lXEnd; cX++)
	  {
	    int yVal = lYStart +
	      (int)rint( ((float)deltaY)*((float)(cX - lXStart))/((float)deltaX) );
	    DrawPixel(cX, yVal, pixValue);
	    cout << cX << " " << yVal << endl;
	  };
      };
}

void Driver_HL::Read_FB(unsigned fbAddress, unsigned &rValue)
{
  rCB.CB_Write(0x21, fbAddress);
  rCB.CB_Write(0x23, 0); //junk value

  WaitOnBR();

  rValue = rCB.CB_Read(0x22);
}

void Driver_HL::Write_FB(unsigned fbAddress, unsigned newValue, bool wait)
{
  rCB.CB_Write(0x11, fbAddress);
  rCB.CB_Write(0x12, newValue);
  rCB.CB_Write(0x13, 0); //trigger write

  if(wait)
    WaitOnBW();
}

void Driver_HL::Block_Set(unsigned fbAddress, unsigned numBytes, unsigned newValue, bool wait)
{
  rCB.CB_Write(0x01, fbAddress);
  rCB.CB_Write(0x02, numBytes );
  rCB.CB_Write(0x03, newValue);
  rCB.CB_Write(0x04, 0); //junk

  if( wait )
    WaitOnBS();
}

// **************** start EngineBase overloads *******************

void Driver_HL::Setup(void)
{
  currentTextureID = 0;

  //cut and paste from EngineXShaderCPU
  /*
  LoadDataRam(0, F2R(rC1.x) );
  LoadDataRam(1, F2R(rC1.y) );
  LoadDataRam(2, F2R(rC1.z) );
  LoadDataRam(3, F2R(rC2.x) );
  LoadDataRam(4, F2R(rC2.y) );
  LoadDataRam(5, F2R(rC2.z) );
  LoadDataRam(6, F2R(rC3.x) );
  LoadDataRam(7, F2R(rC3.y) );
  LoadDataRam(8, F2R(rC3.z) );
  LoadDataRam(9, F2R(t1.x) );
  LoadDataRam(10, F2R(t1.y) );
  LoadDataRam(11, F2R(t2.x) );
  LoadDataRam(12, F2R(t2.y) );
  LoadDataRam(13, F2R(t3.x) );
  LoadDataRam(14, F2R(t3.y) );
  */
  LoadDataRam(15, currentTextureID ); //current texture
  LoadDataRam(16, TEX_START ); //texStart
  LoadDataRam(17, TEX_SIZE );  //texSize 

  LoadDataRam(18, F2R(tForm.m[0][0]) ); //xf11
  LoadDataRam(19, F2R(tForm.m[0][1]) );
  LoadDataRam(20, F2R(tForm.m[0][2]) );
  LoadDataRam(21, F2R(tForm.m[0][3]) );
  LoadDataRam(22, F2R(tForm.m[1][0]) ); //xf21
  LoadDataRam(23, F2R(tForm.m[1][1]) );
  LoadDataRam(24, F2R(tForm.m[1][2]) );
  LoadDataRam(25, F2R(tForm.m[1][3]) );
  LoadDataRam(26, F2R(tForm.m[2][0]) ); //xf31
  LoadDataRam(27, F2R(tForm.m[2][1]) );
  LoadDataRam(28, F2R(tForm.m[2][2]) );
  LoadDataRam(29, F2R(tForm.m[2][3]) );
  LoadDataRam(30, F2R(tForm.m[3][0]) ); //xf41
  LoadDataRam(31, F2R(tForm.m[3][1]) );
  LoadDataRam(32, F2R(tForm.m[3][2]) );
  LoadDataRam(33, F2R(tForm.m[3][3]) );

  LoadDataRam(34, 640); //buffWidth
  LoadDataRam(35, FB_OFFSET); //frame buffer offset
  LoadDataRam(36, DB_OFFSET); //depth buffer offset
}

#if 0
void Driver_HL::TriFace(TCoord rC1, TCoord rC2, TCoord rC3, TexC t1, TexC t2, TexC t3)
{

  //buffer offset.
  int bo = 0;

  //int c_buffer_pos = 0;
  //unsigned bo = 100 + c_buffer_pos*16;

  LoadDataRam(bo+0, F2R(rC1.x) );
  LoadDataRam(bo+1, F2R(rC1.y) );
  LoadDataRam(bo+2, F2R(rC1.z) );
  LoadDataRam(bo+3, F2R(rC2.x) );
  LoadDataRam(bo+4, F2R(rC2.y) );
  LoadDataRam(bo+5, F2R(rC2.z) );
  LoadDataRam(bo+6, F2R(rC3.x) );
  LoadDataRam(bo+7, F2R(rC3.y) );
  LoadDataRam(bo+8, F2R(rC3.z) );
  LoadDataRam(bo+9, F2R(t1.x) );
  LoadDataRam(bo+10, F2R(t1.y) );
  LoadDataRam(bo+11, F2R(t2.x) );
  LoadDataRam(bo+12, F2R(t2.y) );
  LoadDataRam(bo+13, F2R(t3.x) );
  LoadDataRam(bo+14, F2R(t3.y) );
  LoadDataRam(bo+15, currentTextureID ); //current texture

  LoadDataRam(37, 1); //host_command 
  ClearHLTandReset();

  while( !Halted(true) ){};
}

void Driver_HL::Flush(void)
{}
#endif

#if 1
static int c_buffer_pos = 0;
void Driver_HL::TriFace(TCoord rC1, TCoord rC2, TCoord rC3, TexC t1, TexC t2, TexC t3)
{
  if(c_buffer_pos == 16) Flush();  //buffer full

  //buffer offset.
  unsigned bo = 100 + c_buffer_pos*16;

  LoadDataRam(bo+0, F2R(rC1.x) );
  LoadDataRam(bo+1, F2R(rC1.y) );
  LoadDataRam(bo+2, F2R(rC1.z) );
  LoadDataRam(bo+3, F2R(rC2.x) );
  LoadDataRam(bo+4, F2R(rC2.y) );
  LoadDataRam(bo+5, F2R(rC2.z) );
  LoadDataRam(bo+6, F2R(rC3.x) );
  LoadDataRam(bo+7, F2R(rC3.y) );
  LoadDataRam(bo+8, F2R(rC3.z) );
  LoadDataRam(bo+9, F2R(t1.x) );
  LoadDataRam(bo+10, F2R(t1.y) );
  LoadDataRam(bo+11, F2R(t2.x) );
  LoadDataRam(bo+12, F2R(t2.y) );
  LoadDataRam(bo+13, F2R(t3.x) );
  LoadDataRam(bo+14, F2R(t3.y) );
  LoadDataRam(bo+15, currentTextureID ); //current texture

  c_buffer_pos++;
}

void Driver_HL::Flush(void)
{
  LoadDataRam(37, 1); //host_command 
  LoadDataRam(38, c_buffer_pos); //host_argument
  c_buffer_pos = 0;

  ClearHLTandReset();

  while( !Halted(true) ){};
}
#endif

#if 0

void Driver_HL::TriFace(TCoord rC1, TCoord rC2, TCoord rC3, TexC t1, TexC t2, TexC t3)
{
  EngineSWBase::TriFace(rC1, rC2, rC3, t1, t2, t3);
}

void Driver_HL::Flush(void)
{}

#endif

void Driver_HL::ClearFB(void)
{
  cout << "Driver_HL::ClearFB" << endl;
  if(forSim) return;

#if 0
  ClearFB(0);
  return;
#endif


#if 1
  //LoadDataRam(35, (640*240)/4); //frame buffer offset

   LoadDataRam(37, 2); //host_command 
   LoadDataRam(38, 0x0000000); //host_argument
   ClearHLTandReset();
   while( !Halted(true) ){};
#else

  static int blank_id = 0;

  switch(blank_id)
    {
    case 0:
      ClearFB(0);
      break;
    case 1:
      ClearFB(0xE0);
      break;
    case 2:
      ClearFB(0x1C);
      break;
    case 3:
      ClearFB(0x3);
      break;
    };

  blank_id++;
  if(blank_id == 4) blank_id = 0;
#endif
}

#include <unistd.h>

void Driver_HL::ClearDB(void)
{
  cout << "Driver_HL::ClearDB" << endl;
  if( forSim ) return;

  //at -100000
  int value = -10000000;
  unsigned value2;
  assert( sizeof(value) == 4);
  assert( sizeof(value2) == 4 );
  memcpy(&value2, &value, 4);


#if 0
  ClearDB(value2);
#else
  LoadDataRam(37, 3); //host_command 
  LoadDataRam(38, value2 ); //host_argument
  ClearHLTandReset();
  while( !Halted(true) ){};
#endif
}
 
void Driver_HL::UpLoadTexture(Texture &rTexture, unsigned textureID)
{
  unsigned texAddress = 0;
  switch(textureID)
    {
    case 1:
      //texAddress = 0x300000;
      texAddress = TEX_START;
      break;
    case 2:
      //texAddress = 0x380000;
      texAddress = TEX_START + 1*TEX_SIZE;
      break;
    case 3:
      //texAddress = 0x400000;
      texAddress = TEX_START + 2*TEX_SIZE;
      break;
    case 4:
      //texAddress = 0x480000;
      texAddress = TEX_START + 3*TEX_SIZE;
      break;
    case 5:
      //texAddress = 0x500000;
      texAddress = TEX_START + 4*TEX_SIZE;
      break;
    case 6:
      //texAddress = 0x580000;
      texAddress = TEX_START + 5*TEX_SIZE;
      break;
    case 7:
      //texAddress = 0x600000;
      texAddress = TEX_START + 6*TEX_SIZE;
      break;
    default:
      cerr << "invalid texture id mentioned in Driver_HL::UpLoadTexture. aborting." << endl;
      return;
    };

  texAddress *= 4;

  if( !rTexture.pPixelData )
    {
      cerr << "Driver_HL::UpLoadTexture texture data seems to be missing! aborting." << endl;
      return;
    };

  if( rTexture.texWidth != 256 || rTexture.texHeight != 256)
    {
      cerr << "Driver_HL::UpLoadTexture texture has invalid size. aborting." << endl;
      return;
    };

  cout << "Driver_HL::UpLoadTexture " << textureID << " starting..." << endl;
  for(unsigned ii = 0; ii < rTexture.texHeight; ii++)
    for(unsigned jj = 0; jj < rTexture.texWidth; jj++)
      {
	EPIXEL pixValue = rTexture.Lookup(jj, ii);

	/*
	 unsigned redComp   = 0xFF& ((unsigned)( ( (pixValue & EPIXEL_RED_MASK)>>EPIXEL_RED_OFFSET ) ));
	 unsigned greenComp = 0xFF& ((unsigned)( ( (pixValue & EPIXEL_GREEN_MASK)>>EPIXEL_GREEN_OFFSET ) ));
	 unsigned blueComp  = 0xFF& ((unsigned)( ( (pixValue & EPIXEL_BLUE_MASK)>>EPIXEL_BLUE_OFFSET   ) ));
	*/

	EPIXEL retValue = pixValue;

	Write_FB(texAddress, retValue, false);
	texAddress += 4;
      };

  WaitOnBW(); //forces a sync.
  cout << "Driver_HL::UpLoadTexture " << textureID << " done." << endl;
}

void Driver_HL::SetCTexture(unsigned cTextureID)
{
  currentTextureID = cTextureID;
}

// **************** end EngineBase overloads *********************

// ***************** start EDriver overloads ******************
void Driver_HL::PolyScanLine(unsigned sX, long startZ, unsigned eX, long endZ, unsigned yValue, TexC texStart, TexC texInc, EPIXEL pixValue)
{
#if 1
  cout << "Driver_HL::PolyScanLine sX: " << sX << " startZ: " << startZ 
       << " eX: " << eX << " endZ: " << endZ << " yValue: " << yValue
       << " texStart: " << texStart << " texinc: " << texInc << endl;
#endif

  unsigned fbAddress = yValue*640 + sX + 4*FB_OFFSET;
  unsigned dbAddress = 4*(yValue*640 + sX + DB_OFFSET);

  int zIncrement = 0;
  if( eX != sX )
    zIncrement = (endZ - startZ)/((int)(eX - sX));

  unsigned numPixels = eX - sX;

  rCB.CB_Write(0xf0, fbAddress);
  rCB.CB_Write(0xf1, dbAddress);
  rCB.CB_Write(0xf2, startZ);
  rCB.CB_Write(0xf3, zIncrement);

  rCB.CB_Write(0xf4, numPixels); //only lowest 16 bits are used...
  rCB.CB_Write(0xf5, (int)EPixelToHWPixel(pixValue) ); //only lowest 8 bits are used.

  if( currentTextureID != 0 )
    {
      int tc_x_start = rint(texStart.x*65535.00);
      int tc_y_start = rint(texStart.y*65535.00);
      int tc_x_inc   = rint(texInc.x*65535.00);
      int tc_y_inc   = rint(texInc.y*65535.00);

       // int tc_x_start = rint(texStart.x*255.00);
       // int tc_y_start = rint(texStart.y*255.00);
       // int tc_x_inc   = rint(texInc.x*255.00);
       // int tc_y_inc   = rint(texInc.y*255.00);

      rCB.CB_Write(0xf6, ((tc_x_start<<16)&0xFFFF0000) | (tc_y_start&0x0000FFFF) );
      rCB.CB_Write(0xf7, ((tc_x_inc<<16)&0xFFFF0000)   | (tc_y_inc&0x0000FFFF) );
    };

  rCB.CB_Write(0xf8, currentTextureID);

  /*
  localDataRegisters[17] = numPixels&0x000000FF;
  localDataRegisters[16] = (numPixels&0x0000FF00)>>8;
  localDataRegisters[18] = (int)EPixelToHWPixel(pixValue);
  */
  
  rCB.CB_Write(0xF9, 0);

  WaitOnPS();
}

void Driver_HL::ClearFB(EPIXEL pixValue)
{
  Blank(pixValue);
}

void Driver_HL::ClearDB(unsigned depthVal)
{
 
  const unsigned LEN = 64;
  //ii is start offset in bytes, we clear 256 dwords at a time
  for(unsigned ii = 0; ii < 640*480*4; ii += LEN*4)
    {
      //we must wait here because the usb is faster than the BS unit
      Block_Set( (DB_OFFSET*4)+ii, LEN, depthVal, true);
    };
}

// **************** end EngineBase overloads *********************

unsigned char Driver_HL::EPixelToHWPixel(EPIXEL pixValue)
{
  //a very painful pixel conversion
  unsigned char redComp   = (unsigned char)( ( (pixValue & EPIXEL_RED_MASK)>>EPIXEL_RED_OFFSET ) & 0xC0 );
  unsigned char greenComp = (unsigned char)( (( (pixValue & EPIXEL_GREEN_MASK)>>EPIXEL_GREEN_OFFSET ) & 0xE0) >> 2 );
  unsigned char blueComp  = (unsigned char)( (( (pixValue & EPIXEL_BLUE_MASK)>>EPIXEL_BLUE_OFFSET   ) & 0xE0) >> 5 );
  return redComp | greenComp | blueComp;
}

unsigned Driver_HL::HWPixelToBusWidth(unsigned char hwPixel)
{
  return ((unsigned)hwPixel)<<24 | ((unsigned)hwPixel)<<16 | ((unsigned)hwPixel)<<8 | ((unsigned)hwPixel);
}

void Driver_HL::WaitOnBW(void)
{
  for(unsigned ii = 0; ii < 1000; ii++)
    {
      unsigned status = rCB.CB_Read(0xFF);
      if( status & 0x02 ) return;
    }

  cerr << "timeout in Driver_HL::WaitOnBW" << endl;
  assert(false);
}

void Driver_HL::WaitOnBR(void)
{
  for(unsigned ii = 0; ii < 1000; ii++)
    {
      unsigned status = rCB.CB_Read(0xFF);
      if( status & 0x01 ) return;
    }

  cerr << "timeout in Driver_HL::WaitOnBR" << endl;
  assert(false);
}

void Driver_HL::WaitOnBS(void)
{
  for(unsigned ii = 0; ii < 1000; ii++)
    {
      unsigned status = rCB.CB_Read(0xFF);
      if( status & 0x04 ) return;
    }

  cerr << "timeout in Driver_HL::WaitOnBS" << endl;
  assert(false);
}

void Driver_HL::WaitOnPS(void)
{
  for(unsigned ii = 0; ii < 1000000; ii++)
    {
      unsigned status = rCB.CB_Read(0xFF);
      if( status & 0x08 ) return;
      cout << "Driver_HL::WaitOnPS ps is busy: " << ii << endl;
    }

  cerr << "timeout in Driver_HL::WaitOnPS" << endl;
  assert(false);
}
