/*
  file created 2014
  modified
  released to the Public Domain by the author
*/

#include <iostream>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <iomanip>

#include "ShimXShaderCPU.hh"

using namespace std;

// ******************************************

ShimXShaderCPU::ShimXShaderCPU(FB_DEPTH tFBDepth) :
  CPU(), fbDepth(tFBDepth),
  xImgWindow(640,480), currentTex(0), frameBuffer(640, 480),
  refreshOnDMA(true), refreshCount(0), fb(2000000)
{
  for(unsigned ii = 0; ii < fb.size(); ii++)
    fb[ii] = 0;

  visualFBChanged = false;
  refreshCount = 0;
  lExecCount = 0;

  dma_cmd_0_reg = 0;
  dma_addr_0_reg = 0;
  dma_cmd_1_reg = 0;
  dma_addr_1_reg = 0;
  dma_cmd_2_reg = 0;
  dma_addr_2_reg = 0;
  dma_cmd_3_reg = 0;
  dma_addr_3_reg = 0;
}

ShimXShaderCPU::~ShimXShaderCPU(void)
{}

void ShimXShaderCPU::CreateRenderContext(void)
{
  if( !xImgWindow.CreateAll() )
    {
      cerr << "error opening x" << endl;
      assert(false);
    };
  
  frameBuffer.Initialize();
}

void ShimXShaderCPU::DeleteRenderContext(void)
{
  xImgWindow.DeleteAll();
}

void ShimXShaderCPU::ResetAndExecute(void)
{
  cout << "ShimXShaderCPU::ResetAndExecute" << endl;
  Reset();
  Execute();
  //cout << "ShimXShaderCPU::ResetAndExecute done" << endl;
}

void ShimXShaderCPU::LocalWaitForHLT(void)
{}

void ShimXShaderCPU::Flush(void)
{
  //cout << "ShimXShaderCPU::Flush" << endl;
  UpdateFB();
  xImgWindow.UpdateWindow(frameBuffer);
}

unsigned long long ShimXShaderCPU::FullExecCount(void)
{
  return lExecCount;
}

void ShimXShaderCPU::ResetFullExecCount(void)
{
  lExecCount = 0;
}

void ShimXShaderCPU::UpdateFB(void)
{
  if( fbDepth == FB_32BPP )
    {
      for(unsigned ii = 0; ii < 480; ii++)
	for(unsigned jj = 0; jj < 640; jj++)
	  frameBuffer.Pixel(jj, ii) = fb[ii*640+jj];
    };

  if( fbDepth == FB_8BPP )
    {
      for(unsigned ii = 0; ii < 480; ii++)
	for(unsigned jj = 0; jj < 640; jj++)
	  {
	    unsigned pixel_4x_dword_addr = (ii*640+jj);
	    unsigned pixel_dword_addr = pixel_4x_dword_addr>>2;
	    
	    unsigned pix_value_4x = fb[pixel_dword_addr];
	    unsigned pix_value = 0;
	    switch( 0x3&pixel_4x_dword_addr)
	      { //not really respecting endian, but whatever.
	      case 3:
		pix_value = pix_value_4x&0x000000FF;
		break;
	      case 2:
		pix_value = (pix_value_4x&0x0000FF00)>>8;
		break;
	      case 1:
		pix_value = (pix_value_4x&0x00FF0000)>>16;
		break;
	      default: //case 0:
		pix_value = (pix_value_4x&0xFF000000)>>24;
		break;
	      };

	    unsigned red = (pix_value&0xE0)>>5;
	    unsigned green = (pix_value&0x1c)>>2;
	    unsigned blue = (pix_value&0x3);

	    red *= 255; red /= 7; //mul before divide
	    green *= 255; green /= 7; //mul before divide
	    blue *= 255; blue /= 3; //mul before divide

	    EPIXEL pixValue = (red<<24) | (green<<16) | (blue<<8);
	    frameBuffer.Pixel(jj, ii) = pixValue;
	  };
    };

  visualFBChanged = false;
};

CPU::REG_TYPE ShimXShaderCPU::InvalidDataRead(REG_TYPE dwordAddr)
{
  switch(dwordAddr)
    {
    case 0xFFF0:
      return dma_cmd_0_reg;
    case 0xFFF1:
      return dma_addr_0_reg;
    case 0xFFF2:
      return dma_cmd_1_reg;
    case 0xFFF3:
      return dma_addr_1_reg;
    case 0xFFF4:
      return dma_cmd_2_reg;
    case 0xFFF5:
      return dma_addr_2_reg;
    case 0xFFF6:
      return dma_cmd_3_reg;
    case 0xFFF7:
      return dma_addr_3_reg;
    case 0xFFF8:
      return 0; //dma is always not busy in the sim
    default:
      break; //problems...
    };

  return CPU::InvalidDataRead(dwordAddr);
}

void ShimXShaderCPU::InvalidDataWrite(REG_TYPE dwordAddr, REG_TYPE value)
{
  switch(dwordAddr)
    {
    case 0xFFF0:
      dma_cmd_0_reg = value;
      return;
    case 0xFFF1:
      dma_addr_0_reg = value;
      return;
    case 0xFFF2:
      dma_cmd_1_reg = value;
      return;
    case 0xFFF3:
      dma_addr_1_reg = value;
      return;
    case 0xFFF4:
      dma_cmd_2_reg = value;
      return;
    case 0xFFF5:
      dma_addr_2_reg = value;
      return;
    case 0xFFF6:
      dma_cmd_3_reg = value;
      return;
    case 0xFFF7:
      dma_addr_3_reg = value;
      return;
    case 0xFFF8:
      break; //handle dma write below
    default: //problem unknown address
      CPU::InvalidDataWrite(dwordAddr, value);
      return;
    };

  assert(dwordAddr == 0xFFF8);
  //cout << "trapped write to dma unit: " << dwordAddr << endl;

  unsigned numXfers = value;
  if(numXfers > 4)
    {
      cout << "illegal num of xfers" << endl;
      return;
    };

  if( numXfers >= 1) DoDMA(dma_cmd_0_reg, dma_addr_0_reg);

  if( numXfers >= 2) DoDMA(dma_cmd_1_reg, dma_addr_1_reg);

  if( numXfers >= 3) DoDMA(dma_cmd_2_reg, dma_addr_2_reg);

  if( numXfers >= 4) DoDMA(dma_cmd_3_reg, dma_addr_3_reg);
}

void ShimXShaderCPU::DoDMA(REG_TYPE cmdReg, REG_TYPE fbAddr)
{
  //cout << "ShimXShaderCPU::DoDMA " << hex << cmdReg << " " << fbAddr << dec << endl;

  REG_TYPE localAddr = cmdReg&0xFFFF;
  unsigned count = (cmdReg&0xFFF0000)>>16;
  bool toFB = !(cmdReg&0x80000000);

  if(localAddr+count > data_ram.size() ||
     fbAddr+count > fb.size()  ) //|| count > 8  //the 8 was a check to find a bug...
    {
      cout << "invalid dma cmd:" << endl;
      cout << "ShimXShaderCPU::DoDMA localAddr: " << localAddr << " fbAddr: " << fbAddr << " count: " << count << " toFB: " << toFB << endl;
      return;
    };

  localAddr = localAddr;
  //cout << "new localAddr: " << localAddr << endl;

  for(unsigned ii = 0; ii < count; ii++)
    {
      if(toFB)
	{
	  unsigned value = data_ram[localAddr+ii];
	  unsigned fbAd = fbAddr + ii;
	  if(fbAd < 640*480)
	    {
	      visualFBChanged = true;
	      //value = 0xFFFFFFFF;
	    };
	  //cout << "writing value" << value << endl;
	  fb[fbAd] = value;
	}else
	{
	  data_ram[localAddr+ii] = fb[fbAddr + ii];
	};
    };

  //Flush();

  if( refreshOnDMA ) //if not true, maybe no window...
    {
      if(toFB)
	refreshCount++;

      if(refreshCount > 9 )
	{
	  refreshCount = 0;
	  Flush();
	};
    };
}
