-- Copyright 2010,2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.shader_cpu_pack.all;

--for multiplier
library UNISIM;
use UNISIM.VComponents.all;


entity alu is
  port(
    opcode : in std_logic_vector(6 downto 0);

    bus_a : in std_logic_vector(31 downto 0);
    bus_b : in std_logic_vector(31 downto 0);

    imm          : in std_logic_vector(15 downto 0);
    imm_sign_ext : in std_logic_vector(31 downto 0);  --sign extended immediate

    bit_shift_out : out std_logic_vector(31 downto 0);  --other places use this value?

    result : out std_logic_vector(31 downto 0)
    );
end entity;

architecture behaviour of alu is

  signal shift_L, shift_L_2, shift_L_4, shift_L_8, shift_L_16 : std_logic_vector(31 downto 0);
  signal shift_R, shift_R_2, shift_R_4, shift_R_8, shift_R_16 : std_logic_vector(31 downto 0);

  signal bit_shift, bit_shift_2, bit_shift_4, bit_shift_8, bit_shift_16 : std_logic_vector(31 downto 0);

  signal mul_result : std_logic_vector(35 downto 0);

  signal cmp_equal, cmp_gt, cmp_gts, cmp_lt, cmp_lts : std_logic;
  signal cmp_result                                  : std_logic_vector(31 downto 0);

begin

  mul : MULT18X18
    port map(P => mul_result, A => bus_a(17 downto 0), B => bus_b(17 downto 0));
  
  bit_shift_16  <= X"0001_0000"                      when imm(4) = '1' else X"0000_0001";
  bit_shift_8   <= bit_shift_16(23 downto 0) & X"00" when imm(3) = '1' else bit_shift_16;
  bit_shift_4   <= bit_shift_8(27 downto 0) & X"0"   when imm(2) = '1' else bit_shift_8;
  bit_shift_2   <= bit_shift_4(29 downto 0) & "00"   when imm(1) = '1' else bit_shift_4;
  bit_shift     <= bit_shift_2(30 downto 0) & "0"    when imm(0) = '1' else bit_shift_2;
  bit_shift_out <= bit_shift;

  shift_L_16 <= bus_a(15 downto 0) & X"0000"    when imm(4) = '1' else bus_a;
  shift_L_8  <= shift_L_16(23 downto 0) & X"00" when imm(3) = '1' else shift_L_16;
  shift_L_4  <= shift_L_8(27 downto 0) & X"0"   when imm(2) = '1' else shift_L_8;
  shift_L_2  <= shift_L_4(29 downto 0) & "00"   when imm(1) = '1' else shift_L_4;
  shift_L    <= shift_L_2(30 downto 0) & "0"    when imm(0) = '1' else shift_L_2;

  shift_R_16 <= X"0000" & bus_a(31 downto 16)   when imm(4) = '1' else bus_a;
  shift_R_8  <= X"00" & shift_R_16(31 downto 8) when imm(3) = '1' else shift_R_16;
  shift_R_4  <= X"0" & shift_R_8(31 downto 4)   when imm(2) = '1' else shift_R_8;
  shift_R_2  <= "00" & shift_R_4(31 downto 2)   when imm(1) = '1' else shift_R_4;
  shift_R    <= "0" & shift_R_2(31 downto 1)    when imm(0) = '1' else shift_R_2;

  cmp_equal  <= '1' when bus_a = bus_b                     else '0';
  cmp_gt     <= '1' when unsigned(bus_a) > unsigned(bus_b) else '0';
  cmp_gts    <= '1' when signed(bus_a) > signed(bus_b)     else '0';
  cmp_lt     <= '1' when unsigned(bus_a) < unsigned(bus_b) else '0';
  cmp_lts    <= '1' when signed(bus_a) < signed(bus_b)     else '0';
  cmp_result <= X"0000_00" & "000" & cmp_lts & cmp_lt & cmp_gts & cmp_gt & cmp_equal;

  --these are values that will be directly writtten into the destination
  --register only
  result <=
    imm & bus_a(15 downto 0)                               when opcode = INS_LUI else
    bus_a(31 downto 16) & imm                              when opcode = INS_LLI else
    std_logic_vector(unsigned(bus_a) + unsigned(bus_b))    when opcode = INS_ADD else
    std_logic_vector(signed(bus_a) - signed(bus_b))        when opcode = INS_SUB else
    std_logic_vector(signed(bus_a) + signed(imm_sign_ext)) when opcode = INS_ADDL else  --?correct?
    bus_a and bus_b                                        when opcode = INS_AND else
    bus_a or bus_b                                         when opcode = INS_OR else
    bus_a xor bus_b                                        when opcode = INS_XOR else
    not bus_a                                              when opcode = INS_NOT else
    bus_a or bit_shift                                     when opcode = INS_BSET else
    bus_a and not bit_shift                                when opcode = INS_BCLR else
    shift_L                                                when opcode = INS_RSL else
    shift_R                                                when opcode = INS_RSR else
    mul_result(31 downto 0)                                when opcode = INS_MUL else
    cmp_result                                             when opcode = INS_CMP else
    bus_a                                                  when opcode = INS_SRI else
    bus_a                                                  when opcode = INS_SRR else
    bus_a                                                  when opcode = INS_JR else
    (others => '0');

end architecture;
