-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity shader_cb_interface is
  port(
    clock : in std_logic;
    reset : in std_logic;

    cmd_bus_data_in  : in  std_logic_vector(31 downto 0);
    cmd_bus_data_out : out std_logic_vector(31 downto 0);
    cmd_bus_addr     : in  std_logic_vector(7 downto 0);
    cmd_bus_wr       : in  std_logic;
    cmd_bus_strobe   : in  std_logic;

    i_ram_strobe : out std_logic;       --trigger a rw opp of instruction ram
    d_ram_strobe : out std_logic;       --trigger a rw opp of data ram

    i_ram_busy : in std_logic;
    d_ram_busy : in std_logic;

    i_ram_data : in std_logic_vector(31 downto 0);
    d_ram_data : in std_logic_vector(31 downto 0);

    ram_wr_data : out std_logic_vector(31 downto 0);
    ram_addr    : out std_logic_vector(15 downto 0);
    ram_we      : out std_logic;

    pc_addr            : in  std_logic_vector(15 downto 0);  --hackish, but gives some diagnostics.
    shader_halt_status : in  std_logic;
    shader_reset       : out std_logic;
    shader_halt        : out std_logic
    );
end shader_cb_interface;

architecture rtl of shader_cb_interface is

  signal do_output : std_logic;
  signal data_out  : std_logic_vector(31 downto 0);

  signal rw_data    : std_logic_vector(31 downto 0) := X"0000_0000";
  signal rw_address : std_logic_vector(15 downto 0) := X"0000";

  signal id_write : std_logic := '0';

  type CB_INT_STATE is (CB_IDLE, CB_I_READ, CB_I_WRITE, CB_D_READ, CB_D_WRITE);
  signal cb_state : CB_INT_STATE;

  signal local_cpu_reset : std_logic := '0';
  signal local_cpu_halt  : std_logic := '1';
  
begin

  shader_reset <= local_cpu_reset;
  shader_halt  <= local_cpu_halt;

  ram_wr_data <= rw_data;
  ram_addr    <= rw_address;
  ram_we      <= id_write;

  data_out <=
    rw_data                                       when cmd_bus_addr(3 downto 0) = X"0" else
    X"0000" & rw_address                          when cmd_bus_addr(3 downto 0) = X"1" else
    pc_addr & X"000" & "000" & shader_halt_status when cmd_bus_addr(3 downto 0) = X"6" else
    X"0000_0000";
  
  cmd_bus_data_out <= data_out when do_output = '1' else (others => 'Z');

  do_output <= '1' when cmd_bus_strobe = '1' and cmd_bus_addr(7 downto 4) = X"E" else '0';

  process(clock)
    variable next_state     : CB_INT_STATE;
    variable do_strobe_iram : boolean;
    variable do_strobe_dram : boolean;
    variable do_cpu_reset   : boolean;
  begin
    if rising_edge(clock) then
      next_state     := cb_state;
      do_strobe_iram := false;
      do_strobe_dram := false;
      do_cpu_reset   := false;

      if reset = '1' then
        next_state := CB_IDLE;
      else
        next_state := cb_state;

        case cb_state is
          when CB_IDLE =>
            if cmd_bus_strobe = '1' and cmd_bus_addr(7 downto 4) = X"E" and cmd_bus_wr = '1' then
              case cmd_bus_addr(3 downto 0) is
                when X"0" =>            --write an address
                  rw_data <= cmd_bus_data_in;
                when X"1" =>
                  rw_address <= cmd_bus_data_in(15 downto 0);
                when X"2" =>
                  next_state     := CB_I_WRITE;
                  do_strobe_iram := true;
                  id_write       <= '1';
                when X"3" =>
                  next_state     := CB_I_READ;
                  do_strobe_iram := true;
                  id_write       <= '0';
                when X"4" =>
                  next_state     := CB_D_WRITE;
                  do_strobe_dram := true;
                  id_write       <= '1';
                when X"5" =>
                  next_state     := CB_D_READ;
                  do_strobe_dram := true;
                  id_write       <= '0';
                when X"6" =>
                  local_cpu_halt <= '1';
                when X"7" =>
                  local_cpu_halt <= '0';
                when X"8" =>
                  local_cpu_halt <= '0';
                  do_cpu_reset   := true;
                when others =>
                  null;
              end case;
            end if;
          when CB_I_READ =>
            if i_ram_busy = '0' then
              rw_data    <= i_ram_data;
              next_state := CB_IDLE;
            end if;
          when CB_I_WRITE =>
            if i_ram_busy = '0' then
              next_state := CB_IDLE;
            end if;
          when CB_D_READ =>
            if d_ram_busy = '0' then
              rw_data    <= d_ram_data;
              next_state := CB_IDLE;
            end if;
          when CB_D_WRITE =>
            if d_ram_busy = '0' then
              next_state := CB_IDLE;
            end if;
          when others =>
            null;
        end case;
      end if;  --reset

      if do_cpu_reset = true then
        local_cpu_reset <= '1';
      else
        local_cpu_reset <= '0';
      end if;

      if do_strobe_iram = true then
        i_ram_strobe <= '1';
      else
        i_ram_strobe <= '0';
      end if;

      if do_strobe_dram = true then
        d_ram_strobe <= '1';
      else
        d_ram_strobe <= '0';
      end if;

      cb_state <= next_state;
    end if;  --clock
  end process;

end architecture;
