library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.shader_cpu_pack.all;

--this block contains the shader cpu; the brams for the program and memories
--the dma unit and the user control unit (if needed)
entity shader_meta_block is
  port(
    clock : in std_logic;
    reset : in std_logic;

    cmd_bus_data_in  : in  std_logic_vector(31 downto 0);
    cmd_bus_data_out : out std_logic_vector(31 downto 0);
    cmd_bus_addr     : in  std_logic_vector(7 downto 0);
    cmd_bus_wr       : in  std_logic;
    cmd_bus_strobe   : in  std_logic;

    mu_bus_address : in    std_logic_vector(1 downto 0);
    mu_bus_data    : inout std_logic_vector(31 downto 0);
    mu_bus_write   : in    std_logic;
    mu_bus_strobe  : in    std_logic;
    mu_cpu_dma_irq : out   std_logic;
    mu_cpu_dma_sel : in    std_logic
    );
end shader_meta_block;

architecture rtl of shader_meta_block is

  signal cb_shader_reset : std_logic;
  signal cb_shader_halt  : std_logic;
  signal cb_halt_status  : std_logic;

  signal pc          : std_logic_vector(15 downto 0);
  signal instruction : std_logic_vector(31 downto 0);
  
  signal d_ram_addr : std_logic_vector(15 downto 0);
  signal d_ram_dout : std_logic_vector(31 downto 0);
  signal d_ram_din  : std_logic_vector(31 downto 0);
  signal d_ram_we   : std_logic;

  signal dma_d_ram_addr   : std_logic_vector(15 downto 0);
  signal dma_d_ram_dout   : std_logic_vector(31 downto 0);
  signal dma_d_ram_din    : std_logic_vector(31 downto 0);
  signal dma_d_ram_we     : std_logic;
  signal dma_d_ram_strobe : std_logic;
  
  signal ext_data_strobe  : std_logic;
  signal ext_data_address : std_logic_vector(7 downto 0);
  signal ext_data_out     : std_logic_vector(31 downto 0);
  signal ext_data_in      : std_logic_vector(31 downto 0);
  signal ext_data_wr      : std_logic;
  signal ext_data_busy    : std_logic;

  signal cb_int_inst_strobe : std_logic;
  signal cb_int_data_strobe : std_logic;
  signal cb_int_we          : std_logic;
  signal cb_int_addr        : std_logic_vector(15 downto 0);
  signal cb_int_dout        : std_logic_vector(31 downto 0);
  signal cb_int_inst_dout   : std_logic_vector(31 downto 0);
  signal cb_int_data_dout   : std_logic_vector(31 downto 0);
  
begin

  cpu_core : shader_cpu
    port map(clock => clock, reset_ll => reset, logic_reset=>cb_shader_reset, logic_halt=>cb_shader_halt, halt_status=>cb_halt_status,
             prog_address => pc, ins_dword => instruction,
             d_ram_addr => d_ram_addr, d_ram_dout => d_ram_dout, d_ram_din => d_ram_din, d_ram_we => d_ram_we,
             ext_data_strobe => ext_data_strobe, ext_data_address => ext_data_address, ext_data_out => ext_data_out, ext_data_in => ext_data_in,
             ext_data_wr => ext_data_wr, ext_data_busy => ext_data_busy);

  ins_ram : ram_inst
    port map(clock         => clock, cpu_pc => pc, cpu_inst => instruction,
             cb_int_strobe => cb_int_inst_strobe,
             cb_int_we     => cb_int_we,
             cb_int_addr   => cb_int_addr,
             cb_int_dout   => cb_int_inst_dout,
             cb_int_din    => cb_int_dout);

  dma_unit :cpu_dma_unit
  port map(clock=>clock, reset=>reset,
           mu_bus_addr=>mu_bus_address, mu_bus_data=>mu_bus_data, mu_bus_write=>mu_bus_write, mu_bus_strobe=>mu_bus_strobe,
           mu_cpu_irq =>mu_cpu_dma_irq, mu_cpu_sel=>mu_cpu_dma_sel,
           bram_addr => dma_d_ram_addr, bram_din=>dma_d_ram_dout, bram_dout=>dma_d_ram_din, bram_we=>dma_d_ram_we, bram_strobe=>dma_d_ram_strobe,
           ext_cpu_strobe=>ext_data_strobe, ext_cpu_addr=>ext_data_address, ext_cpu_dout=>ext_data_in,
           ext_cpu_din=>ext_data_out, ext_cpu_we=>ext_data_wr, ext_cpu_busy=>ext_data_busy,
           cb_strobe=>cb_int_data_strobe, cb_addr=>cb_int_addr, cb_din=>cb_int_dout, cb_dout=>cb_int_data_dout, cb_we=>cb_int_we);


  data_ram : ram_data
    port map(clock          => clock,
             cpu_addr       => d_ram_addr, cpu_din => d_ram_dout, cpu_dout => d_ram_din, cpu_we => d_ram_we,
             dma_int_strobe => dma_d_ram_strobe, dma_int_we => dma_d_ram_we, dma_int_addr => dma_d_ram_addr,
             dma_int_dout => dma_d_ram_dout, dma_int_din => dma_d_ram_din);

  cb_interface : shader_cb_interface
    port map(
      clock        => clock, reset => reset,
      cmd_bus_data_in => cmd_bus_data_in, cmd_bus_data_out => cmd_bus_data_out, cmd_bus_addr => cmd_bus_addr,
      cmd_bus_wr=>cmd_bus_wr, cmd_bus_strobe => cmd_bus_strobe, 
      i_ram_strobe => cb_int_inst_strobe,
      d_ram_strobe => cb_int_data_strobe,
      i_ram_busy   => '0', d_ram_busy => '0',
      i_ram_data   => cb_int_inst_dout,
      d_ram_data   => cb_int_data_dout,
      ram_wr_data  => cb_int_dout,
      ram_addr     => cb_int_addr,
      ram_we       => cb_int_we,
      pc_addr      => pc, shader_halt_status=>cb_halt_status,
      shader_reset  => cb_shader_reset, shader_halt => cb_shader_halt);

end architecture;
