-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use std.textio.all;
use ieee.std_logic_textio.all;

use work.dma_sim_pack.all;

--this is going to act like a memory controller to the cpu (only client)
--and echo requests

entity tb_mc is
  port(
    clk   : in std_logic;
    reset : in std_logic;

    mu_bus_address : out   std_logic_vector(1 downto 0);
    mu_bus_data    : inout std_logic_vector(31 downto 0);
    mu_bus_write   : out   std_logic;
    mu_bus_strobe  : out   std_logic;
    mu_cpu_dma_irq : in    std_logic;
    mu_cpu_dma_sel : out   std_logic
    );
end tb_mc;

architecture behaviour of tb_mc is

  type MC_INT_STATE is (MC_IDLE, MC_CFG_1, MC_CFG_2, MC_READ_STARTUP, MC_READ, MC_WRITE_STARTUP, MC_WRITE, MC_READ_CLEANUP, MC_CLEANUP);
  signal mc_state : MC_INT_STATE;

  signal mu_addr_reg        : std_logic_vector(1 downto 0);
  signal mu_bus_write_value : std_logic_vector(31 downto 0);

  signal local_cmd_is_write   : std_logic;  --if 1 is write to the fb from the MU
  signal local_cmd_xfer_count : unsigned(11 downto 0);
  signal local_cmd_fb_addr    : std_logic_vector(31 downto 0);

  signal xfer_count : unsigned(11 downto 0);
begin

  mu_cpu_dma_sel <= '1'                when mc_state /= MC_IDLE                                  else '0';
  mu_bus_address <= mu_addr_reg;
  mu_bus_write   <= '1'                when (mc_state = MC_READ) or (mc_state = MC_READ_CLEANUP) else '0';
  mu_bus_data    <= mu_bus_write_value when (mc_state = MC_READ) or (mc_state = MC_READ_CLEANUP) else (others => 'Z');

  process(clk) is
    variable out_line         : line;
    variable next_state       : MC_INT_STATE;
    variable do_inc_xfer      : boolean;
    variable do_reset_xfer    : boolean;
    variable do_mu_bus_strobe : boolean;
    variable read_value       : std_logic_vector(31 downto 0);
  begin

    if rising_edge(clk) then
      next_state       := mc_state;
      do_mu_bus_strobe := false;
      do_reset_xfer    := false;
      do_inc_xfer      := false;

      if reset = '1' then
        next_state    := MC_IDLE;
        mu_addr_reg   <= "00";
        do_reset_xfer := true;
      else
        case mc_state is
          when MC_IDLE =>
            if mu_cpu_dma_irq = '1' then  --start config read
              mu_addr_reg <= "01";
              next_state  := MC_CFG_1;
            end if;
          when MC_CFG_1=>
            local_cmd_is_write   <= mu_bus_data(31);
            local_cmd_xfer_count <= unsigned(mu_bus_data(11 downto 0));
            mu_addr_reg          <= "10";
            next_state           := MC_CFG_2;
          when MC_CFG_2=>
            --mu bus addresses are (still) in bytes, but our
            --remote fb deals in dword increment addresses
            local_cmd_fb_addr <= "00" & mu_bus_data(31 downto 2);
            mu_addr_reg       <= "00";
            do_reset_xfer     := true;
            if local_cmd_xfer_count = X"000" then
              --we have selected the MU by our change in state to MC_CFG_x
              --we will delect the MU by returning to IDLE. the MU should
              --be satisfied.
              next_state := MC_IDLE;
            else
              if local_cmd_is_write = '1' then
                next_state := MC_WRITE_STARTUP;
              else
                next_state := MC_READ_STARTUP;
              end if;
            end if;
          when MC_READ_STARTUP =>
            write(out_line, string'("DMA IN: "));
            hwrite(out_line, local_cmd_fb_addr);
            write(out_line, string'(" "));
            hwrite(out_line, std_logic_vector(local_cmd_xfer_count));
            writeline(output, out_line);
            write(output, string'("dma in start"));
            DMA_In_Start(to_integer(unsigned(local_cmd_fb_addr)), to_integer(local_cmd_xfer_count));
            next_state := MC_READ;
          when MC_READ =>
            read_value := std_logic_vector(to_signed(DMA_Read_Value, mu_bus_write_value'length));
            write(out_line, string'("DMA IN value: "));
            hwrite(out_line, read_value);
            writeline(output, out_line);
            mu_bus_write_value <= read_value;
            do_inc_xfer        := true;
            do_mu_bus_strobe   := true;
            if xfer_count = local_cmd_xfer_count then
              next_state := MC_READ_CLEANUP;
            end if;
          when MC_WRITE_STARTUP =>
            write(out_line, string'("DMA OUT: "));
            hwrite(out_line, local_cmd_fb_addr);
            write(out_line, string'(" "));
            hwrite(out_line, std_logic_vector(local_cmd_xfer_count));
            writeline(output, out_line);
            write(output, string'("dma out start"));
            DMA_Out_Start(to_integer(unsigned(local_cmd_fb_addr)), to_integer(local_cmd_xfer_count));
            next_state       := MC_WRITE;
          when MC_WRITE =>
            write(out_line, string'("DMA OUT value: "));
            hwrite(out_line, mu_bus_data);
            writeline(output, out_line);
            DMA_Write_Value(to_integer(signed(mu_bus_data)));
            do_inc_xfer := true;
            do_mu_bus_strobe := true;
            if xfer_count = local_cmd_xfer_count then
              next_state := MC_CLEANUP;
            else
              next_state := MC_WRITE;
            end if;
          when MC_READ_CLEANUP =>
            do_mu_bus_strobe := true;
            next_state       := MC_CLEANUP;
          when MC_CLEANUP =>
            if local_cmd_is_write = '1' then
              write(output, string'("dma out finish"));
              DMA_Out_Finish;
            else
              write(output, string'("dma in finish"));
              DMA_In_Finish;
            end if;
            next_state := MC_IDLE;
        end case;
      end if;  -- not reset

      mc_state <= next_state;
      if do_mu_bus_strobe = true then
        mu_bus_strobe <= '1';
      else
        mu_bus_strobe <= '0';
      end if;

      if do_reset_xfer = true then
        xfer_count <= X"00" & "0001";
      else
        if do_inc_xfer = true then
          xfer_count <= xfer_count + 1;
        end if;
      end if;
      
    end if;  --clock
  end process;

end behaviour;
