-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--use std.textio.all;
--use ieee.std_logic_textio.all;

entity cpu_dma_unit is
  port(
    clock : in std_logic;
    reset : in std_logic;

    --mu data bus
    mu_bus_addr   : in    std_logic_vector(1 downto 0);
    mu_bus_data   : inout std_logic_vector(31 downto 0);
    mu_bus_write  : in    std_logic;  --if high the MCU is going to write to a mu
    mu_bus_strobe : in    std_logic;
    mu_cpu_irq    : out   std_logic;
    mu_cpu_sel    : in    std_logic;

    --connections to the data BRAM's second port
    bram_addr   : out std_logic_vector(15 downto 0);
    bram_din    : in  std_logic_vector(31 downto 0);
    bram_dout   : out std_logic_vector(31 downto 0);
    bram_we     : out std_logic;
    bram_strobe : out std_logic;

    --connections to the cpu data bus. we have control registers
    --outside the address space of the bram
    --this allows the shader cpu to read and write them
    ext_cpu_strobe : in  std_logic;
    ext_cpu_addr   : in  std_logic_vector(7 downto 0);
    ext_cpu_dout   : out std_logic_vector(31 downto 0);
    ext_cpu_din    : in  std_logic_vector(31 downto 0);
    ext_cpu_we     : in  std_logic;
    ext_cpu_busy   : out std_logic;

    --if the cpu is halted, we allow the command bus
    --to read and write the bram through our port, leaving the 2nd port to the
    --cpu alone
    cb_strobe : in  std_logic;
    cb_addr   : in  std_logic_vector(15 downto 0);
    cb_din    : in  std_logic_vector(31 downto 0);
    cb_dout   : out std_logic_vector(31 downto 0);
    cb_we     : in  std_logic
    );
end cpu_dma_unit;


architecture rtl of cpu_dma_unit is

  signal cmd_0_reg  : std_logic_vector(31 downto 0);
  signal addr_0_reg : std_logic_vector(31 downto 0);
  signal cmd_1_reg  : std_logic_vector(31 downto 0);
  signal addr_1_reg : std_logic_vector(31 downto 0);
  signal cmd_2_reg  : std_logic_vector(31 downto 0);
  signal addr_2_reg : std_logic_vector(31 downto 0);
  signal cmd_3_reg  : std_logic_vector(31 downto 0);
  signal addr_3_reg : std_logic_vector(31 downto 0);

  signal mu_bus_data_in : std_logic_vector(31 downto 0);

  --what dma xfer of the 1-4 inclusive are we currently doing
  signal c_dma_meta_count : unsigned(2 downto 0);

  signal c_dma_is_write_reg : std_logic;
  signal c_dma_fb_addr_reg  : std_logic_vector(31 downto 0);
  signal c_dma_count_reg    : std_logic_vector(11 downto 0);
  --incremented through the dma
  signal c_dma_ram_addr_reg : unsigned(15 downto 0);
  signal c_dma_bram_addr    : unsigned(15 downto 0);  --may the above plus one
  --above is compared to this during xfers.
  signal c_dma_end_addr_reg : unsigned(15 downto 0);

  --we prefetch ram values on bram read because we don't want to wait
  --the extra clock cycle 
  signal dma_read_buffer : std_logic_vector(31 downto 0);

  --a DMA read is a read from the framebuffer and writing into the BRAM
  --a DMA write is a write into the bramebuffer
  type DMA_INT_STATE is (DMA_IDLE, DMA_PRE_XFER_WAIT, DMA_XFER_SETUP, DMA_XFER_READ, DMA_XFER_READ_START,
                         DMA_XFER_WRITE, DMA_XFER_ZOMBIE, DMA_CLEANUP, DMA_EXT_RD, DMA_EXT_WR);
  signal dma_state : DMA_INT_STATE;

  signal dma_start_strobe    : std_logic;
  signal dma_meta_xfer_count : unsigned(2 downto 0);
  signal mcu_saw_irq         : std_logic;

  procedure DMA_Setup(
    variable in_cmd_data      : in  std_logic_vector(31 downto 0);
    variable in_fb_addr       : in  std_logic_vector(31 downto 0);
    signal c_dma_is_write_reg : out std_logic;
    signal c_dma_count_reg    : out std_logic_vector(11 downto 0);
    signal c_dma_fb_addr_reg  : out std_logic_vector(31 downto 0);
    signal c_dma_ram_addr_reg : out unsigned(15 downto 0);
    signal c_dma_end_addr_reg : out unsigned(15 downto 0)
    ) is
  begin
    c_dma_is_write_reg <= not in_cmd_data(31);  --note down below too
    c_dma_count_reg    <= in_cmd_data(27 downto 16);
    c_dma_fb_addr_reg  <= in_fb_addr(29 downto 0) & "00";
    c_dma_ram_addr_reg <= unsigned(in_cmd_data(15 downto 0));
    if in_cmd_data(31) = '1' then
      c_dma_end_addr_reg <= unsigned(in_cmd_data(27 downto 16)) + unsigned(in_cmd_data(15 downto 0)) - 2;
    else
      c_dma_end_addr_reg <= unsigned(in_cmd_data(27 downto 16)) + unsigned(in_cmd_data(15 downto 0));
    end if;
  end DMA_Setup;

  signal dma_active : std_logic;

  signal dma_bram_addr   : std_logic_vector(15 downto 0);
  signal dma_bram_din    : std_logic_vector(31 downto 0);
  signal dma_bram_dout   : std_logic_vector(31 downto 0);
  signal dma_bram_we     : std_logic;
  signal dma_bram_strobe : std_logic;

begin

  ext_cpu_busy <= '0' when dma_state = DMA_IDLE  else '1';
  dma_active   <= '1' when dma_state /= DMA_IDLE else '0';

  --the dma unit completly controls the bram unless it is idle.
  --the cpu should really be halted before attempting to access the dram
  --through the cb.
  bram_addr   <= dma_bram_addr   when dma_active = '1' else cb_addr;
  bram_dout   <= dma_bram_dout   when dma_active = '1' else cb_din;
  bram_we     <= dma_bram_we     when dma_active = '1' else cb_we;
  bram_strobe <= dma_bram_strobe when dma_active = '1' else cb_strobe;

  cb_dout      <= bram_din;
  dma_bram_din <= bram_din;

  dma_bram_addr <= std_logic_vector(c_dma_bram_addr);  --std_logic_vector(c_dma_ram_addr_reg);

  --ext_done <= '0' when dma_state = DMA_EXT_RD or dma_state = DMA_EXT_WR else '1';

  process(mu_bus_strobe, dma_state, c_dma_ram_addr_reg) is
  begin
    if (mu_bus_strobe = '1') and (dma_state = DMA_XFER_WRITE) then
      c_dma_bram_addr <= c_dma_ram_addr_reg + 1;
    else
      c_dma_bram_addr <= c_dma_ram_addr_reg;
    end if;
  end process;

  handle_cr_rw : process(clock) is
    variable do_dma_start_strobe : boolean;
  begin
    if rising_edge(clock) then
      do_dma_start_strobe := false;

      if ext_cpu_we = '1' and ext_cpu_addr(7 downto 4) = X"F" then
        case ext_cpu_addr(3 downto 0) is
          when X"0" =>
            cmd_0_reg <= ext_cpu_din;
          when X"1" =>
            addr_0_reg <= ext_cpu_din;
          when X"2" =>
            cmd_1_reg <= ext_cpu_din;
          when X"3" =>
            addr_1_reg <= ext_cpu_din;
          when X"4" =>
            cmd_2_reg <= ext_cpu_din;
          when X"5" =>
            addr_2_reg <= ext_cpu_din;
          when X"6" =>
            cmd_3_reg <= ext_cpu_din;
          when X"7" =>
            addr_3_reg <= ext_cpu_din;
          when X"8" =>
            if ext_cpu_din(2 downto 0) /= "000" and dma_state = DMA_IDLE then
              dma_meta_xfer_count <= unsigned(ext_cpu_din(2 downto 0));
              do_dma_start_strobe := true;
            end if;
          when others =>
            null;
        end case;
      end if;

      case ext_cpu_addr(7 downto 0) is
        when X"F0" =>
          ext_cpu_dout <= cmd_0_reg;
        when X"F1" =>
          ext_cpu_dout <= addr_0_reg;
        when X"F2" =>
          ext_cpu_dout <= cmd_1_reg;
        when X"F3" =>
          ext_cpu_dout <= addr_1_reg;
        when X"F4" =>
          ext_cpu_dout <= cmd_2_reg;
        when X"F5" =>
          ext_cpu_dout <= addr_2_reg;
        when X"F6" =>
          ext_cpu_dout <= cmd_3_reg;
        when X"F7" =>
          ext_cpu_dout <= addr_3_reg;
        when X"F8" =>
          ext_cpu_dout <= X"0000_000" & "000" & dma_active;
        when others =>
          null;
      end case;

      if do_dma_start_strobe = true then
        dma_start_strobe <= '1';
      else
        dma_start_strobe <= '0';
      end if;
    end if;
  end process;

  process(mu_cpu_sel, mu_bus_write, mu_bus_addr, c_dma_is_write_reg, c_dma_count_reg, c_dma_fb_addr_reg, dma_read_buffer) is
    variable case_input : std_logic_vector(3 downto 0);
  begin
    case_input := (mu_cpu_sel & mu_bus_write & mu_bus_addr);
    case case_input is
      when "1001" =>
        mu_bus_data <= c_dma_is_write_reg & "000" & X"000" & X"0" & c_dma_count_reg;
      when "1010" =>
        mu_bus_data <= c_dma_fb_addr_reg;
      when "1000" =>
        mu_bus_data <= dma_read_buffer;
      when others =>
        mu_bus_data <= (others => 'Z');
    end case;
  end process;

  do_dma : process(clock) is
    variable do_reset_c_dma_meta_count : boolean;
    variable do_inc_c_dma_meta_count   : boolean;
    variable cmd_reg_copy              : std_logic_vector(31 downto 0);
    variable addr_reg_copy             : std_logic_vector(31 downto 0);
    variable next_state                : DMA_INT_STATE;
    variable strobe_bram               : boolean;
    variable strobe_bram_we            : boolean;
    variable do_mu_irq                 : boolean;
    variable do_bram_addr_inc          : boolean;
    variable junk_write_value          : std_logic_vector(31 downto 0) := X"00FF_0000";
  begin
    if rising_edge(clock) then
      next_state                := dma_state;
      strobe_bram               := false;
      strobe_bram_we            := false;
      do_mu_irq                 := false;
      do_bram_addr_inc          := false;
      do_reset_c_dma_meta_count := false;
      do_inc_c_dma_meta_count   := false;

      if reset = '1' then
        next_state := DMA_IDLE;
      else
        case dma_state is
          when DMA_IDLE =>
            mcu_saw_irq <= '0';
            --dma_start_strobe is never set for zero xfers
            if dma_start_strobe = '1' then
              --we start with 1 so at the strobe detection we have the xfer number
              do_reset_c_dma_meta_count := true;
              cmd_reg_copy              := cmd_0_reg;
              addr_reg_copy             := addr_0_reg;
              next_state                := DMA_PRE_XFER_WAIT;
            end if;
          when DMA_PRE_XFER_WAIT =>
            do_mu_irq  := true;
            DMA_Setup(cmd_reg_copy, addr_reg_copy, c_dma_is_write_reg, c_dma_count_reg, c_dma_fb_addr_reg, c_dma_ram_addr_reg, c_dma_end_addr_reg);
            next_state := DMA_XFER_SETUP;
          when DMA_XFER_SETUP =>
            do_mu_irq := true;
            if mu_cpu_sel = '1' then
              --the mcu is going to passivly read our configuration registers
              --over the next few clocks. i need to get ready for the xfer to begin
              --by presenting the data to be read and reset our internal address.
              mcu_saw_irq <= '1';
              if c_dma_is_write_reg = '1' then
                next_state       := DMA_XFER_WRITE;
                --address presented to ram on previous clock
                dma_read_buffer  <= dma_bram_din;
                do_bram_addr_inc := true;
              else
                next_state := DMA_XFER_READ_START;
              end if;
            end if;
          when DMA_XFER_READ_START=>
            if mu_bus_strobe = '1' then
              next_state     := DMA_XFER_READ;
              dma_bram_dout  <= mu_bus_data;
              strobe_bram    := true;
              strobe_bram_we := true;
            end if;
          when DMA_XFER_READ =>         --reading from framebuffer
            --we hold in this state until we hit the proper count or the mu
            --drops our select
            if mu_cpu_sel = '0' then
              next_state := DMA_CLEANUP;
            elsif mu_bus_strobe = '1' then
              dma_bram_dout  <= mu_bus_data;
              strobe_bram    := true;
              strobe_bram_we := true;

              do_bram_addr_inc := true;
              if c_dma_ram_addr_reg = c_dma_end_addr_reg then
                next_state := DMA_XFER_ZOMBIE;
              end if;
            end if;
          when DMA_XFER_WRITE =>        --writing to framebuffer
            --we hold in this state until we hit the proper count or the mu
            --drops our select
            if mu_cpu_sel = '0' then
              next_state := DMA_CLEANUP;
            elsif mu_bus_strobe = '1' then
              strobe_bram      := true;
              dma_read_buffer  <= dma_bram_din;
              do_bram_addr_inc := true;
              if c_dma_ram_addr_reg = c_dma_end_addr_reg then
                next_state := DMA_XFER_ZOMBIE;
              end if;
            end if;
          when DMA_XFER_ZOMBIE =>
            --a state that should not happen (or at least for more than a clock).
            --we consider our xfer done, but the mcu
            --has not dropped our select. basically ignore strobes until select
            --line is dropped
            if mu_cpu_sel = '0' then
              next_state := DMA_CLEANUP;
            end if;
          when DMA_CLEANUP =>
            mcu_saw_irq <= '0';
            next_state  := DMA_IDLE;    --possibly override below
            case c_dma_meta_count is
              when "001" =>
                if (dma_meta_xfer_count = "010") or (dma_meta_xfer_count = "011") or (dma_meta_xfer_count = "100") then
                  cmd_reg_copy            := cmd_1_reg;
                  addr_reg_copy           := addr_1_reg;
                  next_state              := DMA_PRE_XFER_WAIT;
                  do_inc_c_dma_meta_count := true;
                end if;
              when "010" =>
                if (dma_meta_xfer_count = "011") or (dma_meta_xfer_count = "100") then
                  cmd_reg_copy            := cmd_2_reg;
                  addr_reg_copy           := addr_2_reg;
                  next_state              := DMA_PRE_XFER_WAIT;
                  do_inc_c_dma_meta_count := true;
                end if;
              when "011" =>
                if dma_meta_xfer_count = "100" then
                  cmd_reg_copy            := cmd_3_reg;
                  addr_reg_copy           := addr_3_reg;
                  next_state              := DMA_PRE_XFER_WAIT;
                  do_inc_c_dma_meta_count := true;
                end if;
              when others =>
                null;
            --write(output, string'("others"));
            end case;
          --write(output, string'("count test end"));
          when DMA_EXT_RD =>
            next_state := DMA_IDLE;
          when DMA_EXT_WR =>
            next_state := DMA_IDLE;
          when others =>
            null;
        end case;

        if do_reset_c_dma_meta_count = true then
          c_dma_meta_count <= "001";
        elsif do_inc_c_dma_meta_count = true then
          c_dma_meta_count <= c_dma_meta_count + 1;
        end if;

        if do_bram_addr_inc = true then
          c_dma_ram_addr_reg <= c_dma_ram_addr_reg +1;
        end if;

        if do_mu_irq = true then
          mu_cpu_irq <= '1';
        else
          mu_cpu_irq <= '0';
        end if;

        dma_state <= next_state;

        if strobe_bram = true then
          dma_bram_strobe <= '1';
        else
          dma_bram_strobe <= '0';
        end if;

        if strobe_bram_we = true then
          dma_bram_we <= '1';
        else
          dma_bram_we <= '0';
        end if;

      end if;  --reset
    end if;  --clock;
  end process;

end rtl;
