-- Create Date:    22:22:45 02/10/2009
-- modified 2014
-- released to the Public Domain by the author

--this is more complicated than it needs to be because I want it to work at
--high and pixel clock rates. still working on that....

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.tvc_pack.all;

entity disp_driver is
  port (
    tvc_clock : in std_logic;
    tvc_reset : in std_logic;  --this reset is probably coming from another clock domain, needs a sync.

    --we nicly sync these outputs to other tvc components to the tvc clock.
    odd_fill_req    : out std_logic;
    even_fill_req   : out std_logic;
    reset_cntrs_req : out std_logic;  --reset odd/even address counters to display top.

    --odds are, this is a much faster clock than the tvc clock. (but could be slower)
    pixel_clk   : in std_logic;
    pixel_reset : in std_logic;

    --these signals are operating in the pixel clock domain.
    next_pix_addr     : out std_logic_vector(11 downto 0);
    --these vectors should be much wider.
    odd_next_pix_val  : in  std_logic_vector(31 downto 0);  --data from the scanline cache
    even_next_pix_val : in  std_logic_vector(31 downto 0);
    --signals to "DAC" & monitor
    disp_pix_value    : out std_logic_vector(7 downto 0);
    disp_horiz_sync   : out std_logic;
    disp_vert_sync    : out std_logic
    );
end disp_driver;

architecture Behavioral of disp_driver is

  --signals operate at tvc clock domain
  signal odd_fill_seen       : std_logic;
  signal even_fill_seen      : std_logic;
  signal reset_cntrs_seen    : std_logic;
  signal odd_fill_int_req    : std_logic;
  signal even_fill_int_req   : std_logic;
  signal reset_cntrs_int_req : std_logic;

  --signals from clock domain into tvc_clock domain
  signal odd_fill_int_req_PIXCLK    : std_logic;  --internal requests in the pixel clock domain
  signal even_fill_int_req_PIXCLK   : std_logic;
  signal reset_cntrs_int_req_PIXCLK : std_logic;

  signal odd_fill_int_req_TVCCLK    : std_logic;  --internal request in the tvc clock domain
  signal even_fill_int_req_TVCCLK   : std_logic;
  signal reset_cntrs_int_req_TVCCLK : std_logic;

  --signals in the pixel clock domain.
  signal reset_ctrs_sr       : std_logic_vector(3 downto 0);
  signal even_fill_sr        : std_logic_vector(3 downto 0);
  signal odd_fill_sr         : std_logic_vector(3 downto 0);
  signal hsync_counter       : unsigned(11 downto 0);
  signal vsync_counter       : unsigned(11 downto 0);
  signal v_display_pix       : std_logic;  --in a vertical region of the display where we should be clocking out pixels
  signal h_display_pix       : std_logic;  --in a horizontal region of the display where we should be clocking out pixels
  --this must be one clock ahead of the current pixel.
  signal eo_sc_next_pix_addr : std_logic_vector(11 downto 0);

begin

  odd_fill_req    <= odd_fill_int_req;
  even_fill_req   <= even_fill_int_req;
  reset_cntrs_req <= reset_cntrs_int_req;

  next_pix_addr <= eo_sc_next_pix_addr;

  reset_cntrs_int_req_PIXCLK <= reset_ctrs_sr(3);
  odd_fill_int_req_PIXCLK    <= even_fill_sr(3);
  even_fill_int_req_PIXCLK   <= odd_fill_sr(3);

  odd_fill_sync : std_sync
    port map(sig_in => odd_fill_int_req_PIXCLK, sig_out => odd_fill_int_req_TVCCLK,
             clock  => tvc_clock, reset => tvc_reset);

  even_fill_sync : std_sync
    port map(sig_in => even_fill_int_req_PIXCLK, sig_out => even_fill_int_req_TVCCLK,
             clock  => tvc_clock, reset => tvc_reset);

  reset_cntrs_sync : std_sync
    port map(sig_in => reset_cntrs_int_req_PIXCLK, sig_out => reset_cntrs_int_req_TVCCLK,
             clock  => tvc_clock, reset => tvc_reset);

  process(tvc_clock)
  begin
    if tvc_clock = '1' and tvc_clock'event then
      if tvc_reset = '1' then
        odd_fill_int_req    <= '0';
        even_fill_int_req   <= '0';
        reset_cntrs_int_req <= '0';
        odd_fill_seen       <= '0';
        even_fill_seen      <= '0';
        reset_cntrs_seen    <= '0';
      else
        if odd_fill_int_req = '1' then  --have it high only one cycle in the tvc clock domain.
          odd_fill_int_req <= '0';
        else
          if odd_fill_int_req_TVCCLK = '1' then
            if odd_fill_seen = '0' then
              odd_fill_int_req <= '1';
              odd_fill_seen    <= '1';
            end if;
          else
            odd_fill_seen <= '0';
          end if;
        end if;

        if even_fill_int_req = '1' then  --have it high only one cycle in the tvc clock domain.
          even_fill_int_req <= '0';
        else
          if even_fill_int_req_TVCCLK = '1' then
            if even_fill_seen = '0' then
              even_fill_int_req <= '1';
              even_fill_seen    <= '1';
            end if;
          else
            even_fill_seen <= '0';
          end if;
        end if;

        if reset_cntrs_int_req = '1' then  --have it high only one cycle in the tvc clock domain.
          reset_cntrs_int_req <= '0';
        else
          if reset_cntrs_int_req_TVCCLK = '1' then
            if reset_cntrs_seen = '0' then
              reset_cntrs_int_req <= '1';
              reset_cntrs_seen    <= '1';
            end if;
          else
            reset_cntrs_seen <= '0';
          end if;
        end if;
      end if;  --reset
    end if;  --tvc_clock
  end process;

  process(pixel_clk)
    variable long_pixel_value  : std_logic_vector(31 downto 0);
    variable h_scanline_end    : boolean;
    variable v_display_end     : boolean;
    variable do_v_pulse        : boolean;
    variable do_reset_counters : boolean;
    variable do_even_fill      : boolean;
    variable do_odd_fill       : boolean;
  begin
    if pixel_clk = '1' and pixel_clk'event then
      h_scanline_end    := false;
      v_display_end     := false;
      do_reset_counters := false;
      do_even_fill      := false;
      do_odd_fill       := false;

      if pixel_reset = '1' then
        h_scanline_end := true;
        v_display_end  := true;
      --disp_horiz_sync <= '1';
      --disp_vert_sync  <= '1';
      else                              --not reset

        case hsync_counter is
          when X"000" =>
            disp_horiz_sync <= '0';
            h_display_pix   <= '0';
          when X"060" =>
            disp_horiz_sync <= '1';
          when X"070" =>
            h_display_pix <= '1';
          when X"2F0" =>
            h_display_pix <= '0';
            if v_display_pix = '1' then
              if vsync_counter(0) = '1' then
                do_odd_fill := true;
              else
                do_even_fill := true;
              end if;
            end if;
          when X"320" =>
            h_scanline_end := true;
          when others => null;
        end case;

        --many pixel clocks occur at a given vsync counter value
        --this variable ensures a pulse occurs only once per vsync cycle value
        if hsync_counter = X"00F" then  --I think could be any value...
          do_v_pulse := true;
        else
          do_v_pulse := false;
        end if;

        case vsync_counter is
          when X"000" =>                --start the vsync pulse
            disp_vert_sync <= '0';
            v_display_pix  <= '0';
            if do_v_pulse then
              do_reset_counters := true;
            end if;
          when X"002" =>                --end the vsync pulse 
            disp_vert_sync <= '1';
            if do_v_pulse then
              do_even_fill := true;     --pre-fill even scanline
            end if;
          when X"004" =>
            if do_v_pulse then
              do_odd_fill := true;      --pre-fill odd scanline
            end if;
          when X"01D" =>  --we are starting drawing the part of the screen where pixels are displayed 
            v_display_pix <= '1';
          when X"1FD" =>                --stop drawing pixels
            v_display_pix <= '0';
          when X"209" =>                --the end of the full display cycle
            v_display_end := true;
          when others =>
            null;
        end case;

        -- ************ output pixel values ****************
        long_pixel_value := X"0000_0000";
        if v_display_pix = '1' and h_display_pix = '1' then
          if vsync_counter(0) = '1' then
            long_pixel_value := odd_next_pix_val;
          else
            long_pixel_value := even_next_pix_val;
          end if;
          --old pattern generator code
--                      case eo_sc_next_pix_addr(5 downto 4) is
--                      when "00" =>    long_pixel_value := X"c0c0c0c0";
--                      when "01" =>    long_pixel_value := X"38383838";
--                      when "10" =>    long_pixel_value := X"07070707";
--                      when others =>  long_pixel_value := X"00000000";
--                      end case;

          eo_sc_next_pix_addr <= std_logic_vector(unsigned(eo_sc_next_pix_addr) + 1);
        else
          eo_sc_next_pix_addr <= (others => '0');
        end if;

        --the next pix address has not been fetched yet
        --(will be on next clock) therefore need to shift
        --pixel bytes we are outputing by one
        case eo_sc_next_pix_addr(1 downto 0) is
          when "00" =>
            --disp_pix_value <= long_pixel_value(31 downto 24);
            disp_pix_value <= long_pixel_value(7 downto 0);
          when "01" =>
            --disp_pix_value <= long_pixel_value(23 downto 16);
            disp_pix_value <= long_pixel_value(31 downto 24);
          when "10" =>
            --disp_pix_value <= long_pixel_value(15 downto 8);
            disp_pix_value <= long_pixel_value(23 downto 16);
          when "11" =>
            --disp_pix_value <= long_pixel_value(7 downto 0);
            disp_pix_value <= long_pixel_value(15 downto 8);
          when others =>
            null;
        end case;

      end if;  --reset

      if h_scanline_end = true then
        hsync_counter <= X"000";
      else
        hsync_counter <= hsync_counter + 1;
      end if;

      if v_display_end = true then
        vsync_counter <= X"000";
      elsif h_scanline_end = true then
        vsync_counter <= vsync_counter + 1;
      end if;

      if do_reset_counters = true then
        reset_ctrs_sr <= X"F";
      else
        reset_ctrs_sr <= reset_ctrs_sr(2 downto 0) & '0';
      end if;

      if do_even_fill = true then
        even_fill_sr <= X"F";
      else
        even_fill_sr <= even_fill_sr(2 downto 0) & '0';
      end if;

      if do_odd_fill = true then
        odd_fill_sr <= X"F";
      else
        odd_fill_sr <= odd_fill_sr(2 downto 0) & '0';
      end if;
    end if;  --pixel clock
  end process;


end Behavioral;

