-- Create Date:    22:36:41 08/04/2006 
-- modified 2009, 2014
-- released to the Public Domain by the author

----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Design Name: 
-- Module Name:    mcu_sched - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity mcu_sched is
  port (
    tvc_clock : in std_logic;
    tvc_reset : in std_logic;

    mcu_start : out std_logic;
    mcu_busy  : in  std_logic;

    --memory users irq/select lines
    odd_sc_irq    : in  std_logic;
    odd_sc_select : out std_logic;

    even_sc_irq    : in  std_logic;
    even_sc_select : out std_logic;

    block_reader_irq    : in  std_logic;
    block_reader_select : out std_logic;

    block_writer_irq    : in  std_logic;
    block_writer_select : out std_logic;

    rp_writer_irq    : in  std_logic;
    rp_writer_select : out std_logic;

    line_drawer_irq    : in  std_logic;
    line_drawer_select : out std_logic;

    poly_scanline_irq    : in  std_logic;
    poly_scanline_select : out std_logic
    );
end mcu_sched;

architecture Behavioral of mcu_sched is

  type MCU_SCHED_STATE is (MCU_SCHED_IDLE, MCU_SCHED_WAITSTATE, MCU_SCHED_WAITING, MCU_SCHED_END_WAITSTATE, MCU_SCHED_END_WAITSTATE2);
  signal sched_state : MCU_SCHED_STATE := MCU_SCHED_IDLE;

  subtype ID_MU is unsigned(2 downto 0);
  constant NONE_MU    : ID_MU := "000";
  constant ODD_SC_MU  : ID_MU := "001";
  constant EVEN_SC_MU : ID_MU := "010";
  constant BR_MU      : ID_MU := "011";
  constant BW_MU      : ID_MU := "100";
  constant RPW_MU     : ID_MU := "101";
  constant LD_MU      : ID_MU := "110";
  constant PSL_MU     : ID_MU := "111";

  --0 is none 1 is the odd_sc 2 is the even_sc, 3 is the block_reader 4 is the block_writer
  --5 is the rp_writer 6 is the line drawer 7 is the poly_scanline
  signal active_mu : ID_MU := NONE_MU;

  --we service mus in a round-robbin fashion with the exception of the pixel-pusher(s)' scanline caches that have priority
  --this should not ever be 0 or 1, as these are the scanline caches
  --this is the next memory user that should get priority if it and another memory user wants a transfer.
  signal next_mu_to_check : ID_MU := BR_MU;
  
begin

  process(tvc_clock)
    variable new_active_mu : ID_MU;
  begin
    if tvc_clock = '1' and tvc_clock'event then
      new_active_mu := NONE_MU;

      if tvc_reset = '1' then
        sched_state      <= MCU_SCHED_IDLE;
        active_mu        <= NONE_MU;
        next_mu_to_check <= BR_MU;

        mcu_start <= '0';

        odd_sc_select        <= '0';
        even_sc_select       <= '0';
        block_reader_select  <= '0';
        block_writer_select  <= '0';
        rp_writer_select     <= '0';
        line_drawer_select   <= '0';
        poly_scanline_select <= '0';
      else                              --not reset
        case sched_state is
          when MCU_SCHED_IDLE =>
            if odd_sc_irq = '1' then
              new_active_mu := ODD_SC_MU;
            elsif even_sc_irq = '1' then
              new_active_mu := EVEN_SC_MU;
            else
              case next_mu_to_check is
                when BR_MU =>
                  if block_reader_irq = '1' then
                    new_active_mu := BR_MU;
                  elsif block_writer_irq = '1' then
                    new_active_mu := BW_MU;
                  elsif rp_writer_irq = '1' then
                    new_active_mu := RPW_MU;
                  elsif line_drawer_irq = '1' then
                    new_active_mu := LD_MU;
                  elsif poly_scanline_irq = '1' then
                    new_active_mu := PSL_MU;
                  end if;
                when BW_MU =>
                  if block_writer_irq = '1' then
                    new_active_mu := BW_MU;
                  elsif rp_writer_irq = '1' then
                    new_active_mu := RPW_MU;
                  elsif line_drawer_irq = '1' then
                    new_active_mu := LD_MU;
                  elsif poly_scanline_irq = '1' then
                    new_active_mu := PSL_MU;
                  elsif block_reader_irq = '1' then
                    new_active_mu := BR_MU;
                  end if;
                when RPW_MU =>
                  if rp_writer_irq = '1' then
                    new_active_mu := RPW_MU;
                  elsif line_drawer_irq = '1' then
                    new_active_mu := LD_MU;
                  elsif poly_scanline_irq = '1' then
                    new_active_mu := PSL_MU;
                  elsif block_reader_irq = '1' then
                    new_active_mu := BR_MU;
                  elsif block_writer_irq = '1' then
                    new_active_mu := BW_MU;
                  end if;
                when LD_MU =>
                  if line_drawer_irq = '1' then
                    new_active_mu := LD_MU;
                  elsif poly_scanline_irq = '1' then
                    new_active_mu := PSL_MU;
                  elsif block_reader_irq = '1' then
                    new_active_mu := BR_MU;
                  elsif block_writer_irq = '1' then
                    new_active_mu := BW_MU;
                  elsif rp_writer_irq = '1' then
                    new_active_mu := RPW_MU;
                  end if;
                when PSL_MU =>
                  if poly_scanline_irq = '1' then
                    new_active_mu := PSL_MU;
                  elsif block_reader_irq = '1' then
                    new_active_mu := BR_MU;
                  elsif block_writer_irq = '1' then
                    new_active_mu := BW_MU;
                  elsif rp_writer_irq = '1' then
                    new_active_mu := RPW_MU;
                  elsif line_drawer_irq = '1' then
                    new_active_mu := LD_MU;
                  end if;
                when others =>          -- a bug
                  next_mu_to_check <= BR_MU;
              end case;
            end if;

            if new_active_mu /= NONE_MU then
              active_mu <= new_active_mu;
              case new_active_mu is
                when ODD_SC_MU =>
                  odd_sc_select <= '1';
                when EVEN_SC_MU =>
                  even_sc_select <= '1';
                when BR_MU =>
                  block_reader_select <= '1';
                  next_mu_to_check    <= BW_MU;
                when BW_MU =>
                  block_writer_select <= '1';
                  next_mu_to_check    <= RPW_MU;
                when RPW_MU =>
                  rp_writer_select <= '1';
                  next_mu_to_check <= LD_MU;
                when LD_MU =>
                  line_drawer_select <= '1';
                  next_mu_to_check   <= PSL_MU;
                when PSL_MU =>
                  poly_scanline_select <= '1';
                  next_mu_to_check     <= BR_MU;
                when others =>
                  next_mu_to_check <= BR_MU;  --a bug;
              end case;

              sched_state  <= MCU_SCHED_WAITSTATE;
              mcu_start    <= '1';
            end if;
            
          when MCU_SCHED_WAITSTATE =>
            sched_state  <= MCU_SCHED_WAITING;
            mcu_start <= '1';
          when MCU_SCHED_WAITING =>
            mcu_start <= '0';
            if mcu_busy = '0' then
              odd_sc_select        <= '0';
              even_sc_select       <= '0';
              block_reader_select  <= '0';
              block_writer_select  <= '0';
              rp_writer_select     <= '0';
              line_drawer_select   <= '0';
              poly_scanline_select <= '0';
              active_mu            <= NONE_MU;
              sched_state          <= MCU_SCHED_END_WAITSTATE;
            end if;
          when MCU_SCHED_END_WAITSTATE =>
            sched_state <= MCU_SCHED_END_WAITSTATE2;
          when MCU_SCHED_END_WAITSTATE2 =>
            sched_state <= MCU_SCHED_IDLE;
        end case;
      end if;
    end if;
  end process;
end Behavioral;

