-- Create Date:    22:59:02 08/04/2006 
-- modified 2009,2014
-- released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.tvc_pack.all;

-- the memory controller unit (mcu) manages the dram. Because the chip will perpetually
-- be bandwidth starved, it will basiclly set the timing on everything.
-- memory users (mu) will assert that they need reading/writing done on their behalf, and the mcu
-- will preform the desired operations when it can.  the data for reading/writing will be held
-- in the 'calling' mu.  the calling mu's will share an access buss to the mcu.

entity mcu_struct is
  port (
    nx_ram_addr  : out   std_logic_vector(22 downto 0);
    nx_ram_clk   : out   std_logic;
    nx_ram_adv_n : out   std_logic;
    nx_ram_cre   : out   std_logic;
    nx_ram_ce_n  : out   std_logic;
    nx_ram_oe_n  : out   std_logic;
    nx_ram_we_n  : out   std_logic;
    nx_ram_lb_n  : out   std_logic;
    nx_ram_ub_n  : out   std_logic;
    nx_ram_wait  : in    std_logic;
    nx_ram_dq    : inout std_logic_vector(15 downto 0);

    tvc_clock       : in  std_logic;
    tvc_reset       : in  std_logic;
    ram_clock       : in  std_logic;
    ram_clock_adv   : in  std_logic;
    ram_start_delay : out std_logic;

    --the mcu's locic interface to the rest of the tvc           
    --memory user access buss 
    mu_bus_address : out   std_logic_vector(1 downto 0);
    mu_bus_data    : inout std_logic_vector(31 downto 0);
    mu_bus_write   : out   std_logic;  --if high the MCU is going to write to a mu
    --if high the MCU expects the selected MU to present its data on the mu_bus_data or to
    --copy data from the mu_bus_data into its internal cache/registers.
    mu_bus_strobe  : out   std_logic;

    odd_sc_irq           : in  std_logic;
    odd_sc_select        : out std_logic;
    even_sc_irq          : in  std_logic;
    even_sc_select       : out std_logic;
    block_reader_irq     : in  std_logic;
    block_reader_select  : out std_logic;
    block_writer_irq     : in  std_logic;
    block_writer_select  : out std_logic;
    rp_writer_irq        : in  std_logic;
    rp_writer_select     : out std_logic;
    line_drawer_irq      : in  std_logic;
    line_drawer_select   : out std_logic;
    poly_scanline_irq    : in  std_logic;
    poly_scanline_select : out std_logic
    );
end mcu_struct;

architecture rtl of mcu_struct is

  signal sched_start : std_logic;
  signal sched_busy  : std_logic;

  signal mcu_ll_busy : std_logic;

  signal blk_xfer_is_write : std_logic;  --xfer is writing to framebuffer.
  signal blk_rd_req        : std_logic;
  signal blk_wr_req        : std_logic;
  signal blk_xfer_addr     : std_logic_vector(23 downto 2);
  signal blk_xfer_count    : std_logic_vector(11 downto 0);
  signal blk_dout          : std_logic_vector(31 downto 0);
  signal blk_din           : std_logic_vector(31 downto 0);
  signal blk_strobe        : std_logic;

  --not used right now. may be gutted from nexus2_ram_interface
  signal dword_strobe   : std_logic;
  signal dword_addr     : std_logic_vector(23 downto 2);
  signal dword_data_out : std_logic_vector(31 downto 0);
  signal dword_data_in  : std_logic_vector(31 downto 0);
  signal dword_byte_we  : std_logic_vector(3 downto 0);

  type MCU_INT_STATE is (MCU_IDLE, MCU_RD_CFG_1, MCU_RD_CFG_2, MCU_LL_WORKING);
  signal mcu_state : MCU_INT_STATE := MCU_IDLE;

begin

  sched_busy <= '1' when mcu_state /= MCU_IDLE else '0';

  dword_strobe  <= '0';
  dword_addr    <= (others => '0');
  dword_data_in <= (others => '0');
  dword_byte_we <= (others => '0');

  mu_bus_data <= blk_dout when (mcu_state = MCU_LL_WORKING) and (blk_xfer_is_write = '0') else (others => 'Z');
  blk_din     <= mu_bus_data;

  mu_bus_write  <= '0' when mcu_state /= MCU_LL_WORKING else (not blk_xfer_is_write);
  mu_bus_strobe <= blk_strobe;          --we never have to strobe locally

  sched : mcu_sched
    port map(tvc_clock         => tvc_clock, tvc_reset => tvc_reset,
             mcu_start         => sched_start, mcu_busy => sched_busy,
             odd_sc_irq        => odd_sc_irq, odd_sc_select => odd_sc_select, even_sc_irq => even_sc_irq, even_sc_select => even_sc_select,
             block_reader_irq  => block_reader_irq, block_reader_select => block_reader_select,
             block_writer_irq  => block_writer_irq, block_writer_select => block_writer_select,
             rp_writer_irq     => rp_writer_irq, rp_writer_select => rp_writer_select,
             line_drawer_irq   => line_drawer_irq, line_drawer_select => line_drawer_select,
             poly_scanline_irq => poly_scanline_irq, poly_scanline_select => poly_scanline_select);

  mccu_ll : nexus2_ram_interface
    generic map (startup_delay_time => X"000F", start_clear_addr => "111" & X"F_F000")
    port map(busy          => mcu_ll_busy, startup_delay => ram_start_delay,
             blk_rd_req    => blk_rd_req, blk_wr_req => blk_wr_req, blk_xfer_addr => blk_xfer_addr, blk_xfer_des_count => blk_xfer_count,
             blk_dout      => blk_dout, blk_din => blk_din, blk_strobe => blk_strobe,
             dword_strobe  => dword_strobe, dword_addr => dword_addr, dword_data_out => dword_data_out,
             dword_data_in => dword_data_in, dword_byte_we => dword_byte_we,
             nx_ram_addr   => nx_ram_addr, nx_ram_clk => nx_ram_clk, nx_ram_adv_n => nx_ram_adv_n, nx_ram_cre => nx_ram_cre,
             nx_ram_ce_n   => nx_ram_ce_n, nx_ram_oe_n => nx_ram_oe_n, nx_ram_we_n => nx_ram_we_n, nx_ram_lb_n => nx_ram_lb_n,
             nx_ram_ub_n   => nx_ram_ub_n, nx_ram_wait => nx_ram_wait, nx_ram_dq => nx_ram_dq,
             clock_ram     => ram_clock, clock_ram_adv => ram_clock_adv, reset => tvc_reset);

  process(tvc_clock) is
    variable next_state   : MCU_INT_STATE;
    variable do_read_req  : boolean;
    variable do_write_req : boolean;
  begin
    if rising_edge(tvc_clock) then
      do_read_req  := false;
      do_write_req := false;

      next_state := mcu_state;
      if tvc_reset = '1' then
        next_state := MCU_IDLE;
      else
        case mcu_state is
          when MCU_IDLE=>
            if sched_start = '1' then
              next_state     := MCU_RD_CFG_1;
              mu_bus_address <= "01";
            else
              mu_bus_address <= "00";
            end if;
          when MCU_RD_CFG_1=>
            blk_xfer_is_write <= mu_bus_data(31);
            blk_xfer_count    <= mu_bus_data(11 downto 0);
            mu_bus_address    <= "10";
            next_state        := MCU_RD_CFG_2;
          when MCU_RD_CFG_2=>
            blk_xfer_addr  <= mu_bus_data(23 downto 2);
            mu_bus_address <= "00";
            next_state     := MCU_LL_WORKING;

            if blk_xfer_is_write = '1' then
              do_write_req := true;
            else
              do_read_req := true;
            end if;
          when MCU_LL_WORKING=>
            mu_bus_address <= "00";
            if mcu_ll_busy = '0' then
              next_state := MCU_IDLE;
            end if;
        end case;
      end if;  --reset

      if do_read_req = true then
        blk_rd_req <= '1';
      else
        blk_rd_req <= '0';
      end if;

      if do_write_req = true then
        blk_wr_req <= '1';
      else
        blk_wr_req <= '0';
      end if;

      mcu_state <= next_state;
    end if;  --clock

  end process;
  
  
end rtl;

