-- Create Date:    15:37:29 08/13/2011
-- modified 2014
-- released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

use work.tvc_pack.all;

--page13 of datasheet.
-- Fixed-length bursts consist of four, eight, sixteen, or thirty-two words. Continuous
--bursts have the ability to start at a specified address and burst to the end of the 128-word row.


entity nexus2_ram_interface is
  --here to speed simulation. if not added takes too long to simulate
  generic (
    startup_delay_time : std_logic_vector(15 downto 0) := X"FFFF";
    start_clear_addr   : std_logic_vector(22 downto 0) := "000" & X"0_0000");
  port (
    busy          : out std_logic;
    startup_delay : out std_logic;

    blk_rd_req         : in  std_logic;
    blk_wr_req         : in  std_logic;
    blk_xfer_addr      : in  std_logic_vector(23 downto 2);
    blk_xfer_des_count : in  std_logic_vector(11 downto 0);
    blk_dout           : out std_logic_vector(31 downto 0);
    blk_din            : in  std_logic_vector(31 downto 0);
    --external connection should either consume or produce new data
    blk_strobe         : out std_logic;

    --dword read/write interface; will also be used for generic reads and writes.
    --bottom bit is ignored in sync mode
    dword_strobe   : in  std_logic;
    dword_addr     : in  std_logic_vector(23 downto 2);
    dword_data_out : out std_logic_vector(31 downto 0);
    dword_data_in  : in  std_logic_vector(31 downto 0);
    dword_byte_we  : in  std_logic_vector(3 downto 0);

    --interface to the ram chip
    nx_ram_addr  : out std_logic_vector(22 downto 0);
    nx_ram_clk   : out std_logic;
    nx_ram_adv_n : out std_logic;       --address valid active low
    nx_ram_cre   : out std_logic;       --control register enable
    nx_ram_ce_n  : out std_logic;       --chip enable active low
    nx_ram_oe_n  : out std_logic;       --output enable active low
    nx_ram_we_n  : out std_logic;       --write enable active low
    nx_ram_lb_n  : out std_logic;       --lower byte enable active low
    nx_ram_ub_n  : out std_logic;       --upper byte enable active low
    nx_ram_wait  : in  std_logic;       --wait input

    nx_ram_dq : inout std_logic_vector(15 downto 0);

    clock_ram     : in std_logic;
    clock_ram_adv : in std_logic;
    reset         : in std_logic);
end nexus2_ram_interface;

architecture Behavioral of nexus2_ram_interface is
  signal not_clock_ram_adv  : std_logic;
  signal nx_ram_addr_reg    : std_logic_vector(22 downto 0);  --so it can be written out by a different clock.
  signal n2_clk_out_to_obuf : std_logic;

  type R_STATE_TYPE is (R_DELAY, R_CONFIG_1, R_CONFIG_2,
                        R_IDLE, R_SYNC_RW_STARTUP, R_SYNC_READ_DLY, R_SYNC_READ_WAIT, R_SYNC_READ_2, R_SYNC_READ_POST,
                        R_SYNC_WRITE_DLY, R_SYNC_WRITE_WAIT, R_SYNC_WRITE_2,
                        R_BR_STARTUP, R_BR_DLY, R_BR_WAIT, R_BR_READ_ODD, R_BR_READ_EVEN, R_BR_COMPLETE_1, R_BR_COMPLETE_2,
                        R_BW_STARTUP, R_BW_DLY, R_BW_WAIT, R_BW_WRITE_EVEN, R_BW_WRITE_ODD, R_BW_COMPLETE_1, R_BW_COMPLETE_2,
                        R_PRE_READ, R_ASYNC_READ, R_PRE_WRITE, R_ASYNC_WRITE, R_ASYNC_COMPLETE);
  signal r_state             : R_STATE_TYPE := R_IDLE;
  signal after_delay_state   : R_STATE_TYPE := R_IDLE;
  --signal after_async_state : R_STATE_TYPE := R_IDLE; --after an async read or write, return to this state
  constant after_async_state : R_STATE_TYPE := R_CONFIG_2;

  signal delay_ctr  : std_logic_vector(15 downto 0) := X"0000";
  signal delay_time : std_logic_vector(15 downto 0) := X"0000";

  signal prev_dword_addr    : std_logic_vector(23 downto 2);
  signal prev_dword_data_in : std_logic_vector(31 downto 0);
  signal prev_dword_byte_we : std_logic_vector(3 downto 0);

  signal write_is_cre       : std_logic;  --transparent latch (sort of).
  signal dq_ts              : std_logic;  --when goes low, data is presented to ram.
  signal dq_ts_vec          : std_logic_vector(15 downto 0);
  signal dram_read_port_raw : std_logic_vector(15 downto 0);
  signal dram_write_port    : std_logic_vector(15 downto 0);
  signal n2ram_clock_en     : std_logic;
  signal clear_array_addr   : std_logic_vector(22 downto 0);  --count through every address
  signal blk_fill_count     : std_logic_vector(11 downto 0) := (others => '0');
  signal data_out_reg       : std_logic_vector(31 downto 0) := X"0000_0000";

  signal dram_read_reg : std_logic_vector(15 downto 0);  --packed into iob

  signal inc_bread_addr_reg : std_logic := '0';

  signal blk_din_reg : std_logic_vector(31 downto 0);

  attribute iob                         : string;
  attribute equivalent_register_removal : string;

  --nx_ram_clk is generated through an ODDR2 and needs no other special treatment

  --the following cannot be set to FORCE (as I want) because the ODDR2 requires two clock signals -> clock_ram_adv and NOT clock_ram_adv
  --one of the address lines unfortunately shares resources with the nx_ram_clk which is driven by clock_ram
  --due to routing constraints, three clocks cannot be shared between the two pins so
  --we output the address a clock early, to be sure it is held long enough before the clock arrives
  --at the ram chip.
  attribute iob of nx_ram_addr : signal is "FALSE";

  attribute iob of nx_ram_adv_n : signal is "FORCE";
  attribute iob of nx_ram_cre   : signal is "FORCE";
  attribute iob of nx_ram_ce_n  : signal is "FORCE";
  attribute iob of nx_ram_oe_n  : signal is "FORCE";
  attribute iob of nx_ram_we_n  : signal is "FORCE";
  attribute iob of nx_ram_lb_n  : signal is "FORCE";
  attribute iob of nx_ram_ub_n  : signal is "FORCE";
  attribute iob of nx_ram_wait  : signal is "FORCE";

  attribute equivalent_register_removal of nx_ram_lb_n : signal is "no";
  attribute equivalent_register_removal of nx_ram_ub_n : signal is "no";

  --all the lines related to nx_ram_dq
  attribute iob of dram_write_port : signal is "FORCE";
  attribute iob of dram_read_reg   : signal is "TRUE";
  attribute iob of dq_ts_vec       : signal is "TRUE";

--KEEP COMMENTED OUT.
--attribute equivalent_register_removal of dq_ts : signal is "no";
begin
  
  busy <= '1' when (r_state /= R_IDLE) or (dword_strobe = '1') or (blk_rd_req = '1') or (blk_wr_req = '1') else '0';

  not_clock_ram_adv <= not clock_ram_adv;

  n2_ram_clk_OBUF_inst : OBUF
    port map (O => nx_ram_clk, I => n2_clk_out_to_obuf);
  
  ODDR2_inst : ODDR2
    generic map(
      DDR_ALIGNMENT => "NONE",  -- Sets output alignment to "NONE", "C0", "C1" 
      INIT          => '0',  -- Sets initial state of the Q output to '0' or '1'
      SRTYPE        => "SYNC")  -- Specifies "SYNC" or "ASYNC" set/reset
    port map (
      Q  => n2_clk_out_to_obuf,         -- 1-bit output data
      C0 => clock_ram_adv,              -- 1-bit clock input
      C1 => not_clock_ram_adv,          -- 1-bit clock input
      CE => n2ram_clock_en,             -- 1-bit clock enable input
      D0 => '1',             -- 1-bit data input (associated with C0)
      D1 => '0',             -- 1-bit data input (associated with C1)
      R  => '0',                        -- 1-bit reset input
      S  => '0'                         -- 1-bit set input
      );

--this is very important; but is set in the ucf file
--  PULLUP_inst : PULLUP
--   port map (
--      O => nx_ram_wait   -- Pullup on wait
--   );
  
  blk_dout <= data_out_reg;

  dword_data_out <= data_out_reg;

  --dram bus io
  data_o_buff : for i in 0 to 15 generate
  --for all : attribute iob of TS_FF_inst                                       : entity is "TRUE";
  --for all : attribute equivalent_register_removal of TS_FF_inst : entity is "no";
  begin
    OBUFT_inst : OBUFT
      port map (O => nx_ram_dq(i), I => dram_write_port(i), T => dq_ts_vec(i));

    --manually setting up the tristate flip flop  helps iob merging
    TS_FF_inst : FDRSE
      generic map (
        INIT => '1')         -- Initial value of register ('0' or '1')  
      port map (
        Q  => dq_ts_vec(i),             -- Data output
        C  => clock_ram,                -- Clock input
        CE => '1',                      -- Clock enable input
        D  => dq_ts,                    -- Data input
        R  => reset,                    -- Synchronous reset input
        S  => '0'                       -- Synchronous set input
        );

    IBUF_inst : IBUF
      port map (
        O => dram_read_port_raw(i),
        I => nx_ram_dq(i)
        );
  end generate;

  nx_ram_addr <= nx_ram_addr_reg;

  p_bus_main : process(clock_ram) is
    variable next_state           : R_STATE_TYPE;
    variable reset_delay          : boolean;
    variable assert_cre           : boolean;
    variable assert_ce_n          : boolean;
    variable assert_oe_n          : boolean;
    variable assert_we_n          : boolean;
    variable assert_adv_n         : boolean;
    variable assert_lb_n          : boolean;
    variable assert_ub_n          : boolean;
    variable force_dq_ts          : boolean;
    variable ram_clock_on         : boolean;
    variable inc_blk_xfer_count : boolean;
    variable clear_blk_xfer_count : boolean;
    variable assert_blk_strobe    : boolean;
    variable dram_write_port_val  : std_logic_vector(15 downto 0);  --necessary to allow iob packing (no feedback on ff's allowed)
    variable capture_blk_din      : boolean;
  begin
    if clock_ram = '1' and clock_ram'event then
      reset_delay          := false;
      assert_cre           := false;
      assert_adv_n         := false;
      assert_ce_n          := false;
      assert_oe_n          := false;
      assert_we_n          := false;
      assert_lb_n          := true;
      assert_ub_n          := true;
      force_dq_ts          := true;
      ram_clock_on         := false;
      inc_blk_xfer_count   := false;
      clear_blk_xfer_count := false;
      assert_blk_strobe    := false;
      dram_write_port_val  := X"0000";
      capture_blk_din      := false;

      if reset = '1' then
        next_state        := R_DELAY;
        reset_delay       := true;
        after_delay_state <= R_CONFIG_1;
        delay_time        <= startup_delay_time;
        startup_delay     <= '1';
      else
        next_state := r_state;

        case r_state is
          when R_DELAY=>
            if delay_ctr = delay_time then
              next_state := after_delay_state;
            end if;
          when R_CONFIG_1 =>
            --write the bcr for syncronous mode and continuous burst (right now) and wait during delay.
            --changed to active low polarity.
            nx_ram_addr_reg  <= "000" & "10" & "00" & "0" & "0" &"011" & "0" & "0" & "0" & "0" & "0" & "01" & "1" & "111";
            --changed to full drive strength
            --nx_ram_addr_reg <= "000" & "10" & "00" & "0" & "0" &"011" & "0" & "0" & "0" & "0" & "0" & "00" & "1" & "111";
            write_is_cre     <= '1';
            clear_array_addr <= start_clear_addr;
            next_state       := R_PRE_WRITE;
            --after_async_state <= R_CONFIG_2;
            assert_adv_n     := true;
            reset_delay      := true;
          when R_CONFIG_2 =>            --launch/complete a full array read
            if clear_array_addr = "111" & X"F_FFFF" then
              next_state := R_IDLE;
            else                        --write the next address
              next_state   := R_PRE_READ;
              --after_async_state <= R_CONFIG_2;
              assert_adv_n := true;
              assert_ce_n  := true;
              reset_delay  := true;
              assert_oe_n  := true;
            end if;
            nx_ram_addr_reg  <= clear_array_addr;  --no matter what this is safe.
            clear_array_addr <= std_logic_vector(unsigned(clear_array_addr) + 1);  --no matter what this is safe
          when R_IDLE =>
            startup_delay <= '0';
--            if dword_strobe = '1' and dword_addr(30) = '1' then --async 16 bit read/write.
--               nx_ram_addr_reg <= dword_addr(23 downto 1);
--               assert_adv_n := true; --address is latched on rising edge of adv_n
--               assert_ce_n := true;
--               reset_delay := true;
--               dram_write_port <= dword_data_in(15 downto 0); this is now wrong must use var now. --whether we are reading or writing 
--               after_async_state <= R_IDLE; --come back to idle.
--               if dword_addr(31) = '1' then
--                  next_state := R_PRE_WRITE;
--               else
--                  next_state := R_PRE_READ;
--                  assert_oe_n := true;
--               end if;
--            els
            --the check for zero: good idea?
            if dword_strobe = '1' then  --sync 32 bit read/write
              nx_ram_addr_reg <= dword_addr(23 downto 2) & '0';
              ram_clock_on    := true;
              next_state      := R_SYNC_RW_STARTUP;
            elsif blk_rd_req = '1' and blk_xfer_des_count /= X"000" then  --basically the same as a sync read.
              nx_ram_addr_reg <= blk_xfer_addr & '0';
              ram_clock_on    := true;
              next_state      := R_BR_STARTUP;
            elsif blk_wr_req = '1' and blk_xfer_des_count /= X"000" then
              nx_ram_addr_reg <= blk_xfer_addr & '0';
              ram_clock_on    := true;
              next_state      := R_BW_STARTUP;
            end if;
          when R_BR_STARTUP =>
            clear_blk_xfer_count := true;
            assert_ce_n          := true;
            assert_adv_n         := true;
            ram_clock_on         := true;
            reset_delay          := true;
            next_state           := R_BR_DLY;
            assert_oe_n          := true;
          when R_BR_DLY =>              --don't look at wait yet.
            assert_ce_n        := true;
            assert_oe_n        := true;
            ram_clock_on       := true;
            next_state         := R_BR_WAIT;
            inc_blk_xfer_count := true;
          when R_BR_WAIT =>
            assert_ce_n  := true;
            assert_oe_n  := true;
            ram_clock_on := true;
            if nx_ram_wait = '1' then
              --kinda same state as read even
              next_state    := R_BR_READ_ODD;
              dram_read_reg <= dram_read_port_raw;
            --data_out_reg(15 downto 0) <= dram_read_port_raw;
            elsif delay_ctr = X"01FF" then
              data_out_reg <= X"F00F1FF1";
              next_state   := R_IDLE;
            end if;
          when R_BR_READ_ODD =>
            data_out_reg(31 downto 16) <= dram_read_reg;  --move data from iob to internal register
            dram_read_reg              <= dram_read_port_raw;  --get 2nd dword into iob
            inc_blk_xfer_count         := true;
            ram_clock_on               := true;
            assert_ce_n                := true;
            assert_oe_n                := true;
            if blk_fill_count = blk_xfer_des_count then  --we are done -> note we 'stop' early due to late increment
              next_state := R_BR_COMPLETE_1;
            elsif nx_ram_wait = '0' then
              next_state  := R_BR_WAIT;
              reset_delay := true;
            else
              next_state := R_BR_READ_EVEN;
            end if;
          when R_BR_READ_EVEN =>
            reset_delay               := true;
            assert_ce_n               := true;
            assert_oe_n               := true;
            ram_clock_on              := true;
            assert_blk_strobe         := true;  --latch the 32 bit word into the cache on next rising edge.
            next_state                := R_BR_READ_ODD;
            data_out_reg(15 downto 0) <= dram_read_reg;
            dram_read_reg             <= dram_read_port_raw;
          when R_BR_COMPLETE_1 =>     --move last data from iob
            ram_clock_on              := true;
            data_out_reg(15 downto 0) <= dram_read_reg;
            next_state                := R_BR_COMPLETE_2;
          when R_BR_COMPLETE_2 =>
            assert_blk_strobe := true;  --latch final dword out
            next_state        := R_IDLE;
          when R_BW_STARTUP =>
            clear_blk_xfer_count := true;
            assert_ce_n          := true;
            assert_adv_n         := true;
            ram_clock_on         := true;
            reset_delay          := true;
            assert_we_n          := true;
            next_state           := R_BW_DLY;
          when R_BW_DLY =>
            assert_ce_n         := true;
            ram_clock_on        := true;
            assert_we_n         := true;
            capture_blk_din     := true;
            inc_blk_xfer_count := true;
            next_state          := R_BW_WAIT;
          when R_BW_WAIT =>
            --kinda equivilent to write even
            assert_ce_n         := true;
            ram_clock_on        := true;
            dram_write_port_val := blk_din_reg(31 downto 16);
            force_dq_ts         := false;     --present data to chip
            if nx_ram_wait = '1' then
              --kinda equivalent to write odd
              dram_write_port_val := blk_din_reg(15 downto 0);
              capture_blk_din     := true;
              inc_blk_xfer_count := true;
              if blk_fill_count = blk_xfer_des_count then
                next_state := R_BW_COMPLETE_1;
              else
                next_state := R_BW_WRITE_EVEN;
              end if;
            elsif delay_ctr = X"01FF" then
              data_out_reg <= X"F00F1FF1";
              next_state   := R_IDLE;
            end if;
          when R_BW_WRITE_EVEN =>
            dram_write_port_val := blk_din_reg(31 downto 16);
            ram_clock_on        := true;
            assert_ce_n         := true;
            force_dq_ts         := false;     --present data to chip
            next_state          := R_BW_WRITE_ODD;
          when R_BW_WRITE_ODD =>
            dram_write_port_val := blk_din_reg(15 downto 0);
            ram_clock_on        := true;
            assert_ce_n         := true;
            force_dq_ts         := false;     --present data to chip
            capture_blk_din     := true;
            inc_blk_xfer_count := true;
            if blk_fill_count = blk_xfer_des_count then
              next_state := R_BW_COMPLETE_1;  --chip will latch last byte here.
            elsif nx_ram_wait = '0' then
              next_state  := R_BW_WAIT;
              reset_delay := true;
            else
              next_state := R_BW_WRITE_EVEN;
            end if;
          when R_BW_COMPLETE_1 =>
            --kinda equivalent to write odd
            --writing last word.
            force_dq_ts  := false;      --present/hold data to chip
            ram_clock_on := true;
            next_state   := R_BW_COMPLETE_2;
          when R_BW_COMPLETE_2 =>
            ram_clock_on := true;
            next_state   := R_IDLE;
          when R_SYNC_RW_STARTUP =>
            assert_ce_n  := true;
            assert_adv_n := true;
            ram_clock_on := true;
            reset_delay  := true;
            if dword_byte_we /= "0000" then
              assert_we_n := true;
              force_dq_ts := false;
              if dword_byte_we(0) = '0' then
                assert_lb_n := false;
              end if;
              if dword_byte_we(1) = '0' then
                assert_ub_n := false;
              end if;
              dram_write_port_val := dword_data_in(15 downto 0);
              next_state          := R_SYNC_WRITE_DLY;
            else
              next_state  := R_SYNC_READ_DLY;
              assert_oe_n := true;
            end if;
          when R_SYNC_READ_DLY =>       --don't look at wait yet.
            assert_ce_n  := true;
            assert_oe_n  := true;
            ram_clock_on := true;
            next_state   := R_SYNC_READ_WAIT;
          when R_SYNC_READ_WAIT =>
            assert_ce_n  := true;
            assert_oe_n  := true;
            ram_clock_on := true;
            if delay_ctr = X"01FF" then
              data_out_reg <= X"F00F1FF1";
              next_state   := R_IDLE;
            end if;
            if nx_ram_wait = '1' then
              next_state    := R_SYNC_READ_2;
              dram_read_reg <= dram_read_port_raw;
            end if;
          when R_SYNC_READ_2 =>
            --ram_clock_on := true;
            next_state                := R_SYNC_READ_POST;
            data_out_reg(15 downto 0) <= dram_read_reg;
            dram_read_reg             <= dram_read_port_raw;
          when R_SYNC_READ_POST =>
            next_state                 := R_IDLE;
            data_out_reg(31 downto 16) <= dram_read_reg;
          when R_SYNC_WRITE_DLY =>      --don't look at wait yet.
            assert_ce_n  := true;
            ram_clock_on := true;
            force_dq_ts  := false;
            if dword_byte_we(0) = '0' then
              assert_lb_n := false;
            end if;
            if dword_byte_we(1) = '0' then
              assert_ub_n := false;
            end if;
            dram_write_port_val := dword_data_in(15 downto 0);
            next_state          := R_SYNC_WRITE_WAIT;
          when R_SYNC_WRITE_WAIT =>
            assert_ce_n  := true;
            ram_clock_on := true;
            force_dq_ts  := false;
            if nx_ram_wait = '1' then
              next_state := R_SYNC_WRITE_2;
              if dword_byte_we(2) = '0' then
                assert_lb_n := false;
              end if;
              if dword_byte_we(3) = '0' then
                assert_ub_n := false;
              end if;
              dram_write_port_val := dword_data_in(31 downto 16);
            else
              dram_write_port_val := dword_data_in(15 downto 0);
              if dword_byte_we(0) = '0' then
                assert_lb_n := false;
              end if;
              if dword_byte_we(1) = '0' then
                assert_ub_n := false;
              end if;
              if delay_ctr = X"01FF" then
                data_out_reg <= X"F00F0FF0";
                next_state   := R_IDLE;
              end if;
            end if;
          when R_SYNC_WRITE_2 =>
            --dram_write_port_val := dword_data_in(31 downto 16); no force_dq_ts, so presenting data does not matter
            ram_clock_on := true;
            next_state   := R_IDLE;
          when R_PRE_READ =>
            assert_adv_n := true;
            assert_ce_n  := true;
            assert_oe_n  := true;
            if delay_ctr = std_logic_vector(to_unsigned(5, delay_ctr'length)) then
              next_state := R_ASYNC_READ;
            end if;
          when R_ASYNC_READ =>  --release all control lines. have 8ns to get data.
            --data should be availiable.
            dram_read_reg <= dram_read_port_raw;
            --data_out_reg <= X"0000" & dram_read_port_raw;
            next_state    := R_ASYNC_COMPLETE;
            reset_delay   := true;
          when R_PRE_WRITE =>
            if write_is_cre = '1' then
              assert_cre := true;
            end if;
            assert_adv_n := true;
            assert_ce_n  := true;
            assert_we_n  := true;
            force_dq_ts  := false;      --present data to chip
            if delay_ctr = std_logic_vector(to_unsigned(5, delay_ctr'length)) then
              next_state := R_ASYNC_WRITE;
            end if;
          when R_ASYNC_WRITE =>  --realease lines. no requirement to hold dq!
            next_state  := R_ASYNC_COMPLETE;
            reset_delay := true;
          when R_ASYNC_COMPLETE =>  --we should not have this state... --1 clock to allow CE to remain deasserted
            data_out_reg <= X"0000" & dram_read_reg;  --hack! this is done on a write as well!!! (should  not matter though)
            write_is_cre <= '0';
            next_state   := after_async_state;
        end case;
      end if;  --reset

      r_state <= next_state;

      if ram_clock_on = true then
        n2ram_clock_en <= '1';
      else
        n2ram_clock_en <= '0';
      end if;

      if reset_delay = true then
        delay_ctr <= X"0000";
      else
        delay_ctr <= std_logic_vector(unsigned(delay_ctr) + 1);
      end if;

      if assert_cre = true then
        nx_ram_cre <= '1';
      else
        nx_ram_cre <= '0';
      end if;

      if assert_adv_n = true then
        nx_ram_adv_n <= '0';  --for async operation may be driven low always.
      else
        nx_ram_adv_n <= '1';
      end if;

      if assert_ce_n = true then
        nx_ram_ce_n <= '0';
      else
        nx_ram_ce_n <= '1';
      end if;

      if assert_oe_n = true then
        nx_ram_oe_n <= '0';
      else
        nx_ram_oe_n <= '1';
      end if;

      if assert_we_n = true then
        nx_ram_we_n <= '0';
      else
        nx_ram_we_n <= '1';
      end if;

      if assert_lb_n = true then
        nx_ram_lb_n <= '0';
      else
        nx_ram_lb_n <= '1';
      end if;

      if assert_ub_n = true then
        nx_ram_ub_n <= '0';
      else
        nx_ram_ub_n <= '1';
      end if;

      dram_write_port <= dram_write_port_val;

      if force_dq_ts = true then
        dq_ts <= '1';
      else
        dq_ts <= '0';
      end if;

      if clear_blk_xfer_count = true then
        blk_fill_count <= (others => '0');
      elsif inc_blk_xfer_count = true then
        blk_fill_count <= std_logic_vector(unsigned(blk_fill_count) + 1);
      end if;

      if assert_blk_strobe = true then
        blk_strobe <= '1';
      elsif capture_blk_din = true then
        blk_strobe  <= '1';
        blk_din_reg <= blk_din;
      else
        blk_strobe <= '0';
      end if;
      
    end if;  --clock
  end process;

end Behavioral;

