-- file created 2006
-- modified 2009, 2014
-- released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.tvc_pack.all;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
library UNISIM;
use UNISIM.VComponents.all;

entity pix_pusher is
  port (
    mu_bus_address : in    std_logic_vector(1 downto 0);
    mu_bus_data    : inout std_logic_vector(31 downto 0);
    mu_bus_write   : in    std_logic;
    mu_bus_strobe  : in    std_logic;

    odd_sc_irq     : out std_logic;
    odd_sc_select  : in  std_logic;
    even_sc_irq    : out std_logic;
    even_sc_select : in  std_logic;

    tvc_clock : in std_logic;
    tvc_reset : in std_logic;

    disp_raw_clk   : in std_logic;
    disp_raw_reset : in std_logic;

    disp_pix_value  : out std_logic_vector(7 downto 0);
    disp_horiz_sync : out std_logic;
    disp_vert_sync  : out std_logic
    );
end pix_pusher;

architecture Behavioral of pix_pusher is
  type PP_INT_STATE is (PP_PRE_HALT, PP_NORMAL, PP_ODD_FILL, PP_EVEN_FILL);
  signal pp_state : PP_INT_STATE;

  --signal framebuffer_start_addr       : std_logic_vector(23 downto 0);

  signal dd_reset_cntrs_req : std_logic;

  signal dd_even_fill_req        : std_logic;
  signal even_sc_start_fill      : std_logic;
  signal even_sc_start_fill_addr : std_logic_vector(23 downto 0);

  signal dd_odd_fill_req        : std_logic;
  signal odd_sc_start_fill      : std_logic;
  signal odd_sc_start_fill_addr : std_logic_vector(23 downto 0);

  --these signals are operating in the pixel clock domain and connect to the display driver.
  signal pix_addr_sig : std_logic_vector(11 downto 0);
  signal odd_pix_sig  : std_logic_vector(31 downto 0);
  signal even_pix_sig : std_logic_vector(31 downto 0);

  signal pixel_clock : std_logic;
  signal pixel_reset : std_logic;

  signal pixel_reset_in : std_logic;
  signal pixel_dll_locked_out : std_logic;
begin

  --jacks timing.
  --pixel_reset_in <= '1' when (tvc_reset = '1') or (disp_raw_reset = '1') else '0';
  pixel_reset_in <= '1' when (pixel_dll_locked_out /= '1') or (disp_raw_reset = '1') else '0';
  
  --pixel_clock is generated at bottom of file
  rsync : reset_sync
    generic map(init_state => '1')
    port map(clock => pixel_clock, di => pixel_reset_in, do => pixel_reset);

  
  odd_sc_sc : scanline_cache
    port map(mu_bus_address => mu_bus_address, mu_bus_data => mu_bus_data, mu_bus_write => mu_bus_write, mu_bus_strobe => mu_bus_strobe,
             sc_irq         => odd_sc_irq, sc_select => odd_sc_select, tvc_clock => tvc_clock, tvc_reset => tvc_reset,
             sc_start_fill  => odd_sc_start_fill, sc_start_fill_addr => odd_sc_start_fill_addr,
             pixel_clock    => pixel_clock, next_pix_addr => pix_addr_sig, next_pix_values => odd_pix_sig);

  even_sc_sc : scanline_cache
    port map(mu_bus_address => mu_bus_address, mu_bus_data => mu_bus_data, mu_bus_write => mu_bus_write, mu_bus_strobe => mu_bus_strobe,
             sc_irq         => even_sc_irq, sc_select => even_sc_select, tvc_clock => tvc_clock, tvc_reset => tvc_reset,
             sc_start_fill  => even_sc_start_fill, sc_start_fill_addr => even_sc_start_fill_addr,
             pixel_clock    => pixel_clock, next_pix_addr => pix_addr_sig, next_pix_values => even_pix_sig);

  d_driver : disp_driver
    port map(tvc_clock        => tvc_clock, tvc_reset => tvc_reset,
             odd_fill_req     => dd_odd_fill_req, even_fill_req => dd_even_fill_req, reset_cntrs_req => dd_reset_cntrs_req,
             pixel_clk        => pixel_clock, pixel_reset => pixel_reset, next_pix_addr=>pix_addr_sig,
             odd_next_pix_val => odd_pix_sig, even_next_pix_val => even_pix_sig,
             disp_pix_value   => disp_pix_value, disp_horiz_sync => disp_horiz_sync, disp_vert_sync => disp_vert_sync);

  process(tvc_clock)
  begin
    if tvc_clock = '1' and tvc_clock'event then
      if tvc_reset = '1' then
        pp_state                <= PP_PRE_HALT;
        --framebuffer_start_addr <= X"000000";
        even_sc_start_fill      <= '0';
        even_sc_start_fill_addr <= X"000000";
        odd_sc_start_fill       <= '0';
        odd_sc_start_fill_addr  <= X"000000";
      else                              --not a reset
        
        case pp_state is
          when PP_PRE_HALT =>
            pp_state <= PP_NORMAL;
          when PP_NORMAL =>
            if dd_reset_cntrs_req = '1' then
              odd_sc_start_fill_addr  <= X"000280";
              even_sc_start_fill_addr <= X"000000";
            elsif dd_odd_fill_req = '1' then
              odd_sc_start_fill <= '1';
              pp_state          <= PP_ODD_FILL;
            elsif dd_even_fill_req = '1' then
              even_sc_start_fill <= '1';
              pp_state           <= PP_EVEN_FILL;
            end if;
          when PP_ODD_FILL =>
            odd_sc_start_fill      <= '0';
            odd_sc_start_fill_addr <= std_logic_vector(unsigned(odd_sc_start_fill_addr) + 1280);
            pp_state               <= PP_NORMAL;
          when PP_EVEN_FILL =>
            even_sc_start_fill      <= '0';
            even_sc_start_fill_addr <= std_logic_vector(unsigned(even_sc_start_fill_addr) + 1280);
            pp_state                <= PP_NORMAL;
        end case;

      end if;  --reset
    end if;  --clock
  end process;

  DCM_pixel_clk : DCM_SP
    generic map (
      CLKDV_DIVIDE          => 2.0,  --  Divide by: 1.5,2.0,2.5,3.0,3.5,4.0,4.5,5.0,5.5,6.0,6.5
      --     7.0,7.5,8.0,9.0,10.0,11.0,12.0,13.0,14.0,15.0 or 16.0
      CLKFX_DIVIDE          => 4,       --  Can be any interger from 1 to 32
      CLKFX_MULTIPLY        => 2,       --  Can be any integer from 1 to 32
      CLKIN_DIVIDE_BY_2     => false,  --  TRUE/FALSE to enable CLKIN divide by two feature
      CLKIN_PERIOD          => 20.0,    --  Specify period of input clock
      CLKOUT_PHASE_SHIFT    => "NONE",  --  Specify phase shift of "NONE", "FIXED" or "VARIABLE" 
      CLK_FEEDBACK          => "NONE",  --  Specify clock feedback of "NONE", "1X" or "2X" 
      DESKEW_ADJUST         => "SYSTEM_SYNCHRONOUS",  -- "SOURCE_SYNCHRONOUS", "SYSTEM_SYNCHRONOUS" or
                                        --     an integer from 0 to 15
      DLL_FREQUENCY_MODE    => "LOW",  -- "HIGH" or "LOW" frequency mode for DLL
      DUTY_CYCLE_CORRECTION => true,  --  Duty cycle correction, TRUE or FALSE
      PHASE_SHIFT           => 0,  --  Amount of fixed phase shift from -255 to 255
      STARTUP_WAIT          => true)  --  Delay configuration DONE until DCM_SP LOCK, TRUE/FALSE
    port map (
      --CLK0 => CLK0,     -- 0 degree DCM CLK ouptput
      --CLK180 => CLK180, -- 180 degree DCM CLK output
      --CLK270 => CLK270, -- 270 degree DCM CLK output
      --CLK2X => CLK2X,   -- 2X DCM CLK output
      --CLK2X180 => CLK2X180, -- 2X, 180 degree DCM CLK out
      --CLK90 => CLK90,   -- 90 degree DCM CLK output
      --CLKDV => CLKD,V   -- Divided DCM CLK out (CLKDV_DIVIDE)
      CLKFX => pixel_clock,               -- DCM CLK synthesis out (M/D)
      --CLKFX180 => clk_ram_adv, -- 180 degree CLK synthesis out
      LOCKED => pixel_dll_locked_out, -- DCM LOCK status output
      --PSDONE => PSDONE, -- Dynamic phase adjust done output
      --STATUS => STATUS, -- 8-bit DCM status bits output
      --CLKFB => CLKFB,   -- DCM clock feedback
      CLKIN => disp_raw_clk,        -- Clock input (from IBUFG, BUFG or DCM)
     --PSCLK => PSCLK,   -- Dynamic phase adjust clock input
     --PSEN => PSEN,     -- Dynamic phase adjust enable input
     --PSINCDEC => PSINCDEC, -- Dynamic phase adjust increment/decrement
     RST => tvc_reset        -- DCM asynchronous reset input
      );


end Behavioral;
