-- Create Date:    11:25:24 08/05/2006
-- modified 2009, 2014
-- released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity poly_scanline is
  port (
    --memory user access bus 
    mu_bus_address : in    std_logic_vector(1 downto 0);
    mu_bus_data    : inout std_logic_vector(31 downto 0);
    mu_bus_write   : in    std_logic;  --if high the MCU is going to write to a mu
    mu_bus_strobe  : in    std_logic;

    poly_scanline_irq    : out std_logic;
    poly_scanline_select : in  std_logic;

    tvc_clock : in std_logic;
    tvc_reset : in std_logic;

    --interface to the eppcu
    cmd_bus_data   : in std_logic_vector(31 downto 0);
    cmd_bus_addr   : in std_logic_vector(7 downto 0);
    cmd_bus_wr     : in std_logic;
    cmd_bus_strobe : in std_logic;

    idle : out std_logic
    );
end poly_scanline;

architecture Behavioral of poly_scanline is

  type PS_INT_STATE is (PS_IDLE, PS_STARTUP, PS_STARTUP_1, PS_FETCH_FB_WAIT, PS_FETCH_FB, PS_FETCH_DB_WAIT, PS_FETCH_DB,
                        PS_COMPUTE_1, PS_COMPUTE_1A, PS_COMPUTE_2,
                        PS_TEX_ADDRESS_1, PS_TEX_ADDRESS_2, PS_TEX_ADDRESS_3, PS_TEX_FETCH_WAIT, PS_TEX_FETCH,
                        PS_COMPLETE_COMPUTE_LL,
                        PS_WRITE_FB_WAIT, PS_WRITE_FB, PS_WRITE_DB_WAIT, PS_WRITE_DB,
                        PS_COMPLETE_COMPUTE);
  signal ps_state : PS_INT_STATE := PS_IDLE;

  signal scanline_remaining : std_logic_vector(15 downto 0) := (others => '0');

  signal mu_bus_data_out : std_logic_vector(31 downto 0) := (others => '0');

  signal work_curr_num   : std_logic_vector(4 downto 0)  := (others => '0');  --the current pixel in the cache (used in compute)
  signal work_num_pixels : std_logic_vector(4 downto 0)  := (others => '0');  --how many pixels in the poly caches. extra bit is tacky but important
  signal work_fb_address : std_logic_vector(31 downto 0) := (others => '0');
  signal work_db_address : std_logic_vector(31 downto 0) := (others => '0');

  signal compute_fb_address : std_logic_vector(31 downto 0);
  
  signal curr_z_depth : std_logic_vector(31 downto 0) := (others => '0');  --a signed value
  signal z_increment  : std_logic_vector(31 downto 0) := (others => '0');  -- a signed value
  signal pix_value    : std_logic_vector(7 downto 0)  := (others => '0');

  signal tc_x       : std_logic_vector(15 downto 0) := (others => '0');
  signal tc_y       : std_logic_vector(15 downto 0) := (others => '0');
  signal tc_x_inc   : std_logic_vector(15 downto 0) := (others => '0');
  signal tc_y_inc   : std_logic_vector(15 downto 0) := (others => '0');
  signal texture_id : std_logic_vector(2 downto 0)  := (others => '0');

  component poly_cache is
    generic(WIDTH : positive);
    port(
      addr         : in  std_logic_vector(3 downto 0);
      data_out     : out std_logic_vector(WIDTH-1 downto 0);
      data_in      : in  std_logic_vector(WIDTH-1 downto 0);
      write_enable : in  std_logic;
      clk          : in  std_logic
      );
  end component;

  signal fbc_data_out : std_logic_vector(31 downto 0) := (others => '0');
  signal fbc_data_in  : std_logic_vector(31 downto 0) := (others => '0');
  signal fbc_addr     : unsigned(3 downto 0)          := (others => '0');
  signal fbc_write    : std_logic                     := '0';

  signal dc_data_out : std_logic_vector(31 downto 0) := (others => '0');
  signal dc_data_in  : std_logic_vector(31 downto 0) := (others => '0');
  signal dc_addr     : unsigned(3 downto 0)          := (others => '0');
  signal dc_write    : std_logic                     := '0';

  signal mult_bit_mask : std_logic_vector(7 downto 0);
  signal mult_n1_full  : std_logic_vector(15 downto 0);
  signal tex_address   : unsigned(23 downto 0);  --should be 31, but running out of space in fpga..

  signal depth_test_valid : std_logic := '0';  --we should update the front buffer and depth buffer
  signal depth_test_valid_once : std_logic;

  signal xfer_config_reg  : std_logic_vector(31 downto 0);
  signal xfer_fb_addr_reg : std_logic_vector(31 downto 0);

  procedure Setup_FB_Xfer(
    constant is_read        : in  boolean;
    signal work_num_pixels  : in  std_logic_vector(4 downto 0);
    signal work_fb_address  : in  std_logic_vector(31 downto 0);
    signal xfer_config_reg  : out std_logic_vector(31 downto 0);
    signal xfer_fb_addr_reg : out std_logic_vector(31 downto 0)) is
    variable cmd : std_logic_vector(7 downto 0);
    variable end_addr : unsigned(31 downto 0);
    variable dword_xfers : unsigned(6 downto 0);
    variable odd_start : boolean;
    variable odd_end : boolean;
  begin
    if is_read = true then
      cmd := X"00";
    else
      cmd := X"80";
    end if;

    --this is pretty ugly.
    end_addr := unsigned(work_fb_address) + unsigned(work_num_pixels);

    odd_start := false;
    odd_end := false;

    if work_fb_address(1 downto 0) /= "00" then
      odd_start := true;
    end if;

    if end_addr(1 downto 0) /= "00" then
      odd_end := true;
    end if;

    if odd_start = true and odd_end = true then
      dword_xfers := unsigned(work_num_pixels(4 downto 2)) + 2;
    elsif ((odd_start = true) and (odd_end = false)) or ((odd_start = false) and (odd_end = true)) then
      dword_xfers := unsigned(work_num_pixels(4 downto 2)) + 1;
    else
      dword_xfers := "0000" & unsigned(work_num_pixels(4 downto 2));
    end if;
    
    xfer_config_reg <= cmd & X"00_00" & "0" & std_logic_vector(dword_xfers);
    xfer_fb_addr_reg <= work_fb_address;
  end Setup_FB_Xfer;

    procedure Setup_DB_Xfer(
    constant is_read        : in  boolean;
    signal work_num_pixels  : in  std_logic_vector(4 downto 0);
    signal work_db_address  : in  std_logic_vector(31 downto 0);
    signal xfer_config_reg  : out std_logic_vector(31 downto 0);
    signal xfer_fb_addr_reg : out std_logic_vector(31 downto 0)) is
    variable cmd : std_logic_vector(7 downto 0);
  begin
    if is_read = true then
      cmd := X"00";
    else
      cmd := X"80";
    end if;
    xfer_config_reg  <= cmd & X"0000" & "000" & work_num_pixels(4 downto 0);
    xfer_fb_addr_reg <= work_db_address;
  end Setup_DB_Xfer;
  
begin
  fb_cache : poly_cache generic map (WIDTH => 32)
    port map(addr => std_logic_vector(fbc_addr), data_out => fbc_data_in, data_in => fbc_data_out, write_enable => fbc_write, clk => tvc_clock);

  depth_cache : poly_cache generic map (WIDTH => 32)
    port map(addr => std_logic_vector(dc_addr), data_out => dc_data_in, data_in => dc_data_out, write_enable => dc_write, clk => tvc_clock);

  idle <= '1' when ps_state = PS_IDLE else '0';

  process(poly_scanline_select, mu_bus_write, mu_bus_address, xfer_config_reg, xfer_fb_addr_reg, mu_bus_data_out) is
    variable case_input : std_logic_vector(3 downto 0);
  begin
    case_input := poly_scanline_select & mu_bus_write & mu_bus_address;
    case case_input is
      when "1001" =>
        mu_bus_data <= xfer_config_reg;
      when "1010" =>
        mu_bus_data <= xfer_fb_addr_reg;
      when "1000" =>
        mu_bus_data <= mu_bus_data_out;
      when others =>
        mu_bus_data <= (others => 'Z');
    end case;
  end process;

  process(ps_state, fbc_data_in, dc_data_in) is
  begin
    case ps_state is
      when PS_WRITE_FB=>
        mu_bus_data_out <= fbc_data_in;
      when PS_WRITE_DB=>
        mu_bus_data_out <= dc_data_in;
      when others=>
        mu_bus_data_out <= (others => '0');
    end case;
  end process;

  poly_scanline_irq <= '1' when (ps_state = PS_FETCH_FB_WAIT) or (ps_state = PS_FETCH_DB_WAIT) or (ps_state = PS_TEX_FETCH_WAIT)
                       or (ps_state = PS_WRITE_FB_WAIT) or (ps_state = PS_WRITE_DB_WAIT) else '0';

  process (tvc_clock) is
    variable do_fbc_write : boolean;
    variable do_dc_write  : boolean;
    variable next_state   : PS_INT_STATE;
  begin
    if tvc_clock = '1' and tvc_clock'event then
      do_fbc_write := false;
      do_dc_write  := false;
      next_state   := ps_state;

      if tvc_reset = '1' then
        next_state := PS_IDLE;
      else                              --not reset
        case ps_state is
          when PS_IDLE =>
            if cmd_bus_strobe = '1' and cmd_bus_wr = '1' then
              case cmd_bus_addr is
                when X"F0" =>
                  work_fb_address <= cmd_bus_data;
                when X"F1" =>
                  work_db_address <= cmd_bus_data;
                when X"F2" =>
                  curr_z_depth <= cmd_bus_data;
                when X"F3" =>
                  z_increment <= cmd_bus_data;
                when X"F4" =>
                  scanline_remaining <= cmd_bus_data(15 downto 0);
                when X"F5" =>
                  pix_value <= cmd_bus_data(7 downto 0);
                when X"F6" =>
                  tc_x <= cmd_bus_data(31 downto 16);
                  tc_y <= cmd_bus_data(15 downto 0);
                when X"F7" =>
                  tc_x_inc <= cmd_bus_data(31 downto 16);
                  tc_y_inc <= cmd_bus_data(15 downto 0);
                when X"F8" =>
                  texture_id <= cmd_bus_data(2 downto 0);
                when X"F9" =>
                  --note we don't care about command bus value
                  --if scanline length is zero, we end right back up in idle.
                  next_state := PS_STARTUP;
                when others =>
                  null;
              end case;
            end if;
          when PS_STARTUP =>            --also used for loop termination
            if scanline_remaining = X"0000" then
              next_state := PS_IDLE;
            else
              next_state := PS_STARTUP_1;
            end if;
            if scanline_remaining > X"000F" then
              work_num_pixels <= "1" & X"0";
            else
              work_num_pixels <= "0" & scanline_remaining(3 downto 0);
            end if;
          when PS_STARTUP_1 =>
            next_state := PS_FETCH_FB_WAIT;
            Setup_FB_Xfer(true, work_num_pixels, work_fb_address, xfer_config_reg, xfer_fb_addr_reg);
            fbc_addr <= (others=>'0');
          when PS_FETCH_FB_WAIT =>
            if poly_scanline_select = '1' then
              next_state := PS_FETCH_FB;
            end if;
          when PS_FETCH_FB =>
            if poly_scanline_select = '0' then
              next_state       := PS_FETCH_DB_WAIT;
              Setup_DB_Xfer(true, work_num_pixels, work_db_address, xfer_config_reg, xfer_fb_addr_reg);
              dc_addr <= (others=>'0');
            end if;
            if mu_bus_strobe = '1' then
              fbc_data_out <= mu_bus_data;
              fbc_addr     <= fbc_addr + 1;
              do_fbc_write := true;
            end if;
          when PS_FETCH_DB_WAIT =>
            if poly_scanline_select = '1' then
              next_state := PS_FETCH_DB;
            end if;
          when PS_FETCH_DB =>
            if poly_scanline_select = '0' then
              next_state    := PS_COMPUTE_1;
              work_curr_num <= "0" & X"0";
              depth_test_valid_once <= '0';
              compute_fb_address <= work_fb_address;
            end if;
            if mu_bus_strobe = '1' then
              dc_data_out <= mu_bus_data;
              dc_addr     <= dc_addr + 1;
              do_dc_write := true;
            end if;
          when PS_COMPUTE_1 =>
            fbc_addr <= unsigned( compute_fb_address(5 downto 2) );
            dc_addr  <= unsigned(work_curr_num(3 downto 0));

            case texture_id is
              when "001" =>
                --tex_address <= X"30_0000";
                tex_address <= X"0C_3500";
              when "010" =>
                --tex_address <= X"38_0000";
                tex_address <= X"10_3500";
              when "011" =>
                --tex_address <= X"40_0000";
                tex_address <= X"14_3500";
              when "100" =>
                --tex_address <= X"48_0000";
                tex_address <= X"18_3500";
              when "101" =>
                --tex_address <= X"50_0000";
                tex_address <= X"1C_3500";
              when "110" =>
                --tex_address <= X"58_0000";
                tex_address <= X"20_3500";
              when "111" =>
                --tex_address <= X"60_0000";
                tex_address <= X"24_3500";
              when others =>
                null;
            end case;

            if signed(curr_z_depth) > signed(dc_data_in) then
              depth_test_valid <= '1';
              depth_test_valid_once <= '1';
            else
              depth_test_valid <= '0';
            end if;
            next_state := PS_COMPUTE_1A;
          when PS_COMPUTE_1A =>         --this state is a hack.
            if depth_test_valid = '0' or texture_id = "000" then
              next_state := PS_COMPUTE_2;
            else
              next_state    := PS_TEX_ADDRESS_1;
              mult_bit_mask <= "00000001";
              mult_n1_full  <= X"01" & X"00";
            end if;
          when PS_TEX_ADDRESS_1 =>
            if (tc_y(15 downto 8) and mult_bit_mask) = mult_bit_mask then
              tex_address <= tex_address + unsigned(mult_n1_full & "00");
            end if;
            mult_n1_full  <= mult_n1_full(14 downto 0) & '0';  --left shift
            mult_bit_mask <= mult_bit_mask(6 downto 0) & '0';  --left shift
            if mult_bit_mask(7) = '1' then            --final bit
              next_state := PS_TEX_ADDRESS_2;
            end if;
          when PS_TEX_ADDRESS_2 =>
            tex_address <= tex_address + unsigned(tc_x(15 downto 8) & "00");
            next_state  := PS_TEX_ADDRESS_3;
          when PS_TEX_ADDRESS_3 =>
            xfer_config_reg  <= X"0000000" & "0001";  --fetch one 32 bit value
            xfer_fb_addr_reg <= X"00" & "00" & std_logic_vector(tex_address(23 downto 2));
            next_state       := PS_TEX_FETCH_WAIT;
          when PS_TEX_FETCH_WAIT =>
            if poly_scanline_select = '1' then
              next_state := PS_TEX_FETCH;
            end if;
          when PS_TEX_FETCH =>
            if poly_scanline_select = '0' then
              next_state := PS_COMPUTE_2;
            end if;
            if mu_bus_strobe = '1' then
              pix_value <= mu_bus_data(31 downto 30) & mu_bus_data(23 downto 21) & mu_bus_data(15 downto 13);
              --pix_value <= mu_bus_data(31 downto 29) & mu_bus_data(24 downto 22) & mu_bus_data(15 downto 14);
            end if;
          when PS_COMPUTE_2 =>
            if depth_test_valid = '1' then
              case work_fb_address(1 downto 0) is
                when "00" =>
                  fbc_data_out <= pix_value & fbc_data_in(23 downto 0);
                when "01" =>
                  fbc_data_out <= fbc_data_in(31 downto 24) & pix_value & fbc_data_in(15 downto 0);
                when "10" =>
                  fbc_data_out <= fbc_data_in(31 downto 16) & pix_value & fbc_data_in(7 downto 0);
                when others =>          --'11'
                  fbc_data_out <= fbc_data_in(31 downto 8) & pix_value;
              end case;

              dc_data_out  <= curr_z_depth;
              do_fbc_write := true;
              do_dc_write  := true;
            end if;

            --fbc_data_out <= X"FFFF_FFFF";
            --do_fbc_write := true;

            curr_z_depth <= std_logic_vector(signed(curr_z_depth) + signed(z_increment));
            if texture_id /= "000" then
              tc_x <= std_logic_vector(signed(tc_x) + signed(tc_x_inc));
              tc_y <= std_logic_vector(signed(tc_y) + signed(tc_y_inc));
            end if;

            if work_curr_num = work_num_pixels then
              if depth_test_valid_once = '1' then
                next_state := PS_COMPLETE_COMPUTE_LL;
              else
                next_state := PS_COMPLETE_COMPUTE;
              end if;
            else
              compute_fb_address <= std_logic_vector(unsigned(compute_fb_address) + 1);
              work_curr_num <= std_logic_vector(unsigned(work_curr_num) + 1);
              next_state    := PS_COMPUTE_1;
            end if;
          when PS_COMPLETE_COMPUTE_LL =>
            next_state       := PS_WRITE_FB_WAIT;
            Setup_FB_Xfer(false, work_num_pixels, work_fb_address, xfer_config_reg, xfer_fb_addr_reg);
            fbc_addr         <= (others=>'0');
          when PS_WRITE_FB_WAIT =>
            if poly_scanline_select = '1' then
              next_state := PS_WRITE_FB;
            end if;
          when PS_WRITE_FB =>
            if poly_scanline_select = '0' then
              next_state       := PS_WRITE_DB_WAIT;
              Setup_DB_Xfer(false, work_num_pixels, work_db_address, xfer_config_reg, xfer_fb_addr_reg);
              dc_addr          <= (others=>'0');
            end if;
            if mu_bus_strobe = '1' then
              fbc_addr <= fbc_addr + 1;
            end if;
          when PS_WRITE_DB_WAIT=>
            if poly_scanline_select = '1' then
              next_state := PS_WRITE_DB;
            end if;
          when PS_WRITE_DB=>
            if poly_scanline_select = '0' then
              next_state := PS_COMPLETE_COMPUTE;
            end if;
            if mu_bus_strobe = '1' then
              dc_addr <= dc_addr + 1;
            end if;
          when PS_COMPLETE_COMPUTE =>  --maybe could/should be pushed into WRITE_DB
            scanline_remaining <= std_logic_vector(unsigned(scanline_remaining) - unsigned(work_num_pixels));
            work_fb_address    <= std_logic_vector(unsigned(work_fb_address) + unsigned(work_num_pixels));
            work_db_address    <= std_logic_vector(unsigned(work_db_address) + unsigned(work_num_pixels & "00"));
            next_state         := PS_STARTUP;
        end case;
      end if;  --not reset

      ps_state <= next_state;

      if do_fbc_write = true then
        fbc_write <= '1';
      else
        fbc_write <= '0';
      end if;

      if do_dc_write = true then
        dc_write <= '1';
      else
        dc_write <= '0';
      end if;
    end if;  --clock
  end process;

end Behavioral;

