-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library UNISIM;
use UNISIM.VComponents.all;

use work.shader_cpu_pack.all;

entity ram_data is
  port(
    clock : in std_logic;

    cpu_addr : in  std_logic_vector(15 downto 0);
    cpu_din  : in  std_logic_vector(31 downto 0);
    cpu_dout : out std_logic_vector(31 downto 0);
    cpu_we   : in  std_logic;

    dma_int_strobe : in  std_logic;
    dma_int_we     : in  std_logic;
    dma_int_addr   : in  std_logic_vector(15 downto 0);
    dma_int_dout   : out std_logic_vector(31 downto 0);
    dma_int_din    : in  std_logic_vector(31 downto 0)
    );
end ram_data;

architecture rtl of ram_data is
  signal local_dma_we : std_logic;

  signal cpu_we_0, cpu_we_1, cpu_we_2, cpu_we_3         : std_logic;
  signal dma_we_0, dma_we_1, dma_we_2, dma_we_3         : std_logic;
  signal do_a_0, do_a_1, do_a_2, do_a_3 : std_logic_vector(31 downto 0);
  signal do_b_0, do_b_1, do_b_2, do_b_3 : std_logic_vector(31 downto 0);

  --preserve the register read at the rising edge to mux out the proper value
  --in next clock period when another address might be being presented
  signal addr_read_a_reg : std_logic_vector(15 downto 0);
  signal addr_read_b_reg : std_logic_vector(15 downto 0);
  
begin

  process(clock) is
  begin
    if rising_edge(clock) then
      addr_read_a_reg <= cpu_addr;
      addr_read_b_reg <= dma_int_addr;
    end if;
  end process;

  local_dma_we <= '1' when dma_int_we = '1' and  dma_int_strobe = '1' else '0';

  dma_we_0 <= local_dma_we when M_0( dma_int_addr ) else '0';
  dma_we_1 <= local_dma_we when M_1( dma_int_addr ) else '0';
  dma_we_2 <= local_dma_we when M_2( dma_int_addr ) else '0';
  dma_we_3 <= local_dma_we when M_3( dma_int_addr ) else '0';

  cpu_we_0 <= cpu_we when M_0( cpu_addr ) else '0';
  cpu_we_1 <= cpu_we when M_1( cpu_addr ) else '0';
  cpu_we_2 <= cpu_we when M_2( cpu_addr ) else '0';
  cpu_we_3 <= cpu_we when M_3( cpu_addr ) else '0';

  cpu_dout <=
    do_a_0 when M_0( addr_read_a_reg ) else
    do_a_1 when M_1( addr_read_a_reg ) else
    do_a_2 when M_2( addr_read_a_reg ) else
    do_a_3 when M_3( addr_read_a_reg ) else
    X"0000_0000";

  dma_int_dout <=
    do_b_0 when M_0( addr_read_b_reg ) else
    do_b_1 when M_1( addr_read_b_reg ) else
    do_b_2 when M_2( addr_read_b_reg ) else
    do_b_3 when M_3( addr_read_b_reg ) else
    X"0000_0000";

  in_ram_0 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => cpu_din, WEA => cpu_we_0, DOA => do_a_0,
             ADDRA => cpu_addr(8 downto 0), DIB => dma_int_din, ADDRB => dma_int_addr(8 downto 0), WEB => dma_we_0, DOB => do_b_0);

  in_ram_1 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => cpu_din, WEA => cpu_we_1, DOA => do_a_1,
             ADDRA => cpu_addr(8 downto 0), DIB => dma_int_din, ADDRB => dma_int_addr(8 downto 0), WEB => dma_we_1, DOB => do_b_1);

  in_ram_2 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => cpu_din, WEA => cpu_we_2, DOA => do_a_2,
             ADDRA => cpu_addr(8 downto 0), DIB => dma_int_din, ADDRB => dma_int_addr(8 downto 0), WEB => dma_we_2, DOB => do_b_2);

  in_ram_3 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => cpu_din, WEA => cpu_we_3, DOA => do_a_3,
             ADDRA => cpu_addr(8 downto 0), DIB => dma_int_din, ADDRB => dma_int_addr(8 downto 0), WEB => dma_we_3, DOB => do_b_3);


end architecture;
