-- file created 2006
-- modified 2009, 2014
-- released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
library UNISIM;
use UNISIM.VComponents.all;

entity scanline_cache is
  port(
    mu_bus_address : in    std_logic_vector(1 downto 0);
    mu_bus_data    : inout std_logic_vector(31 downto 0);
    mu_bus_write   : in    std_logic;
    mu_bus_strobe  : in    std_logic;

    sc_irq    : out std_logic;
    sc_select : in  std_logic;

    tvc_clock : in std_logic;
    tvc_reset : in std_logic;

    sc_start_fill      : in std_logic;  --change the state to ask the mcu with data starting from the address in sc_start_fill_addr
    sc_start_fill_addr : in std_logic_vector(23 downto 0);

    --these signals only connect to the second port of the sram and operate in the pixel clock domain
    --they are mostly ignored by the logic in this class.
    pixel_clock     : in  std_logic;
    next_pix_addr   : in  std_logic_vector(11 downto 0);
    next_pix_values : out std_logic_vector(31 downto 0)  -- when in drain mode this is 'updated' on every clock.
    );
end scanline_cache;

architecture Behavioral of scanline_cache is

  type SC_CURR_STATE is (SC_IDLE, SC_WAITING, SC_MCU_DRIVEN);
  signal c_state : SC_CURR_STATE;

  --signals for the block ram
  signal brWE  : std_logic;
  signal brEN  : std_logic;
  signal brRST : std_logic;

  signal brDI : std_logic_vector(31 downto 0);
  --signal brDO : std_logic_vector(31 downto 0);

  signal brADDR : unsigned(7 downto 0);

  signal scanline_addr_start : std_logic_vector(23 downto 0);  --a local copy of the data from sc_start_fill_addr
  
begin

  high_word : RAMB4_S16_S16
    port map(WEA   => brWE, ENA => brEN, RSTA => brRST, CLKA => tvc_clock,
             ADDRA => std_logic_vector(brADDR), DIA => brDI(31 downto 16),  --DOA=>brDO(31 downto 16)
             WEB   => '0', ENB => '1', RSTB => '0', CLKB => pixel_clock,
             ADDRB => next_pix_addr(9 downto 2), DIB => (others => '0'),
             DOB   => next_pix_values(31 downto 16));
  low_word : RAMB4_S16_S16
    port map(WEA   => brWE, ENA => brEN, RSTA => brRST, CLKA => tvc_clock,
             ADDRA => std_logic_vector(brADDR), DIA => brDI(15 downto 0),  --DOA=>brDO(15 downto 0) );                         
             WEB   => '0', ENB => '1', RSTB => '0', CLKB => pixel_clock,
             ADDRB => next_pix_addr(9 downto 2), DIB => (others => '0'),
             DOB   => next_pix_values(15 downto 0));

  brDI  <= mu_bus_data;
  brWE  <= '1' when (c_state = SC_MCU_DRIVEN) and (mu_bus_write = '1') and (mu_bus_strobe = '1') else '0';
  brEN  <= '1';
  brRST <= '0';

  sc_irq <= '1' when c_state = SC_WAITING else '0';

  process(sc_select, mu_bus_write, mu_bus_address, scanline_addr_start) is
    variable case_input : std_logic_vector(3 downto 0);
  begin
    case_input := sc_select & mu_bus_write & mu_bus_address;
    case case_input is
      when "1001" =>
        mu_bus_data <= X"00" & X"0000A0";  --X"0000A0";
      when "1010" =>
        mu_bus_data <= X"00" & scanline_addr_start;
      when others =>
        mu_bus_data <= (others => 'Z');
    end case;
  end process;

  process(tvc_clock) is
    variable next_state : SC_CURR_STATE;
  begin
    if tvc_clock = '1' and tvc_clock'event then
      next_state := c_state;

      if tvc_reset = '1' then
        next_state := SC_IDLE;
      else
        case c_state is
          when SC_IDLE =>
            if sc_start_fill = '1' then
              scanline_addr_start <= sc_start_fill_addr;
              next_state          := SC_WAITING;
              brADDR              <= (others => '0');
            end if;
          when SC_WAITING =>
            if sc_select = '1' then
              next_state := SC_MCU_DRIVEN;
            end if;
          when SC_MCU_DRIVEN =>
            if sc_select = '0' then
              next_state := SC_IDLE;
            end if;
            if mu_bus_strobe = '1' then
              brADDR <= brADDR + 1;
            end if;
        end case;
      end if;  --reset
      c_state <= next_state;
    end if;  --clock
  end process;
end Behavioral;
