-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use std.textio.all;
use ieee.std_logic_textio.all;

use work.shader_cpu_pack.all;

entity shader_cpu is
  port(
    clock    : in std_logic;
    reset_ll : in std_logic;            --whole chip reset

    logic_reset : in  std_logic;
    logic_halt  : in  std_logic;
    halt_status : out std_logic;

    prog_address : out std_logic_vector(15 downto 0);
    ins_dword    : in  std_logic_vector(31 downto 0);

    d_ram_addr : out std_logic_vector(15 downto 0);
    d_ram_dout : out std_logic_vector(31 downto 0);
    d_ram_din  : in  std_logic_vector(31 downto 0);
    d_ram_we   : out std_logic;

    ext_data_strobe  : out std_logic;
    ext_data_address : out std_logic_vector(7 downto 0);
    ext_data_out     : out std_logic_vector(31 downto 0);
    ext_data_in      : in  std_logic_vector(31 downto 0);
    ext_data_wr      : out std_logic;
    ext_data_busy    : in  std_logic
    );
end entity;

architecture behavioural of shader_cpu is
  signal cpu_reset : std_logic;

  --control signals
  --most are outputs from retire
  signal ins_halted    : std_logic;
  --signal ins_halted_reg : std_logic;
  signal jump_new_addr : std_logic_vector(15 downto 0);
  signal jump_set_pc   : std_logic;
  signal hold          : std_logic;

  signal ls_busy : std_logic;

  signal a_reg_value : std_logic_vector(31 downto 0);
  signal b_reg_value : std_logic_vector(31 downto 0);

  --signals and registers related to pipeline implementation

  -- ****************************
  --the output of the fetch stage
  -- ****************************

  --duplicate of value in output registers of iram -> no buffering
  signal fetched_insn           : std_logic_vector(31 downto 0);
  signal fetch_output_is_bubble : std_logic;
  --the address of the above instruction. not used here for debugging
  signal fetched_addr           : std_logic_vector(15 downto 0);
  signal fetch_collision_in     : std_logic;

  signal buffer_out_is_bubble : std_logic;
  -- ****************************
  -- start of the compute & load/store stage
  -- ****************************

  --the input of the compute stage. copy of above value unless in reset.
  signal compute_insn_in      : std_logic_vector(31 downto 0);
  signal compute_insn_prev    : std_logic_vector(31 downto 0);
  --we ripple these values through the pipeline for debugging.
  --they will be stripped out by implementation
  signal compute_is_bubble_in : std_logic;
  signal compute_addr_in      : std_logic_vector(15 downto 0);

  --the compute instruction first goes through the decoder
  --these signals are the output of the decoder.
  signal dec_opp          : std_logic_vector(6 downto 0);
  signal dec_a_reg_addr   : std_logic_vector(2 downto 0);
  signal dec_b_reg_addr   : std_logic_vector(2 downto 0);
  signal dec_c_reg_addr   : std_logic_vector(2 downto 0);
  signal dec_imm          : std_logic_vector(15 downto 0);
  signal dec_imm_sign_ext : std_logic_vector(31 downto 0);

  --signal reg_c_value    : std_logic_vector(31 downto 0);
  signal reg_c_addr            : std_logic_vector(2 downto 0);
  signal maybe_a_reg_collision : std_logic;
  signal maybe_b_reg_collision : std_logic;
  signal reg_collision         : std_logic;
  --compute stage local. used to possibly skip a load or store
  --not triggered in compiled code (yet), but certainly possible
  --to code in assembly.
  signal compute_skip_insn_reg : std_logic;
  signal ls_skip_ins           : std_logic;
  signal compute_a_reg_used    : std_logic;
  signal compute_b_reg_used    : std_logic;

  --outputs of the compute stage
  signal compute_reg_c_addr_out   : std_logic_vector(2 downto 0);
  signal compute_reg_c_value_out  : std_logic_vector(31 downto 0);
  signal compute_skip_next_insn   : std_logic;
  signal compute_ls_mem_value_out : std_logic_vector(31 downto 0);
  signal compute_is_bubble_out    : std_logic;

  --inputs of the retire state
  signal retire_skip_insn_reg : std_logic;
  signal retire_in_opp_reg    : std_logic_vector(6 downto 0);
  signal retire_in_imm_reg    : std_logic_vector(15 downto 0);
  signal retire_in_c_value    : std_logic_vector(31 downto 0);
  signal retire_in_c_addr     : std_logic_vector(2 downto 0);
  signal retire_in_mem_value  : std_logic_vector(31 downto 0);
  signal retire_in_pc_addr    : std_logic_vector(15 downto 0);  --debugging only.
  signal retire_is_bubble_in  : std_logic;

  signal retire_out_is_bubble : std_logic;  --debuggin only

  signal just_stored      : std_logic;
  signal just_stored_addr : std_logic_vector(2 downto 0);

  --runs internally between the retire stage and the register bank
  signal do_reg_c_write : std_logic;
  signal retire_c_value : std_logic_vector(31 downto 0);
begin

  cpu_reset   <= '1' when reset_ll = '1' or logic_reset = '1' else '0';
  halt_status <= ins_halted;

  fetch_stage : fetch
    port map(clock    => clock, reset => cpu_reset, pc_to_fetch => prog_address, ins_fetched => ins_dword,
             new_addr => jump_new_addr, set_pc => jump_set_pc, collision_in => fetch_collision_in,
             hold_pc  => hold, instruction => fetched_insn, out_is_bubble => fetch_output_is_bubble, fetched_addr => fetched_addr);


  --the input of the compute stage. copy of value from fetch stage unless in reset.
  --after a jump we have to eat an extra instruction.
  --note that none of the skip instructions have pipeline delaying side effects
  --like causing register conflicts or causing jumps themselves. this means
  --they can be ignored here.

  --this is the buffer stage. not complex enough to warrant its own file
  process(clock) is
  begin
    if rising_edge(clock) then
      if hold /= '1' and reg_collision /= '1' then
        compute_addr_in      <= fetched_addr;
        compute_insn_in      <= fetched_insn;
        buffer_out_is_bubble <= fetch_output_is_bubble;
      end if;
    end if;
  end process;

  --compute_addr_in      <= fetched_addr;
  --compute_insn_in      <= fetched_insn;
  --compute_is_bubble_in <= '1' when jump_set_pc = '1' else buffer_out_is_bubble;
  compute_is_bubble_in <= '1' when jump_set_pc = '1' else buffer_out_is_bubble;

  fetch_collision_in <= reg_collision;

  decode : decoder
    port map(ins_dword_in => compute_insn_in, opp => dec_opp, a_reg_addr => dec_a_reg_addr, b_reg_addr => dec_b_reg_addr,
             c_reg_addr   => dec_c_reg_addr, imm => dec_imm, imm_sign_ext => dec_imm_sign_ext);

  --This is a bug. We don't use reg_a and reg_b on every instruction.
  --We must ignore collisions with registers we don't use.
  maybe_a_reg_collision <= '1' when ((just_stored_addr = dec_a_reg_addr) and (compute_a_reg_used = '1'))                     else '0';
  maybe_b_reg_collision <= '1' when ((just_stored_addr = dec_b_reg_addr) and (compute_b_reg_used = '1'))                     else '0';
  reg_collision         <= '1' when (just_stored = '1') and ((maybe_a_reg_collision = '1') or (maybe_b_reg_collision = '1')) else '0';

  hold        <= '1' when (ins_halted = '1') or (logic_halt = '1') or (ls_busy = '1')                                                                        else '0';
  ls_skip_ins <= '1' when (compute_skip_insn_reg = '1') or (compute_is_bubble_in = '1') or (reg_collision = '1') or (ins_halted = '1') or (logic_halt = '1') else '0';

  compute_stage : compute
    port map(clock       => clock, reset => cpu_reset, opcode => dec_opp, a_reg => dec_a_reg_addr, b_reg => dec_b_reg_addr,
             imm         => dec_imm, imm_sign_ext => dec_imm_sign_ext, hold => hold,
             a_reg_value => a_reg_value, b_reg_value => b_reg_value,
             a_reg_used  => compute_a_reg_used, b_reg_used => compute_b_reg_used,
             alu_result  => compute_reg_c_value_out, skip_next_insn => compute_skip_next_insn);

  ls_unit : load_store
    port map (clock           => clock, reset => cpu_reset, skip_ins => ls_skip_ins,
              ins_opp         => dec_opp, ins_imm => dec_imm, ins_imm_sign_ext => dec_imm_sign_ext,
              a_reg_value     => a_reg_value, b_reg_value => b_reg_value, value_read => compute_ls_mem_value_out, busy => ls_busy,
              d_ram_addr      => d_ram_addr, d_ram_dout => d_ram_dout, d_ram_din => d_ram_din, d_ram_we => d_ram_we,
              ext_data_strobe => ext_data_strobe, ext_data_address => ext_data_address, ext_data_out => ext_data_out,
              ext_data_in     => ext_data_in, ext_data_wr => ext_data_wr, ext_data_busy => ext_data_busy);

  --cpu reset is redundant here.
  --if previous instruction was a jump, our results are not going to be used.
  compute_is_bubble_out <= '1' when (compute_is_bubble_in = '1') or (reg_collision = '1') or (cpu_reset = '1') or (jump_set_pc = '1') or (hold = '1') else '0';

  process(clock) is
    variable out_line : line;
  begin
    if rising_edge(clock) then
      if ins_halted /= '1' then

        --if retire_is_bubble_in /= '1' then
        --  write(out_line, string'("RET "));
        --  hwrite(out_line, retire_in_pc_addr);
        --  writeline(output, out_line);
        --end if;

        if compute_is_bubble_out = '0' and compute_skip_next_insn = '1' then
          compute_skip_insn_reg <= '1';
        else
          compute_skip_insn_reg <= '0';
        end if;

        if compute_is_bubble_out = '1' or retire_skip_insn_reg = '1' then
          retire_is_bubble_in <= '1';
        else
          retire_is_bubble_in <= '0';
        end if;

        if compute_is_bubble_out = '0' then
          retire_skip_insn_reg <= compute_skip_next_insn;
        end if;

        retire_in_pc_addr   <= compute_addr_in;
        retire_in_opp_reg   <= dec_opp;
        retire_in_imm_reg   <= dec_imm;
        retire_in_c_addr    <= dec_c_reg_addr;
        retire_in_c_value   <= compute_reg_c_value_out;
        retire_in_mem_value <= compute_ls_mem_value_out;
      end if;  --not halted
    end if;  --clock
  end process;

--3rd stage below

  just_stored_addr <= retire_in_c_addr;
  just_stored      <= '1' when do_reg_c_write = '1' else '0';

--operates at the level of retire, but port a & b addresses are from compute
--state and those values feed back to the compute stage
  regs : reg_bank
    port map(clock       => clock,
             port_a_addr => dec_a_reg_addr, port_a => a_reg_value, port_b_addr => dec_b_reg_addr, port_b => b_reg_value,
             port_c_addr => retire_in_c_addr, port_c => retire_c_value, WE => do_reg_c_write);

  retire_stage : retire
    port map(clock           => clock, reset => cpu_reset, is_bubble => retire_is_bubble_in,
             insn            => retire_in_opp_reg, insn_imm => retire_in_imm_reg, read_value_in => retire_in_mem_value, c_value_in => retire_in_c_value,
             reg_write_value => retire_c_value, do_reg_write => do_reg_c_write,
             do_jump         => jump_set_pc, jump_addr => jump_new_addr, do_halt => ins_halted);

end behavioural;

