-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;

package shader_cpu_pack is
  subtype cpu_inst_type is std_logic_vector(6 downto 0);
  constant INS_NOP   : cpu_inst_type := "0000000";
  constant INS_LUI   : cpu_inst_type := "0000001";
  constant INS_LLI   : cpu_inst_type := "0000010";
  constant INS_ADD   : cpu_inst_type := "0000011";
  constant INS_SUB   : cpu_inst_type := "0000100";
  constant INS_ADDL  : cpu_inst_type := "0000101";
  constant INS_AND   : cpu_inst_type := "0000110";
  constant INS_OR    : cpu_inst_type := "0000111";
  constant INS_XOR   : cpu_inst_type := "0001000";
  constant INS_NOT   : cpu_inst_type := "0001001";
  constant INS_BSET  : cpu_inst_type := "0001010";
  constant INS_BCLR  : cpu_inst_type := "0001011";
  constant INS_RSL   : cpu_inst_type := "0001100";
  constant INS_RSR   : cpu_inst_type := "0001101";
  constant INS_MUL   : cpu_inst_type := "0010000";
  constant INS_CMP   : cpu_inst_type := "0110000";
  constant INS_SRI   : cpu_inst_type := "1000000";
  constant INS_SRR   : cpu_inst_type := "1000001";
  constant INS_LRI   : cpu_inst_type := "1000010";
  constant INS_LRR   : cpu_inst_type := "1000011";
  constant INS_SEQZ  : cpu_inst_type := "1010000";
  constant INS_SNEQZ : cpu_inst_type := "1010001";
  constant INS_SBSET : cpu_inst_type := "1010010";
  constant INS_SBCLR : cpu_inst_type := "1010011";
  constant INS_JI    : cpu_inst_type := "1100000";
  constant INS_JR    : cpu_inst_type := "1100001";
  constant INS_HLT   : cpu_inst_type := "1110000";

  component cpu_dma_unit is
    port(
      clock : in std_logic;
      reset : in std_logic;

      --mu data bus
      mu_bus_addr   : in    std_logic_vector(1 downto 0);
      mu_bus_data   : inout std_logic_vector(31 downto 0);
      mu_bus_write  : in    std_logic;  --if high the MCU is going to write to a mu
      mu_bus_strobe : in    std_logic;
      mu_cpu_irq    : out   std_logic;
      mu_cpu_sel    : in    std_logic;

      --connections to the data BRAM's second port
      bram_addr   : out std_logic_vector(15 downto 0);
      bram_din    : in  std_logic_vector(31 downto 0);
      bram_dout   : out std_logic_vector(31 downto 0);
      bram_we     : out std_logic;
      bram_strobe : out std_logic;

      --connections to the cpu address bus. we have control registers
      --outside the address space of the bram
      --this allows the shader cpu to read and write them
      ext_cpu_strobe : in  std_logic;
      ext_cpu_addr   : in  std_logic_vector(7 downto 0);
      ext_cpu_dout   : out std_logic_vector(31 downto 0);
      ext_cpu_din    : in  std_logic_vector(31 downto 0);
      ext_cpu_we     : in  std_logic;
      ext_cpu_busy   : out std_logic;

      --if the cpu is halted, we allow the command bus
      --to read and write the bram through our port, leaving the 2nd port to the
      --cpu alone
      cb_strobe : in  std_logic;
      cb_addr   : in  std_logic_vector(15 downto 0);
      cb_din    : in  std_logic_vector(31 downto 0);
      cb_dout   : out std_logic_vector(31 downto 0);
      cb_we     : in  std_logic
      );
  end component;

  component shader_cpu is
    port(
      clock    : in std_logic;
      reset_ll : in std_logic;

      logic_reset : in  std_logic;
      logic_halt  : in  std_logic;
      halt_status : out std_logic;

      prog_address : out std_logic_vector(15 downto 0);
      ins_dword    : in  std_logic_vector(31 downto 0);

      d_ram_addr : out std_logic_vector(15 downto 0);
      d_ram_dout : out std_logic_vector(31 downto 0);
      d_ram_din  : in  std_logic_vector(31 downto 0);
      d_ram_we   : out std_logic;

      ext_data_strobe  : out std_logic;
      ext_data_address : out std_logic_vector(7 downto 0);
      ext_data_out     : out std_logic_vector(31 downto 0);
      ext_data_in      : in  std_logic_vector(31 downto 0);
      ext_data_wr      : out std_logic;
      ext_data_busy    : in  std_logic
      );
  end component;

  component alu is
    port(
      opcode        : in  std_logic_vector(6 downto 0);
      bus_a         : in  std_logic_vector(31 downto 0);
      bus_b         : in  std_logic_vector(31 downto 0);
      imm           : in  std_logic_vector(15 downto 0);
      imm_sign_ext  : in  std_logic_vector(31 downto 0);
      bit_shift_out : out std_logic_vector(31 downto 0);
      result        : out std_logic_vector(31 downto 0)
      );
  end component;

  component fetch is
    port(
      clock         : in  std_logic;
      reset         : in  std_logic;
      pc_to_fetch   : out std_logic_vector(15 downto 0);
      ins_fetched   : in  std_logic_vector(31 downto 0);
      new_addr      : in  std_logic_vector(15 downto 0);
      set_pc        : in  std_logic;
      collision_in  : in  std_logic;
      hold_pc       : in  std_logic;
      instruction   : out std_logic_vector(31 downto 0);
      out_is_bubble : out std_logic;
      fetched_addr  : out std_logic_vector(15 downto 0)
      );
  end component;

  component pc is
    port(
      clock : in std_logic;
      reset : in std_logic;

      new_addr : in std_logic_vector(15 downto 0);
      set_pc   : in std_logic;

      hold_pc       : in std_logic;
      hold_pc_value : in std_logic_vector(15 downto 0);

      pc_value : out std_logic_vector(15 downto 0)
      );
  end component;

  component decoder is
    port(
      ins_dword_in : in std_logic_vector(31 downto 0);

      opp        : out std_logic_vector(6 downto 0);
      a_reg_addr : out std_logic_vector(2 downto 0);
      b_reg_addr : out std_logic_vector(2 downto 0);
      c_reg_addr : out std_logic_vector(2 downto 0);

      imm          : out std_logic_vector(15 downto 0);
      imm_sign_ext : out std_logic_vector(31 downto 0)
      );
  end component;

  component compute is
    port(
      clock : in std_logic;
      reset : in std_logic;

      opcode       : in std_logic_vector(6 downto 0);
      a_reg        : in std_logic_vector(2 downto 0);
      b_reg        : in std_logic_vector(2 downto 0);
      imm          : in std_logic_vector(15 downto 0);
      imm_sign_ext : in std_logic_vector(31 downto 0);
      hold         : in std_logic;

      a_reg_value : in std_logic_vector(31 downto 0);
      b_reg_value : in std_logic_vector(31 downto 0);

      a_reg_used     : out std_logic;
      b_reg_used     : out std_logic;
      alu_result     : out std_logic_vector(31 downto 0);
      skip_next_insn : out std_logic
      );
  end component;

  component load_store is
    port(
      clock : in std_logic;
      reset : in std_logic;

      skip_ins         : in std_logic;
      ins_opp          : in std_logic_vector(6 downto 0);
      ins_imm          : in std_logic_vector(15 downto 0);
      ins_imm_sign_ext : in std_logic_vector(31 downto 0);

      a_reg_value : in  std_logic_vector(31 downto 0);
      b_reg_value : in  std_logic_vector(31 downto 0);
      value_read  : out std_logic_vector(31 downto 0);
      busy        : out std_logic;      --somehow must be implemented.

      d_ram_addr : out std_logic_vector(15 downto 0);
      d_ram_dout : out std_logic_vector(31 downto 0);
      d_ram_din  : in  std_logic_vector(31 downto 0);
      d_ram_we   : out std_logic;

      --at upper 8 bits of cpu address space
      ext_data_strobe  : out std_logic;  --perform a read or write
      ext_data_address : out std_logic_vector(7 downto 0);
      ext_data_out     : out std_logic_vector(31 downto 0);
      ext_data_in      : in  std_logic_vector(31 downto 0);
      ext_data_wr      : out std_logic;
      ext_data_busy    : in  std_logic
      );
  end component;

  component retire is
    port(
      clock : in std_logic;
      reset : in std_logic;

      insn      : in std_logic_vector(6 downto 0);
      insn_imm  : in std_logic_vector(15 downto 0);
      is_bubble : in std_logic;

      read_value_in : in std_logic_vector(31 downto 0);
      c_value_in    : in std_logic_vector(31 downto 0);

      reg_write_value : out std_logic_vector(31 downto 0);
      do_reg_write    : out std_logic;

      do_jump   : out std_logic;
      jump_addr : out std_logic_vector;

      do_halt : out std_logic
      );
  end component;


  component std_reg is
    generic (
      REG_ADDR : std_logic_vector(2 downto 0)
      );

    port(
      clock : in std_logic;

      a_out_port : out std_logic_vector(31 downto 0);
      b_out_port : out std_logic_vector(31 downto 0);

      do_write       : in std_logic;
      reg_write_addr : in std_logic_vector(2 downto 0);
      c_in_port      : in std_logic_vector(31 downto 0)
      );
  end component;

  component reg_bank is
    port(
      clock       : in  std_logic;
      port_a_addr : in  std_logic_vector(2 downto 0);
      port_b_addr : in  std_logic_vector(2 downto 0);
      port_a      : out std_logic_vector(31 downto 0);
      port_b      : out std_logic_vector(31 downto 0);
      port_c_addr : in  std_logic_vector(2 downto 0);
      port_c      : in  std_logic_vector(31 downto 0);
      WE          : in  std_logic
      );
  end component;

  component shader_cb_interface is
    port(
      clock : in std_logic;
      reset : in std_logic;

      cmd_bus_data_in  : in  std_logic_vector(31 downto 0);
      cmd_bus_data_out : out std_logic_vector(31 downto 0);
      cmd_bus_addr     : in  std_logic_vector(7 downto 0);
      cmd_bus_wr       : in  std_logic;
      cmd_bus_strobe   : in  std_logic;

      i_ram_strobe : out std_logic;
      d_ram_strobe : out std_logic;

      i_ram_busy : in std_logic;
      d_ram_busy : in std_logic;

      i_ram_data : in std_logic_vector(31 downto 0);
      d_ram_data : in std_logic_vector(31 downto 0);

      ram_wr_data : out std_logic_vector(31 downto 0);
      ram_addr    : out std_logic_vector(15 downto 0);
      ram_we      : out std_logic;

      pc_addr            : in  std_logic_vector(15 downto 0);
      shader_halt_status : in  std_logic;
      shader_reset       : out std_logic;
      shader_halt        : out std_logic
      );
  end component;

  component ram_inst is
    port(
      clock : in std_logic;

      cpu_pc   : in  std_logic_vector(15 downto 0);
      cpu_inst : out std_logic_vector(31 downto 0);

      cb_int_strobe : in  std_logic;
      cb_int_we     : in  std_logic;
      cb_int_addr   : in  std_logic_vector(15 downto 0);
      cb_int_dout   : out std_logic_vector(31 downto 0);
      cb_int_din    : in  std_logic_vector(31 downto 0)
      );
  end component;

  component ram_data is
    port(
      clock : in std_logic;

      cpu_addr : in  std_logic_vector(15 downto 0);
      cpu_din  : in  std_logic_vector(31 downto 0);
      cpu_dout : out std_logic_vector(31 downto 0);
      cpu_we   : in  std_logic;

      dma_int_strobe : in  std_logic;
      dma_int_we     : in  std_logic;
      dma_int_addr   : in  std_logic_vector(15 downto 0);
      dma_int_dout   : out std_logic_vector(31 downto 0);
      dma_int_din    : in  std_logic_vector(31 downto 0)
      );
  end component;

  function M_0(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_1(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_2(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_3(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_4(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_5(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_6(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_7(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_8(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_9(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_A(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_B(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_C(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_D(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_E(sig_in : std_logic_vector(15 downto 0))
    return boolean;

  function M_F(sig_in : std_logic_vector(15 downto 0))
    return boolean;

end;  --package

package body shader_cpu_pack is

  function M_0(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0000000";
  end M_0;

  function M_1(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0000001";
  end M_1;

  function M_2(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0000010";
  end M_2;

  function M_3(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0000011";
  end M_3;

  function M_4(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0000100";
  end M_4;

  function M_5(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0000101";
  end M_5;

  function M_6(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0000110";
  end M_6;

  function M_7(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0000111";
  end M_7;

  function M_8(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0001000";
  end M_8;

  function M_9(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0001001";
  end M_9;

  function M_A(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0001010";
  end M_A;

  function M_B(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0001011";
  end M_B;

  function M_C(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0001100";
  end M_C;

  function M_D(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0001101";
  end M_D;

  function M_E(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0001110";
  end M_E;

  function M_F(sig_in : std_logic_vector(15 downto 0))
    return boolean is
  begin
    return sig_in(15 downto 9) = "0001111";
  end M_F;

end;  --body
