--file created 2006
--modified 2009,2014
--released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.tvc_pack.all;
use work.shader_pack.all;

entity tvc is
  port (
    tvc_clock     : in  std_logic;
    ram_clock     : in  std_logic;
    ram_clock_adv : in  std_logic;
    tvc_reset_in  : in  std_logic;
    tvc_reset_out : out std_logic;      --we hold reset internally much longer

    disp_raw_clk   : in std_logic;
    disp_raw_reset : in std_logic;

    disp1_vgaR  : out std_logic_vector(2 downto 0);
    disp1_vgaG  : out std_logic_vector(2 downto 0);
    disp1_vgaB  : out std_logic_vector(2 downto 1);
    disp1_hSync : out std_logic;
    disp1_vSync : out std_logic;

    --directions are in the direction of the FUs
    cmd_bus_data_in  : in  std_logic_vector(31 downto 0);
    cmd_bus_addr     : in  std_logic_vector(7 downto 0);
    cmd_bus_strobe   : in  std_logic;
    cmd_bus_wr       : in  std_logic;
    cmd_bus_data_out : out std_logic_vector(31 downto 0);
    busy_bits        : out std_logic_vector(31 downto 0);

    mem_addr     : out   std_logic_vector(22 downto 0);
    nx_ram_clk   : out   std_logic;
    nx_ram_adv_n : out   std_logic;
    nx_ram_cre   : out   std_logic;
    nx_ram_ce_n  : out   std_logic;
    nx_ram_oe_n  : out   std_logic;
    nx_ram_we_n  : out   std_logic;
    nx_ram_lb_n  : out   std_logic;
    nx_ram_ub_n  : out   std_logic;
    nx_ram_wait  : in    std_logic;
    mem_dq       : inout std_logic_vector(15 downto 0)
    );
end tvc;

architecture Behavioral of tvc is

  signal tvc_reset       : std_logic;
  signal ram_start_delay : std_logic;
  signal disp1_pix_value : std_logic_vector(7 downto 0);

  --start mcu related signals
  signal mu_bus_address : std_logic_vector(1 downto 0);
  signal mu_bus_data    : std_logic_vector(31 downto 0);
  signal mu_bus_write   : std_logic;
  signal mu_bus_strobe  : std_logic;

  signal odd_sc_irq           : std_logic;
  signal odd_sc_select        : std_logic;
  signal even_sc_irq          : std_logic;
  signal even_sc_select       : std_logic;
  signal block_reader_irq     : std_logic;
  signal block_reader_select  : std_logic;
  signal block_writer_irq     : std_logic;
  signal block_writer_select  : std_logic;
  signal block_set_irq        : std_logic;
  signal block_set_select     : std_logic;
  signal cpu_dma_irq          : std_logic;
  signal cpu_dma_select       : std_logic;
  signal poly_scanline_irq    : std_logic;
  signal poly_scanline_select : std_logic;
  
begin
  busy_bits(31 downto 4) <= X"0000000";

  tvc_reset     <= '1' when (tvc_reset_in = '1') or (ram_start_delay = '1') else '0';
  tvc_reset_out <= tvc_reset;

  disp1_vgaR <= disp1_pix_value(7 downto 5);
  disp1_vgaG <= disp1_pix_value(4 downto 2);
  disp1_vgaB <= disp1_pix_value(1 downto 0);

  --odd_sc_irq <= '0';
  --even_sc_irq <= '0';
  ppusher : pix_pusher
    port map(mu_bus_address  => mu_bus_address, mu_bus_data => mu_bus_data, mu_bus_write => mu_bus_write, mu_bus_strobe => mu_bus_strobe,
             odd_sc_irq      => odd_sc_irq, odd_sc_select => odd_sc_select, even_sc_irq => even_sc_irq, even_sc_select => even_sc_select,
             tvc_clock       => tvc_clock, tvc_reset => tvc_reset, disp_raw_clk => disp_raw_clk, disp_raw_reset => disp_raw_reset, disp_pix_value => disp1_pix_value,
             disp_horiz_sync => disp1_hSync, disp_vert_sync => disp1_vSync);

  mem_control_unit : mcu_struct         --uses non-delayed reset
    port map(nx_ram_addr       => mem_addr, nx_ram_clk => nx_ram_clk, nx_ram_adv_n => nx_ram_adv_n, nx_ram_cre => nx_ram_cre, nx_ram_ce_n => nx_ram_ce_n,
             nx_ram_oe_n       => nx_ram_oe_n, nx_ram_we_n => nx_ram_we_n, nx_ram_lb_n => nx_ram_lb_n, nx_ram_ub_n => nx_ram_ub_n, nx_ram_wait => nx_ram_wait, nx_ram_dq => mem_dq,
             tvc_clock         => tvc_clock, tvc_reset => tvc_reset_in, ram_clock => ram_clock, ram_clock_adv => ram_clock_adv, ram_start_delay => ram_start_delay,
             mu_bus_address    => mu_bus_address, mu_bus_data => mu_bus_data, mu_bus_write => mu_bus_write, mu_bus_strobe => mu_bus_strobe,
             odd_sc_irq        => odd_sc_irq, odd_sc_select => odd_sc_select, even_sc_irq => even_sc_irq, even_sc_select => even_sc_select,
             block_reader_irq  => block_reader_irq, block_reader_select => block_reader_select,
             block_writer_irq  => block_writer_irq, block_writer_select => block_writer_select,
             rp_writer_irq     => block_set_irq, rp_writer_select => block_set_select,
             line_drawer_irq   => cpu_dma_irq, line_drawer_select => cpu_dma_select,
             poly_scanline_irq => poly_scanline_irq, poly_scanline_select => poly_scanline_select);

  --start memory users

  br_mu : block_reader
    port map(mu_bus_address   => mu_bus_address, mu_bus_data => mu_bus_data, mu_bus_write => mu_bus_write, mu_bus_strobe => mu_bus_strobe,
             block_reader_irq => block_reader_irq, block_reader_select => block_reader_select,
             tvc_clock        => tvc_clock, tvc_reset => tvc_reset,
             cmd_bus_data     => cmd_bus_data_in, cmd_bus_addr => cmd_bus_addr, cmd_bus_wr=>cmd_bus_wr, cmd_bus_strobe => cmd_bus_strobe,
             cmd_bus_data_out => cmd_bus_data_out, idle => busy_bits(0));

  bw_mu : block_writer
    port map(mu_bus_address   => mu_bus_address, mu_bus_data => mu_bus_data, mu_bus_write => mu_bus_write, mu_bus_strobe => mu_bus_strobe,
             block_writer_irq => block_writer_irq, block_writer_select => block_writer_select,
             tvc_clock        => tvc_clock, tvc_reset => tvc_reset,
             cmd_bus_data     => cmd_bus_data_in, cmd_bus_addr => cmd_bus_addr, cmd_bus_wr=>cmd_bus_wr, cmd_bus_strobe => cmd_bus_strobe,
             idle             => busy_bits(1));

  bs_mu : block_set
    port map(mu_bus_address => mu_bus_address, mu_bus_data => mu_bus_data, mu_bus_write => mu_bus_write, mu_bus_strobe => mu_bus_strobe,
             block_set_irq  => block_set_irq, block_set_select => block_set_select,
             tvc_clock      => tvc_clock, tvc_reset => tvc_reset,
             cmd_bus_data   => cmd_bus_data_in, cmd_bus_addr => cmd_bus_addr, cmd_bus_wr=>cmd_bus_wr, cmd_bus_strobe => cmd_bus_strobe,
             idle           => busy_bits(2));

  ps_mu : poly_scanline
    port map(mu_bus_address    => mu_bus_address, mu_bus_data => mu_bus_data, mu_bus_write => mu_bus_write, mu_bus_strobe => mu_bus_strobe,
             poly_scanline_irq => poly_scanline_irq, poly_scanline_select => poly_scanline_select, tvc_clock => tvc_clock, tvc_reset => tvc_reset,
             cmd_bus_data      => cmd_bus_data_in, cmd_bus_addr => cmd_bus_addr, cmd_bus_wr=>cmd_bus_wr, cmd_bus_strobe => cmd_bus_strobe,
             idle              => busy_bits(3));

  smb : shader_meta_block
    port map(clock           => tvc_clock, reset => tvc_reset,
             cmd_bus_data_in => cmd_bus_data_in, cmd_bus_data_out => cmd_bus_data_out, cmd_bus_addr => cmd_bus_addr,
             cmd_bus_wr      => cmd_bus_wr, cmd_bus_strobe => cmd_bus_strobe,
             mu_bus_address  => mu_bus_address, mu_bus_data => mu_bus_data, mu_bus_write => mu_bus_write, mu_bus_strobe => mu_bus_strobe,
             mu_cpu_dma_irq  => cpu_dma_irq, mu_cpu_dma_sel => cpu_dma_select);

end Behavioral;
