--file created 2014
--modified
--released to the Public Domain by the author

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--Q&D drive the command bus

entity usb_cb_master is
  port(
    clock : in std_logic;
    reset : in std_logic;

    --status bits of functional units.
    busy_bits        : in  std_logic_vector(31 downto 0);
    cmd_bus_data_in  : in  std_logic_vector(31 downto 0);
    cmd_bus_data_out : out std_logic_vector(31 downto 0);
    cmd_bus_addr     : out std_logic_vector(7 downto 0);
    cmd_bus_wr       : out std_logic;
    cmd_bus_strobe   : out std_logic;

    usb_data_in    : in  std_logic_vector(7 downto 0);
    usb_strobe_in  : in  std_logic;
    usb_in_busy    : out std_logic; --I am not paying attention to the strobe_in line
    usb_data_out   : out std_logic_vector(8 downto 0); --top bit is packet end
    usb_strobe_out : out std_logic;
    usb_out_busy   : in  std_logic
    );
end usb_cb_master;

architecture rtl of usb_cb_master is
  --data we read from control bus or pull from junk_test_reg
  signal usb_to_write_reg : std_logic_vector(31 downto 0);

  --lets me test rw of spi. on cb addres FE
  signal junk_test_reg : std_logic_vector(31 downto 0) := X"12345678";

  type RD_SM is (RD_IDLE, RD_READ_CB, RD_STARTUP, RD_WRITE_1, RD_WRITE_2, RD_WRITE_3, RD_WRITE_4, RD_WRITE_TAIL, RD_BUSY);
  signal cbr_state : RD_SM;

  signal cb_write_strobe : std_logic;
  signal cb_read_strobe  : std_logic;

  signal cb_address    : std_logic_vector(7 downto 0);
  signal cb_write_data : std_logic_vector(31 downto 0);

  signal noticed_strobe : std_logic;
  signal got_data       : std_logic;

  signal packet_accepted : std_logic;
  
  signal bb_0 : std_logic_vector(7 downto 0);  --newest data (tail)
  signal bb_1 : std_logic_vector(7 downto 0);  --data_3
  signal bb_2 : std_logic_vector(7 downto 0);  --data_2
  signal bb_3 : std_logic_vector(7 downto 0);  --data_1
  signal bb_4 : std_logic_vector(7 downto 0);  --data_0
  signal bb_5 : std_logic_vector(7 downto 0);  --address
  signal bb_6 : std_logic_vector(7 downto 0);  --command
  signal bb_7 : std_logic_vector(7 downto 0);  --oldest data (header)
  
begin
  cmd_bus_data_out <= cb_write_data;
  cmd_bus_addr     <= cb_address;
  usb_in_busy <= '1' when usb_strobe_in = '1' or got_data = '1' or cb_write_strobe = '1' or cb_read_strobe = '1' or cbr_state /= RD_IDLE or reset = '1' else '0';
  --usb_in_busy <= '0';
  
  process(clock) is
    variable strobe_got_data : boolean;
  begin
    if rising_edge(clock) then
      strobe_got_data := false;

      if reset = '1' then
        noticed_strobe <= '0';
        bb_0           <= X"00";  --just clear one of a pre-existing magic number.
      else
        if usb_strobe_in = '1' and noticed_strobe = '0' then
          bb_0            <= usb_data_in;
          bb_1            <= bb_0;
          bb_2            <= bb_1;
          bb_3            <= bb_2;
          bb_4            <= bb_3;
          bb_5            <= bb_4;
          bb_6            <= bb_5;
          bb_7            <= bb_6;
          strobe_got_data := true;
          noticed_strobe  <= '1';
        elsif usb_strobe_in = '0' then
          noticed_strobe <= '0';
        end if;
      end if;

      if packet_accepted = '1' then --signal from next level of pipeline.
        bb_6 <= X"00";
      end if;
      
      if strobe_got_data = true then
        got_data <= '1';
      else
        got_data <= '0';
      end if;
    end if;
  end process;

  process(clock) is
    variable do_cb_write_strobe : boolean;
    variable do_cb_read_strobe  : boolean;
    variable do_packet_accepted : boolean;
  begin
    if rising_edge(clock) then
      do_cb_write_strobe := false;
      do_cb_read_strobe  := false;
      do_packet_accepted := false;

      if got_data = '1' then
        if (bb_0 = X"FF") and (bb_7 = X"AA") then  --assume we got a valid command.
          cb_write_data <= bb_1 & bb_2 & bb_3 & bb_4;
          cb_address    <= bb_5;
          --discard this chunk of data for next clock. will have to match AA &
          --FF again
          do_packet_accepted := true;
          if bb_6 = X"F0" then
            do_cb_write_strobe := true;
          elsif bb_6 = X"0F" then
            do_cb_read_strobe := true;
          end if;
        end if;
      end if;

      if do_packet_accepted = true then
        packet_accepted <= '1';
      else
        packet_accepted <= '0';
      end if;
      
      if do_cb_write_strobe = true then
        cb_write_strobe <= '1';
      else
        cb_write_strobe <= '0';
      end if;

      if do_cb_read_strobe = true then
        cb_read_strobe <= '1';
      else
        cb_read_strobe <= '0';
      end if;
    end if;  --clock
  end process;

  --reads take longer than writes. it may be possible
  --for a write to stomp on an in process read.
  process(clock) is
    variable do_cb_write_strobe  : boolean;
    variable do_cb_read_strobe   : boolean;
    variable do_usb_write_strobe : boolean;
    variable next_state          : RD_SM;
    variable after_rd_busy_state : RD_SM;
  begin
    if rising_edge(clock) then
      next_state          := cbr_state;
      do_cb_write_strobe  := false;
      do_cb_read_strobe   := false;
      do_usb_write_strobe := false;

      if reset = '1' then
        next_state := RD_IDLE;
      else
        case cbr_state is
          when RD_IDLE =>
            if cb_write_strobe = '1' then
              if cb_address = X"FE" then
                junk_test_reg <= cb_write_data;
              else
                do_cb_write_strobe := true;
              end if;
            elsif cb_read_strobe = '1' then
              if cb_address = X"FF" then
                usb_to_write_reg <= busy_bits;
                next_state       := RD_STARTUP;
              elsif cb_address = X"FE" then
                usb_to_write_reg <= junk_test_reg;
                next_state       := RD_STARTUP;
              else
                do_cb_read_strobe := true;
                next_state        := RD_READ_CB;
              end if;
            end if;
          when RD_READ_CB =>
            usb_to_write_reg <= cmd_bus_data_in;
            next_state       := RD_STARTUP;
          when RD_STARTUP =>
            usb_data_out        <= '0' & X"AA";
            do_usb_write_strobe := true;
            after_rd_busy_state := RD_WRITE_1;
            next_state          := RD_BUSY;
          when RD_WRITE_1 =>
            usb_data_out        <= '0' & usb_to_write_reg(31 downto 24);
            do_usb_write_strobe := true;
            after_rd_busy_state := RD_WRITE_2;
            next_state          := RD_BUSY;
          when RD_WRITE_2 =>
            usb_data_out        <= '0' & usb_to_write_reg(23 downto 16);
            do_usb_write_strobe := true;
            after_rd_busy_state := RD_WRITE_3;
            next_state          := RD_BUSY;
          when RD_WRITE_3 =>
            usb_data_out        <= '0' & usb_to_write_reg(15 downto 8);
            do_usb_write_strobe := true;
            after_rd_busy_state := RD_WRITE_4;
            next_state          := RD_BUSY;
          when RD_WRITE_4 =>
            usb_data_out        <= '0' & usb_to_write_reg(7 downto 0);
            do_usb_write_strobe := true;
            after_rd_busy_state := RD_WRITE_TAIL;
            next_state          := RD_BUSY;
          when RD_WRITE_TAIL =>
            usb_data_out        <= '1' & X"FF";
            do_usb_write_strobe := true;
            after_rd_busy_state := RD_IDLE;
            next_state          := RD_BUSY;
          when RD_BUSY =>
            if usb_out_busy = '0' then
              next_state := after_rd_busy_state;
            end if;
        end case;
      end if;

      cbr_state <= next_state;

      if do_usb_write_strobe = true then
        usb_strobe_out <= '1';
      else
        usb_strobe_out <= '0';
      end if;
      
      if do_cb_write_strobe = true then
        cmd_bus_strobe <= '1';
        cmd_bus_wr     <= '1';
      elsif do_cb_read_strobe = true then
        cmd_bus_strobe <= '1';
        cmd_bus_wr     <= '0';
      else
        cmd_bus_strobe <= '0';
        cmd_bus_wr     <= '0';
      end if;
    end if;
  end process;

end rtl;
