--file created 2014
--modified
--released to the Public Domain by the author

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.tvc_pack.all;

--timing is designed for an 80MHz clock
--slower clocks fine (but slower)
--usb clock is ignored

entity usb_io_async is
  port(
    clock : in std_logic;
    reset : in std_logic;

    --signals to/from usb
    usb_ifclk   : in    std_logic;
    usb_fd      : inout std_logic_vector(7 downto 0);
    usb_fifo_ad : out   std_logic_vector(1 downto 0);
    usb_flag_a  : in    std_logic;
    usb_flag_b  : in    std_logic;
    usb_flag_c  : in    std_logic;
    usb_int0_n  : out   std_logic;
    usb_slcs_n  : out   std_logic;
    usb_slrd    : out   std_logic;
    usb_slwr    : out   std_logic;
    usb_sloe    : out   std_logic;
    usb_pktend  : out   std_logic;

    --all syncronous to clock above
    --data in (to write to usb)
    fifo2_inbusy  : out std_logic;      --writes are ignored if high
    fifo2_pktend  : in  std_logic;
    fifo2_inbyte  : in  std_logic_vector(7 downto 0);
    fifo2_inflag  : in  std_logic;
    --data out (read from usb)
    fifo0_outbyte : out std_logic_vector(7 downto 0);
    fifo0_outflag : out std_logic;      --strobed only once.
    fifo0_busy    : in  std_logic  --I won't read more usb data if this is high.
    );
end usb_io_async;

architecture rtl of usb_io_async is

  type U_STATE_INT is (U_START, U_WR_1, U_WR_2, U_WR_3, U_WR_4, U_WR_END_1, U_WR_END_2, U_RD_1, U_RD_2, U_RD_3);
  signal c_state : U_STATE_INT := U_START;

  signal tmr : unsigned(3 downto 0);

  signal data_oe            : std_logic;
  signal data_out_reg       : std_logic_vector(7 downto 0);
  signal data_out_is_pktend : std_logic;
  signal inflag_noticed     : std_logic;

  signal fifo0_outbyte_reg : std_logic_vector(7 downto 0);

  signal cdiv      : unsigned(3 downto 0) := (others => '0');
  signal div_clock : std_logic;

  signal usb_flag_a_sync : std_logic;
  signal usb_flag_b_sync : std_logic;
  signal usb_flag_c_sync : std_logic;
  signal usb_fd_sync     : std_logic_vector(7 downto 0);

  --read some usb data if we can accept it
  procedure MaybeStartRead(
    signal fifo0_busy    : in  std_logic;
    variable next_state  : out U_STATE_INT;
    variable tmr_reset   : out boolean;
    signal usb_fifo_addr : out std_logic_vector(1 downto 0)) is
  begin
    next_state    := U_RD_1;
    tmr_reset     := true;              --if going to U_START, we don't care.
    usb_fifo_addr <= "00";
  end procedure;
  
begin
  usb_slcs_n <= '0';                    --always selected
  usb_int0_n <= '1';                    --no irq.

  usb_fd <= data_out_reg when data_oe = '1' else (others => 'Z');

  --we will go busy immediatly following a strobe.
  fifo2_inbusy <= '1' when fifo2_inflag = '1' or inflag_noticed = '1' else '0';

  fifo0_outbyte <= fifo0_outbyte_reg;

  U3 : std_sync port map(sig_in => usb_flag_a, sig_out => usb_flag_a_sync, clock => clock, reset => reset);
  U4 : std_sync port map(sig_in => usb_flag_b, sig_out => usb_flag_b_sync, clock => clock, reset => reset);
  U5 : std_sync port map(sig_in => usb_flag_c, sig_out => usb_flag_c_sync, clock => clock, reset => reset);

  g1 : for b in 0 to 7 generate
    U2 : std_sync port map(sig_in => usb_fd(b), sig_out => usb_fd_sync(b), clock => clock, reset => reset);
  end generate;


  process (clock) is
    variable tmr_reset            : boolean;
    variable next_state           : U_STATE_INT;
    variable strobe_oe            : boolean;
    variable strobe_usb_oe        : boolean;
    variable strobe_fifo0_outflag : boolean;
    variable strobe_pktend        : boolean;
    variable do_usb_wr            : boolean;
    variable do_usb_rd            : boolean;
  begin
    if rising_edge(clock) then
      tmr_reset            := false;
      next_state           := c_state;
      strobe_oe            := false;
      strobe_usb_oe        := false;
      strobe_fifo0_outflag := false;
      strobe_pktend        := false;
      do_usb_wr            := false;
      do_usb_rd            := false;

      if reset = '1' then  --this module probably does not need a reset
        next_state     := U_START;
        inflag_noticed <= '0';
        tmr_reset      := true;
      else
        
        if fifo2_inflag = '1' and inflag_noticed = '0' then
          data_out_reg       <= fifo2_inbyte;
          data_out_is_pktend <= fifo2_pktend;
          inflag_noticed     <= '1';
        end if;

        case c_state is
          when U_START =>
            --we have data to write and space to put it.
            if inflag_noticed = '1' and usb_flag_b_sync = '0' then
              usb_fifo_ad <= "10";
              tmr_reset   := true;
              next_state  := U_WR_1;
            else
              --we could read data if some exists
              MaybeStartRead(fifo0_busy, next_state, tmr_reset, usb_fifo_ad);
            end if;
          when U_WR_1 =>                --holding address line
            if tmr = "1000" then
              next_state := U_WR_2;
              tmr_reset  := true;
            end if;
          when U_WR_2 =>
            --assert write; actual xfer happens on falling edge
            strobe_oe := true;
            do_usb_wr := true;
            if tmr = "1000" then
              next_state := U_WR_3;
              tmr_reset  := true;
            end if;
          when U_WR_3 =>
            strobe_oe := true;
            --continue holding data lines
            --with wr released. data latched on falling edge.
            if tmr = "1000" then
              if data_out_is_pktend = '1' then
                next_state := U_WR_4;
                tmr_reset  := true;
              else
                next_state := U_WR_END_1;
                tmr_reset  := true;
              end if;
            end if;
          when U_WR_4 =>
            strobe_pktend := true;
            if tmr = "1000" then
              tmr_reset  := true;
              next_state := U_WR_END_1;
            end if;
          when U_WR_END_1 =>            --dont tweak address lines for a clock
            if tmr = "1000" then
              next_state := U_WR_END_2;
            end if;
          when U_WR_END_2 =>
            inflag_noticed <= '0';
            MaybeStartRead(fifo0_busy, next_state, tmr_reset, usb_fifo_ad);
          when U_RD_1 =>
            if tmr = "1000" then
              --there is usb data to read we can accept the data.
              if usb_flag_c_sync = '0' and fifo0_busy = '0' then
                next_state := U_RD_2;
                tmr_reset  := true;
              else
                tmr_reset  := true;
                next_state := U_START;
              end if;
            end if;
          when U_RD_2 =>
            strobe_usb_oe := true;
            do_usb_rd     := true;
            if tmr = "1000" then
              fifo0_outbyte_reg    <= usb_fd_sync;
              strobe_fifo0_outflag := true;
              next_state           := U_RD_3;
              tmr_reset            := true;
            end if;
          when U_RD_3 =>
            strobe_fifo0_outflag := true;
            --holding addr while releaseing usb_rd and ub_oe
            if tmr = "1000" then
              tmr_reset  := true;
              next_state := U_START;
            end if;
        end case;
      end if;  --reset

      if strobe_pktend = true then
        usb_pktend <= '1';
      else
        usb_pktend <= '0';
      end if;

      if strobe_fifo0_outflag = true then
        fifo0_outflag <= '1';
      else
        fifo0_outflag <= '0';
      end if;

      if strobe_usb_oe = true then
        usb_sloe <= '1';
      else
        usb_sloe <= '0';
      end if;

      if do_usb_wr = true then
        usb_slwr <= '1';
      else
        usb_slwr <= '0';
      end if;

      if do_usb_rd = true then
        usb_slrd <= '1';
      else
        usb_slrd <= '0';
      end if;

      if strobe_oe = true then
        data_oe <= '1';
      else
        data_oe <= '0';
      end if;

      if tmr_reset = true then
        tmr <= (others => '0');
      else
        tmr <= tmr + 1;
      end if;

      c_state <= next_state;
      
    end if;  --clock
  end process;

end architecture;
