--file created 2014
--modified
--released to the Public Domain by the author

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity usb_master is
  port(
    clk   : in std_logic;
    reset : in std_logic;

    cmd_wr_strobe    : in  std_logic;
    cmd_rd_strobe    : in  std_logic;
    cb_address       : in  std_logic_vector(7 downto 0);
    cb_data_to_write : in  std_logic_vector(31 downto 0);
    cb_read_data     : out std_logic_vector(31 downto 0);

    usb_busy : out std_logic;

    usb_fd     : inout std_logic_vector(7 downto 0);
    usb_faddr  : in    std_logic_vector(1 downto 0);
    usb_slrd   : in    std_logic;
    usb_slwr   : in    std_logic;
    usb_sloe   : in    std_logic;
    usb_slcs   : in    std_logic;
    usb_int0   : in    std_logic;
    usb_pktend : in    std_logic;
    usb_flaga  : out   std_logic;
    usb_flagb  : out   std_logic;
    usb_flagc  : out   std_logic;
    usb_ifclk  : out   std_logic
    );
end usb_master;

architecture tb of usb_master is
  signal usb_ifclk_i : std_logic := '0';

  signal have_data_to_write : std_logic := '0';

  signal out_data_reg : std_logic_vector(7 downto 0);

  signal write_seen : std_logic;
  signal c_w_byte   : unsigned(3 downto 0) := X"0";
  signal w_reg_0    : std_logic_vector(7 downto 0);  --command 0xF0 wr ; 0x0F rd
  signal w_reg_1    : std_logic_vector(7 downto 0);  --cb address
  signal w_reg_2    : std_logic_vector(7 downto 0);  --data
  signal w_reg_3    : std_logic_vector(7 downto 0);  --data
  signal w_reg_4    : std_logic_vector(7 downto 0);  --data
  signal w_reg_5    : std_logic_vector(7 downto 0);  --data

  signal read_seen : std_logic;
  signal c_r_byte  : unsigned(3 downto 0)          := X"0";
  signal read_reg  : std_logic_vector(31 downto 0) := (others => '0');
  
begin
  usb_ifclk <= usb_ifclk_i;
  usb_fd    <= out_data_reg when usb_sloe = '1'           else (others => 'Z');
  usb_flaga <= '0';
  usb_flagb <= '0';  --we have space availiable for a device write
  usb_flagc <= '0'          when have_data_to_write = '1' else '1';  --we have data availiable for a device read

  usb_busy <= '1' when (have_data_to_write = '1') or (cmd_rd_strobe = '1') or (cmd_wr_strobe = '1') else '0';

  cb_read_data <= read_reg;

  usb_clock : process
  begin
    --ugly but pauses output
    wait for 600 us;
    wait for 0.901 us;  --hopefully async enough for a good test
    usbclkloop : loop
      wait for 11.11 ns;
      usb_ifclk_i <= not usb_ifclk_i;
    end loop usbclkloop;
    wait;
  end process;


  usb_read : process (clk) is
  begin
    if rising_edge(clk) then

      if (have_data_to_write = '0') then
        if cmd_wr_strobe = '1' then
          w_reg_0            <= X"F0";
          w_reg_1            <= cb_address;
          w_reg_2            <= cb_data_to_write(7 downto 0);
          w_reg_3            <= cb_data_to_write(15 downto 8);
          w_reg_4            <= cb_data_to_write(23 downto 16);
          w_reg_5            <= cb_data_to_write(31 downto 24);
          have_data_to_write <= '1';
        elsif cmd_rd_strobe = '1' then
          w_reg_0            <= X"0F";
          w_reg_1            <= cb_address;
          w_reg_2            <= (others => '1');
          w_reg_3            <= (others => '1');
          w_reg_4            <= (others => '1');
          w_reg_5            <= (others => '1');
          have_data_to_write <= '1';
        end if;
      end if;

      --a read should not happen unless have_data_to_write = '1'
      if usb_slrd = '1' and read_seen = '0' then
        read_seen <= '1';

        case c_w_byte is
          when X"0" =>
            out_data_reg <= X"AA";
          when X"1" =>
            out_data_reg <= w_reg_0;    --write
          when X"2" =>
            out_data_reg <= w_reg_1;    --reg
          when X"3" =>
            out_data_reg <= w_reg_2;
          when X"4" =>
            out_data_reg <= w_reg_3;
          when X"5" =>
            out_data_reg <= w_reg_4;
          when X"6" =>
            out_data_reg <= w_reg_5;
          when X"7" =>
            out_data_reg <= X"FF";
          when others =>
            null;
        end case;
      elsif usb_slrd = '0' and read_seen = '1' then
        if c_w_byte = X"7" then
          have_data_to_write <= '0';
          c_w_byte           <= X"0";
        else
          c_w_byte <= c_w_byte + 1;
        end if;
        read_seen <= '0';
      elsif usb_slrd = '0' then
        read_seen <= '0';
      end if;
    end if;
  end process;

  usb_write : process (usb_ifclk_i) is
  begin
    if rising_edge(usb_ifclk_i) then

      if usb_slwr = '1' and write_seen = '0' then
        write_seen <= '1';
      elsif usb_slwr = '0' and write_seen = '1' then
        --data strobed on falling edge of write
        case c_r_byte is
          when X"0" =>
            assert usb_fd = X"AA" report "first read byte not AA" severity warning;
          when X"1" =>
            read_reg(31 downto 24) <= usb_fd;  --write
          when X"2" =>
            read_reg(23 downto 16) <= usb_fd;  --reg
          when X"3" =>
            read_reg(15 downto 8) <= usb_fd;
          when X"4" =>
            read_reg(7 downto 0) <= usb_fd;
          when X"5" =>
            assert usb_fd = X"FF" report "last read byte not FF" severity warning;
          when others =>
            null;
        end case;

        if c_w_byte = X"5" then
          c_r_byte <= X"0";
        else
          c_r_byte <= c_r_byte + 1;
        end if;
        write_seen <= '0';
      elsif usb_slwr = '0' then
        write_seen <= '0';
      end if;
    end if;
  end process;

end tb;
